/**
 * Copyright (c) 2005-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.ComponentContext;
import javax.jbi.component.ComponentLifeCycle;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.ObjectName;
import javax.xml.namespace.QName;

import org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.Description;
import org.ow2.easywsdl.wsdl.api.Endpoint;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.Interceptor;
import org.ow2.petals.component.framework.api.PetalsComponent;
import org.ow2.petals.component.framework.api.Wsdl;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorException;
import org.ow2.petals.component.framework.jbidescriptor.JBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.ComponentInterceptor;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.jbidescriptor.generated.Settablelong;
import org.ow2.petals.component.framework.mbean.InterceptorManager;
import org.ow2.petals.component.framework.mbean.MBeanHelper;
import org.ow2.petals.component.framework.mbean.RuntimeConfigurationNotifier;
import org.ow2.petals.component.framework.monitoring.Monitoring;
import org.ow2.petals.component.framework.process.JBIAcceptorManager;
import org.ow2.petals.component.framework.process.JBIProcessorManager;
import org.ow2.petals.component.framework.process.async.AsyncMessageManager;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.component.framework.util.ServiceEndpointKey;
import org.ow2.petals.component.framework.util.WSDLUtilImpl;
import org.ow2.petals.probes.api.exceptions.MultipleProbesFactoriesFoundException;
import org.ow2.petals.probes.api.exceptions.NoProbesFactoryFoundException;
import org.ow2.petals.probes.api.exceptions.ProbeException;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;

import com.ebmwebsourcing.easycommons.lang.StringHelper;

/**
 * An abstract class that can be used to easily create a new JBI Component.
 * <p>
 * It implements the Component interface.<br />
 * It allow to manage directly endpoints.<br />
 * To do this, simply put into the resources of the project a valid wsdl named
 * 'component.wsdl'
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 */
public abstract class AbstractComponent implements PetalsComponent {

    /**
     * Default value of the sampling duration used by response time probe.
     */
    private final static long DEFAULT_MONITORING_SAMPLING_PERIOD = 300;

    /**
     * The Component's Context
     */
    protected ComponentContext context;

    /**
     * The message exchange acceptor manager
     */
    private JBIAcceptorManager acceptorManager;

    /**
     * The asynchronous message handler
     */
    private AsyncMessageManager asyncManager;

    /**
     * The component's delivery channel
     */
    private DeliveryChannel channel;

    /**
     * The component specific configuration in extensions form.
     */
    private ConfigurationExtensions componentExtensions;

    /**
     * The list of endpoints, managed directly by the component
     */
    private Map<ServiceEndpointKey, ServiceEndpoint> componentServiceEndpoints;

    /**
     * Manage the interceptors at runtime, exposed as MBean
     */
    private InterceptorManager interceptorManager;

    /**
     * All the component interceptors. Key is the name of the interceptor
     */
    private Map<String, Interceptor> interceptors;

    /**
     * The component configuration
     */
    private Jbi jbiComponentConfiguration;

    /**
     * The component's logger
     */
    private Logger logger;

    /**
     * The description and the document of the native WSDL managed internally by
     * the component
     */
    private Wsdl nativeWsdl;

    /**
     * The processor manager in charge of handling JBI Message Exchanges
     */
    private JBIProcessorManager processorManager;

    /**
     * Manage runtime configuration notifications, exposed as MBean
     */
    private RuntimeConfigurationNotifier runtimeConfigurationNotifier;

    /**
     * The component's service unit manager
     */
    private AbstractServiceUnitManager suManager;

    /**
     * The timer used as sampler by response time probes
     */
    private Timer probesTimer;

    /**
     * The monitoring MBean
     */
    private Monitoring monitoringMbean;

    /**
     * The object name of the monitoring MBean
     */
    private ObjectName monitoringMbeanName;

    public AsyncMessageManager getAsyncManager() {
        return this.asyncManager;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#getChannel()
     */
    public DeliveryChannel getChannel() {
        return this.channel;
    }

    public org.ow2.petals.component.framework.jbidescriptor.generated.Component getComponentConfiguration() {
        return this.jbiComponentConfiguration.getComponent();
    }

    public ConfigurationExtensions getComponentExtensions() {
        return this.componentExtensions;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#getContext()
     */
    public ComponentContext getContext() {
        return this.context;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#getExtensionMBeanName()
     */
    public ObjectName getExtensionMBeanName() {
        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.Component#getInterceptor(java.lang
     * .String)
     */
    public Interceptor getInterceptor(final String name) {
        Interceptor interceptor = null;

        if (this.interceptors != null) {
            interceptor = this.interceptors.get(name);
        }
        return interceptor;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#getInterceptors()
     */
    public Map<String, Interceptor> getInterceptors() {
        return this.interceptors;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#getLifeCycle()
     */
    public ComponentLifeCycle getLifeCycle() {
        return this;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#getLogger()
     */
    public final Logger getLogger() {
        return this.logger;
    }

    /**
     * Get the nativeWsdl
     * 
     * @return the nativeWsdl
     */
    public Wsdl getNativeWsdl() {
        return this.nativeWsdl;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.Component#getServiceDescription
     * (javax.jbi.servicedesc.ServiceEndpoint)
     */
    public Document getServiceDescription(final ServiceEndpoint endpoint) {
        if (endpoint == null) {
            return null;
        }
        
        if ((this.componentServiceEndpoints != null)
                && !this.componentServiceEndpoints.isEmpty()
                && (this.componentServiceEndpoints.get(new ServiceEndpointKey(endpoint
                        .getServiceName(), endpoint.getEndpointName())) != null)) {

            if ((this.nativeWsdl != null)
                    && WSDLUtilImpl.isDescriptionContaining(
                            this.nativeWsdl.getDescription(), endpoint)) {
                return this.nativeWsdl.getDocument();
            }
        }

        return this.suManager.getServiceDescription(endpoint);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.Component#getServiceUnitManager()
     */
    public javax.jbi.component.ServiceUnitManager getServiceUnitManager() {
        return this.suManager;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.Component#init(javax.jbi.component
     * .ComponentContext)
     */
    public final void init(final ComponentContext compContext) throws JBIException {
        this.context = compContext;
        this.logger = this.context.getLogger("", this.getResourceBundleName());
        this.logger.info("Init component...");
        this.channel = this.context.getDeliveryChannel();
        // Get the component configuration
        final File jbiXmlFile = new File(compContext.getInstallRoot() + File.separator
                + Constants.Component.META_INF + File.separator + Constants.Component.JBI_XML);
        try {
            // CDK configuration
            final FileInputStream jbiInputStream = new FileInputStream(jbiXmlFile);
            try {
                this.jbiComponentConfiguration = JBIDescriptorBuilder
                        .buildJavaJBIDescriptor(jbiInputStream);
            } finally {
                try {
                    jbiInputStream.close();
                } catch (final IOException e) {
                    // do nothing
                }
            }

            // Component specific configuration
            if ((this.jbiComponentConfiguration.getComponent().getPropertiesFile() != null)
                    && !StringHelper.isNullOrEmpty(this.jbiComponentConfiguration.getComponent()
                            .getPropertiesFile().getValue())) {
                this.componentExtensions = new ConfigurationExtensions(
                        this.jbiComponentConfiguration.getComponent().getAny(),
                        this.jbiComponentConfiguration.getComponent().getPropertiesFile()
                                .getValue());
            } else {
                this.componentExtensions = new ConfigurationExtensions(
                        this.jbiComponentConfiguration.getComponent().getAny());
            }
            this.runtimeConfigurationNotifier = new RuntimeConfigurationNotifier(
                    this.jbiComponentConfiguration, jbiXmlFile, this.logger);
            this.interceptors = this.loadInterceptors();
            this.interceptorManager = new InterceptorManager(this, this.jbiComponentConfiguration,
                    jbiXmlFile);

            // Initialize the response time probe sampler. MUST be done before
            // the creation of the monitoring MBean that requires it.
            this.probesTimer = new Timer(this.context.getComponentName() + " - Monitoring sampler",
                    true);

            // create monitoring mbean and probes
            try {
                this.monitoringMbean = this.createMonitoringMBean();
            } catch (final MultipleProbesFactoriesFoundException e1) {
                this.getLogger().log(Level.WARNING,
                        "Technical monitoring deactivated because of the following error", e1);
            } catch (final NoProbesFactoryFoundException e1) {
                this.getLogger().log(Level.WARNING,
                        "Technical monitoring deactivated because of the following error", e1);
            }

            // Necessary resources must be instantiated only once during the
            // life-cycle of the component
            if (this.suManager == null) {
                this.suManager = this.createServiceUnitManager();
            }
            final File wsdlFile = new File(compContext.getWorkspaceRoot() + File.separator
                    + Constants.Component.NATIVE_SERVICE_FOLDER + File.separator
                    + Constants.Component.COMPONENT_WSDL);
            if (wsdlFile != null && wsdlFile.exists()) {
                final Description description = WSDLUtilImpl.createWsdlDescription(
                        wsdlFile.toURI().toURL());
                this.nativeWsdl = new ComponentWsdl(description);
            }

            // initialize the service endpoint list managed by the component
            this.componentServiceEndpoints = new ConcurrentHashMap<ServiceEndpointKey, ServiceEndpoint>();

        } catch (final CDKJBIDescriptorException e) {
            throw new JBIException(e);
        } catch (final FileNotFoundException e) {
            throw new JBIException(e);
        } catch (final PEtALSCDKException e) {
            throw new JBIException(e);
        } catch (final WSDLException e) {
            throw new JBIException(e);
        } catch (MalformedURLException e) {
            throw new JBIException(e);
        }

        this.logger.fine("Calling initialization method on component implementation...");
        this.doInit();
        this.logger.fine("Specific initialization done.");

        try {
            this.monitoringMbean.init();
        } catch (final ProbeException e) {
            this.getLogger().warning("Error initializing probes: " + e.getMessage());
        }

        // Since the worker manager instantiates the JBIListeners, we must
        // initialize it after the component specific initialization. This is
        // because the JBIListener implementation can access to the specific
        // component implementation which is done in the previous doInit()
        // call.
        this.processorManager = new JBIProcessorManager(this, this.runtimeConfigurationNotifier,
                this.logger);
        this.processorManager.init();
        this.acceptorManager = new JBIAcceptorManager(this, this.processorManager,
                this.runtimeConfigurationNotifier, this.logger);
        this.acceptorManager.init();
        this.asyncManager = new AsyncMessageManager(this, this.processorManager, this.logger);

        // Registers the monitoring MBean
        this.monitoringMbeanName = this.context.getMBeanNames().createCustomComponentMBeanName(
                "monitoring_" + this.context.getComponentName());
        MBeanHelper.registerMBean(this.context.getMBeanServer(), this.monitoringMbeanName,
                this.monitoringMbean);

        this.logger.info("Component initialized");
    }

    /**
     * Create the monitoring Mbean
     * 
     * @return The monitoring Mbean
     * @throws NoProbesFactoryFoundException
     *             No probe implementation found in the classloader
     * @throws MultipleProbesFactoriesFoundException
     *             Several probe implementations found in the classloader
     */
    protected Monitoring createMonitoringMBean() throws MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        return new Monitoring(this.getProbesTimer(), this.getResponseTimeProbeSamplePeriod());
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.Component#isExchangeWithConsumerOkay
     * (javax.jbi.servicedesc.ServiceEndpoint,
     * javax.jbi.messaging.MessageExchange)
     */
    public boolean isExchangeWithConsumerOkay(final ServiceEndpoint endpoint,
            final MessageExchange exchange) {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.Component#isExchangeWithProviderOkay
     * (javax.jbi.servicedesc.ServiceEndpoint,
     * javax.jbi.messaging.MessageExchange)
     */
    public boolean isExchangeWithProviderOkay(final ServiceEndpoint endpoint,
            final MessageExchange exchange) {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.Component#resolveEndpointReference
     * (org.w3c.dom.DocumentFragment)
     */
    public ServiceEndpoint resolveEndpointReference(final DocumentFragment epr) {
        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ComponentLifeCycle#shutDown()
     */
    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#shutDown()
     */
    public final void shutDown() throws JBIException {
        this.logger.info("Shut down component...");
        try {
            this.logger.fine("Calling specific shut down...");
            this.doShutdown();
            this.logger.fine("Specific shut down done.");
            this.processorManager.shutdown();
            this.acceptorManager.shutdown();
            this.asyncManager.shutdown();

            // Shutdown the probes
            try {
                this.monitoringMbean.shutdown();
            } catch (final ProbeException e) {
                this.getLogger().warning("Error shutdowning probes: " + e.getMessage());
            }
            MBeanHelper.unregisterMBean(this.context.getMBeanServer(), this.monitoringMbeanName);

            // Free the response time probe sampler
            this.probesTimer.cancel();

        } finally {
            // remove unnecessary resources
            this.nativeWsdl = null;
            this.channel = null;
            this.runtimeConfigurationNotifier = null;
            this.interceptors = null;
            this.interceptorManager = null;
            this.acceptorManager = null;
            this.processorManager = null;
            this.asyncManager = null;
        }
        this.logger.info("Component shut down");
    }

    @Override
    protected void finalize() throws Throwable {
        JBIDescriptorBuilder.close();
        super.finalize();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#start()
     */
    public final void start() throws JBIException {
        this.logger.info("Start component...");
        if (this.channel == null) {
            this.channel = this.context.getDeliveryChannel();
        }
        this.registerRuntimeConfigMBean();
        this.registerInterceptorManagerMBean();
        this.processorManager.start();
        this.acceptorManager.start();
        this.asyncManager.start();
        this.activateServiceEndpointList();
        this.logger.fine("Calling specific start...");
        this.doStart();

        // Start the probes. Probes are started in last position to have all
        // components already started
        try {
            this.monitoringMbean.start();
        } catch (final ProbeException e) {
            this.getLogger().warning("Error starting probes: " + e.getMessage());
        }

        this.logger.info("Component started");
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#stop()
     */
    public final void stop() throws JBIException {
        this.logger.info("Stop component...");
        try {

            // Stop the probes. Probes are stopped in first position to have all
            // components always running if needed
            try {
                this.monitoringMbean.stop();
            } catch (ProbeException e) {
                this.getLogger().warning("Error stopping probes: " + e.getMessage());
            }

            this.logger.fine("Calling specific stop...");
            this.doStop();
            this.logger.fine("Specific stop done.");
        } finally {
            this.deactivateServiceEndpointList();
            this.unregisterRuntimeConfigurationMBean();
            this.unregisterInterceptorManagerMBean();
            // stop first the acceptor Manager as it aggregates the worker
            // manager
            this.acceptorManager.stop();
            this.processorManager.stop();
            this.asyncManager.stop();
        }
        this.logger.info("Component stopped");
    }

    /**
     * Can be overridden to provide a specific service unit manager
     * 
     * @return the service unit manager used by this component.
     */
    protected abstract AbstractServiceUnitManager createServiceUnitManager();

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.doInit(). This method is called at the end of the component
     * classical initialization. When this method is called, the component's
     * delivery channel, context, logger and service unit manager are available.
     * 
     */
    protected void doInit() throws JBIException {
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.doShutdown(). This method is called at the end of the
     * component classical shutdown.
     * 
     */
    protected void doShutdown() throws JBIException {
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.doStart(). This method is called at the end of the component
     * classical start.
     * 
     */
    protected void doStart() throws JBIException {
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.doStop(). This method is called at the end of the component
     * classical stop.
     * 
     */
    protected void doStop() throws JBIException {
    }

    /**
     * Can be overridden to return a custom resource bundle name
     * 
     * @return the resource bundle name or null if no resource bundle name has
     *         to be used (default value)
     */
    protected String getResourceBundleName() {
        return null;
    }

    /**
     * Store wsdl and then activate endpoint in the JBI context
     * 
     * @param serviceName
     * @param endpointName
     * @throws JBIException
     */
    public ServiceEndpoint activateEndpoint(final QName serviceName, final String endpointName)
            throws JBIException {

        // Fake EP for method getServicedescription...
        ServiceEndpointKey key = new ServiceEndpointKey(serviceName, endpointName);
        this.componentServiceEndpoints.put(key, new ComponentServiceEndpoint(serviceName,
                endpointName));
        final ServiceEndpoint se = this.context.activateEndpoint(serviceName, endpointName);
        this.componentServiceEndpoints.put(key, se);
        return se;
    }

    /**
     * Activate the internal service endpoint list managed directly by the
     * component. It contains the specific endpoints.
     * 
     * @throws JBIException
     */
    private void activateServiceEndpointList() throws JBIException {
        QName serviceName = null;
        String endpointName = null;
        List<Endpoint> endpointList = null;

        if (this.nativeWsdl != null) {
            endpointList = WSDLUtilImpl.getEndpointList(
                    this.nativeWsdl.getDescription());
            if (endpointList != null) {
                final Iterator<Endpoint> iterator = endpointList.iterator();
                if (iterator != null) {
                    while (iterator.hasNext()) {
                        final Endpoint endpoint = iterator.next();
                        if (endpoint != null) {
                            serviceName = endpoint.getService().getQName();
                            endpointName = endpoint.getName();
                            this.activateEndpoint(serviceName, endpointName);
                        }
                    }
                }
            }
        }
    }

    /**
     * Deactivate the internal service endpoint list managed directly by the
     * component.
     * 
     * @throws JBIException
     */
    private void deactivateServiceEndpointList() throws JBIException {
        if (this.componentServiceEndpoints != null) {
            for (final ServiceEndpointKey key : this.componentServiceEndpoints.keySet()) {
                this.deactivateEndpoint(this.componentServiceEndpoints.get(key));
            }
            // finally clear remaining endpoints if all is ok...
            this.componentServiceEndpoints.clear();
        }
    }

    public void deactivateEndpoint(final ServiceEndpoint serviceEndpoint) throws JBIException {
        // remove from container
        this.context.deactivateEndpoint(serviceEndpoint);
        // remove from local cache
        this.componentServiceEndpoints.remove(new ServiceEndpointKey(serviceEndpoint
                .getServiceName(), serviceEndpoint.getEndpointName()));
    }

    /**
     * Load the interceptors if there is any.
     * 
     * @param configuration
     */
    private final Map<String, Interceptor> loadInterceptors() {
        final Map<String, Interceptor> interceptors = new HashMap<String, Interceptor>();

        if (this.jbiComponentConfiguration.getComponent().getComponentInterceptors() != null) {
            for (final ComponentInterceptor componentInterceptor : this.jbiComponentConfiguration
                    .getComponent().getComponentInterceptors().getInterceptor()) {
                try {
                    this.logger.info("Loading interceptor : " + componentInterceptor.getName());
                    Class<?> clazz = Thread.currentThread().getContextClassLoader()
                            .loadClass(componentInterceptor.getClazz());
                    final Interceptor itc = (Interceptor) clazz.newInstance();

                    ConfigurationExtensions configurationExtensions;
                    if ((this.jbiComponentConfiguration.getComponent().getPropertiesFile() != null)
                            && !StringHelper.isNullOrEmpty(this.jbiComponentConfiguration
                                    .getComponent().getPropertiesFile().getValue())) {
                        configurationExtensions = new ConfigurationExtensions(
                                componentInterceptor.getParam(), this.jbiComponentConfiguration
                                        .getComponent().getPropertiesFile().getValue());
                    } else {
                        configurationExtensions = new ConfigurationExtensions(
                                componentInterceptor.getParam());
                    }

                    itc.init(this, componentInterceptor.getName(), componentInterceptor.isActive(),
                            configurationExtensions, this.logger);
                    interceptors.put(componentInterceptor.getName(), itc);
                } catch (final PEtALSCDKException e) {
                    this.logger.log(Level.WARNING, "Could not load or initialize interceptor '"
                            + componentInterceptor.getName() + "' from class '"
                            + componentInterceptor.getClazz()
                            + "', this interceptor is skipped. Cause : ", e);
                } catch (ClassNotFoundException e) {
                    this.logger.log(Level.WARNING, "Could not load or initialize interceptor '"
                            + componentInterceptor.getName() + "' from class '"
                            + componentInterceptor.getClazz()
                            + "', this interceptor is skipped. Cause : ", e);
                } catch (InstantiationException e) {
                    this.logger.log(Level.WARNING, "Could not load or initialize interceptor '"
                            + componentInterceptor.getName() + "' from class '"
                            + componentInterceptor.getClazz()
                            + "', this interceptor is skipped. Cause : ", e);
                } catch (IllegalAccessException e) {
                    this.logger.log(Level.WARNING, "Could not load or initialize interceptor '"
                            + componentInterceptor.getName() + "' from class '"
                            + componentInterceptor.getClazz()
                            + "', this interceptor is skipped. Cause : ", e);
                } catch (ClassCastException e) {
                    this.logger.log(Level.WARNING, "Could not load or initialize interceptor '"
                            + componentInterceptor.getName() + "' from class '"
                            + componentInterceptor.getClazz()
                            + "', this interceptor is skipped. Cause : ", e);
                }
            }
        }

        return interceptors;
    }

    /**
     * 
     * @throws JBIException
     */
    private void registerInterceptorManagerMBean() throws JBIException {
        final ObjectName mbeanName = this.context.getMBeanNames().createCustomComponentMBeanName(
                "interceptors_" + this.context.getComponentName());

        this.interceptorManager.setMBeanName(mbeanName);

        MBeanHelper
                .registerMBean(this.context.getMBeanServer(), mbeanName, this.interceptorManager);
    }

    /**
     * Register the {@link RuntimeConfigurationMBean}
     * 
     * @throws JBIException
     */
    private void registerRuntimeConfigMBean() throws JBIException {
        final ObjectName runtimeConfigurationMBeanName = this.context.getMBeanNames()
                .createCustomComponentMBeanName(
                        "runtime_configuration_" + this.context.getComponentName());

        this.runtimeConfigurationNotifier.setMBeanName(runtimeConfigurationMBeanName);

        MBeanHelper.registerMBean(this.context.getMBeanServer(), runtimeConfigurationMBeanName,
                this.runtimeConfigurationNotifier);
    }

    /**
     * 
     * @throws JBIException
     */
    private void unregisterInterceptorManagerMBean() throws JBIException {
        MBeanHelper.unregisterMBean(this.context.getMBeanServer(),
                this.interceptorManager.getMBeanName());
    }

    /**
     * Unregister the runtime configuration MBean
     * 
     * @throws JBIException
     */
    private void unregisterRuntimeConfigurationMBean() throws JBIException {
        MBeanHelper.unregisterMBean(this.context.getMBeanServer(),
                this.runtimeConfigurationNotifier.getMBeanName());
    }
    
    /**
     * Return the JBI processor manager
     * 
     * @return the JBI processor manager
     */
    public JBIProcessorManager getProcessorManager() {
        return this.processorManager;
    }

    /**
     * @return The timer used as sampler by response time probes
     */
    public Timer getProbesTimer() {
        return this.probesTimer;
    }

    public Monitoring getMonitoringBean() {
        return this.monitoringMbean;
    }

    /**
     * @return The sampling duration used by response time probes, in
     *         milliseconds
     */
    public long getResponseTimeProbeSamplePeriod() {
        final Settablelong value = this.getComponentConfiguration().getMonitoringSamplingPeriod();
        return (value == null ? DEFAULT_MONITORING_SAMPLING_PERIOD : value.getValue()) * 1000;
    }

    private static class ComponentServiceEndpoint implements ServiceEndpoint {
        private QName s = null;

        private String e = null;

        public ComponentServiceEndpoint(final QName serviceName, final String endpointName) {
            this.s = serviceName;
            this.e = endpointName;
        }

        public DocumentFragment getAsReference(final QName operationName) {
            return null;
        }

        public String getEndpointName() {
            return this.e;
        }

        public QName[] getInterfaces() {
            return null;
        }

        public QName getServiceName() {
            return this.s;
        }
    }
}
