/**
 * Copyright (c) 2007-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.InstallationContext;
import javax.management.ObjectName;

import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorException;
import org.ow2.petals.component.framework.jbidescriptor.JBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;

/**
 * A simple bootstrap implementation.
 * <p>
 * This bootstrap just copy the JBI extensions in a properties file that will be used by components.
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Roland Naudin - EBM WebSourcing
 */
public class JBIBootstrap implements javax.jbi.component.Bootstrap {

    private static final String LOGGER_BOOTSTRAP_PREFIX = "bootstrap.";

    /**
     * The component installation context
     */
    protected InstallationContext installContext;

    /**
     * The logger
     */
    private Logger logger;

    /**
     * The component configuration
     */
    private Jbi jbiComponentConfiguration;

    /**
     * Returns null by default as this class does not provides JMX extension
     * MBean
     * 
     * @see javax.jbi.component.ComponentLifeCycle#getExtensionMBeanName()
     */
    public ObjectName getExtensionMBeanName() {
        return null;
    }

    /**
     * @return the componentConfiguration
     */
    public Jbi getJbiComponentConfiguration() {
        return this.jbiComponentConfiguration;
    }

    /**
     * 
     * @see javax.jbi.component.Bootstrap#init(javax.jbi.component.
     *      InstallationContext)
     */
    public void init(InstallationContext installContext) throws JBIException {
        this.installContext = installContext;
        try {
            // the restriction on the method getLogger has been unactivated in
            // PEtALS container to be able to manage the logger of the bootstrap
            this.logger = this.installContext.getContext().getLogger(
                    LOGGER_BOOTSTRAP_PREFIX + installContext.getComponentName(), null);
        } catch (IllegalStateException e) {
            // not a PEtALS container
            this.logger = Logger.getLogger(LOGGER_BOOTSTRAP_PREFIX
                    + installContext.getComponentName());
        }

        if (installContext.isInstall()) {
            // We are in an installation context

            // Get the component configuration
            final File jbiXmlFile = new File(installContext.getInstallRoot() + File.separator
                    + Constants.Component.META_INF + File.separator + Constants.Component.JBI_XML);
            try {
                FileInputStream jbiInputStream = new FileInputStream(jbiXmlFile);
                try {
                    this.jbiComponentConfiguration = JBIDescriptorBuilder
                            .buildJavaJBIDescriptor(jbiInputStream);
                } finally {
                    jbiInputStream.close();
                }
            } catch (final CDKJBIDescriptorException e) {
                throw new JBIException(e);
            } catch (IOException e) {
                throw new JBIException(e);
            }

            // do additional configuration from component implementation. (
            // this method is overridden in the default bootstrap)
            doInit();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Bootstrap#onInstall()
     */
    public void onInstall() throws JBIException {
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Bootstrap#onUninstall()
     */
    public void onUninstall() throws JBIException {
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Bootstrap#cleanUp()
     */
    public void cleanUp() throws JBIException {
    }

    public InstallationContext getInstallContext() {
        return installContext;
    }

    public Logger getLogger() {
        return logger;
    }

    /**
     * Can be overridden.
     * 
     * @throws Exception
     */
    protected void doInit() throws JBIException {
    }

    @Override
    protected void finalize() throws Throwable {
        JBIDescriptorBuilder.close();
        super.finalize();
    }

}
