/**
 * Copyright (c) 2007-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.interceptor;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import org.ow2.petals.component.framework.api.Interceptor;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.jbidescriptor.generated.SUInterceptor;

/**
 * A manager of interceptors.
 * <p>
 * One message interceptor manager per Consume / Provide extension.
 * </p>
 * 
 * @author Christophe Hamerling - EBM WebSourcing
 */
public class MessageInterceptorManager {

    /**
     * The input interceptors to be used. The Map is ordered, the iterator used
     * to iterate over the entries will return the {@link Interceptor} in the
     * user defined order.
     */
    private Map<String, Interceptor> inputOrderedInterceptors;

    /**
     * The input accept interceptors to be used. The Map is ordered, the
     * iterator used to iterate over the entries will return the {@link
     * Interceptor} in the user defined order.
     */
    private Map<String, Interceptor> inputResponseOrderedInterceptors;

    /**
     * The output interceptors to be used. The Map is ordered, the iterator used
     * to iterate over the entries will return the {@link Interceptor} in the
     * user defined order.
     */
    private Map<String, Interceptor> outputOrderedInterceptors;

    /**
     * The output response interceptors to be used. The Map is ordered, the
     * iterator used to iterate over the entries will return the {@link
     * Interceptor} in the user defined order.
     */
    private Map<String, Interceptor> outputResponseOrderedInterceptors;

    /**
     * Additional configuration at SU level for input interceptors
     */
    private Map<String, ConfigurationExtensions> inputAdditionalConfiguration;

    /**
     * Additional configuration at SU level for input response interceptors
     */
    private Map<String, ConfigurationExtensions> inputResponseAdditionalConfiguration;

    /**
     * Additional configuration at SU level for output interceptors
     */
    private Map<String, ConfigurationExtensions> outputAdditionalConfiguration;

    /**
     * Additional configuration at SU level for output response interceptors
     */
    private Map<String, ConfigurationExtensions> outputResponseAdditionalConfiguration;

    /**
     * Logger
     */
    private Logger logger;

    /**
     * The component properties File path
     */
    private String propertiesFilePath;

    /**
     * Creates a new instance of {@link MessageInterceptorManager}
     * 
     * @param logger
     */
    public MessageInterceptorManager(final Logger logger) {
        this.logger = logger;
        this.inputOrderedInterceptors = new LinkedHashMap<String, Interceptor>();
        this.inputResponseOrderedInterceptors = new LinkedHashMap<String, Interceptor>();
        this.outputOrderedInterceptors = new LinkedHashMap<String, Interceptor>();
        this.outputResponseOrderedInterceptors = new LinkedHashMap<String, Interceptor>();

        this.inputAdditionalConfiguration = new HashMap<String, ConfigurationExtensions>();
        this.inputResponseAdditionalConfiguration = new HashMap<String, ConfigurationExtensions>();
        this.outputAdditionalConfiguration = new HashMap<String, ConfigurationExtensions>();
        this.outputResponseAdditionalConfiguration = new HashMap<String, ConfigurationExtensions>();
    }

    /**
     * Creates a new instance of {@link MessageInterceptorManager}
     * 
     * @param logger
     */
    public MessageInterceptorManager(String propertiesFilePath, final Logger logger) {
        this(logger);
        this.propertiesFilePath = propertiesFilePath;
    }

    /**
     * Initialization
     * 
     * @param componentInterceptors
     * @param localConfiguration
     * @param inOrder
     * @param outOrder
     */
    public void init(final Map<String, Interceptor> componentInterceptors,
            final List<SUInterceptor> inConfiguration, final List<SUInterceptor> outConfiguration,
            final List<SUInterceptor> inResponseConfiguration,
            final List<SUInterceptor> outResponseConfiguration) {
        logger.fine("Initializing interceptor manager");

        // accept
        this.initInterceptor(componentInterceptors, inConfiguration, inputOrderedInterceptors,
                inputAdditionalConfiguration);
        this.initInterceptor(componentInterceptors, inResponseConfiguration,
                inputResponseOrderedInterceptors, inputResponseAdditionalConfiguration);

        // send
        this.initInterceptor(componentInterceptors, outConfiguration, outputOrderedInterceptors,
                outputAdditionalConfiguration);
        this.initInterceptor(componentInterceptors, outResponseConfiguration,
                outputResponseOrderedInterceptors, outputResponseAdditionalConfiguration);

    }

    /**
     * Initialize the interceptor manager. Put all the interceptors in the good
     * order.
     * 
     * @param componentInterceptor
     *      the JBI component level interceptors. Must be non null.
     * @param localInterceptor
     *      the interceptors to be setted
     * @param orderedInterceptorsConfiguration
     *      the interceptors configuration
     */
    protected void initInterceptor(final Map<String, Interceptor> componentInterceptors,
            final List<SUInterceptor> originalInterceptorsConfiguration,
            Map<String, Interceptor> orderedInterceptors,
            Map<String, ConfigurationExtensions> additionalInterceptorsConfiguration) {

        logger.fine("Initializing interceptors...");

        // add the component active interceptors in the component defined
        // order. The component interceptors have higher priority
        for (final Interceptor interceptor : componentInterceptors.values()) {
            if (interceptor.isActive()) {
                orderedInterceptors.put(interceptor.getName(), interceptor);
            }
        }

        if (originalInterceptorsConfiguration != null) {
            // set additional configuration
            for (SUInterceptor originalSUInterceptor : originalInterceptorsConfiguration) {
                String itcName = originalSUInterceptor.getName();
                ConfigurationExtensions extensions;
                try {
                    if (this.propertiesFilePath != null) {
                        extensions = new ConfigurationExtensions(originalSUInterceptor.getParam(),
                                this.propertiesFilePath);
                    } else {
                        extensions = new ConfigurationExtensions(originalSUInterceptor.getParam());
                    }
                    additionalInterceptorsConfiguration.put(itcName, extensions);
                } catch (PEtALSCDKException e) {
                    logger.warning("Failed to set additional configuration for interceptor '"
                            + itcName + "'. Skip it");
                }

                // get the component interceptor if it exists
                final Interceptor interceptor = componentInterceptors.get(itcName);
                if (interceptor != null && !orderedInterceptors.containsKey(itcName)) {
                    orderedInterceptors.put(interceptor.getName(), interceptor);
                } else {
                    logger.warning("Interceptor '" + itcName
                            + "' not found in component available interceptors. Skip it");
                }
            }
        }
    }

    /**
     * FIXME : Check state
     * 
     * @param interceptor
     */
    public void addGlobalInterceptor(Interceptor interceptor) {
        this.inputOrderedInterceptors.put(interceptor.getName(), interceptor);
        this.inputResponseOrderedInterceptors.put(interceptor.getName(), interceptor);
        this.outputOrderedInterceptors.put(interceptor.getName(), interceptor);
        this.outputResponseOrderedInterceptors.put(interceptor.getName(), interceptor);
    }

    /**
     * @return the inputOrderedInterceptors
     */
    public Map<String, Interceptor> getInputOrderedInterceptors() {
        return this.inputOrderedInterceptors;
    }

    /**
     * @return the outputOrderedInterceptors
     */
    public Map<String, Interceptor> getOutputOrderedInterceptors() {
        return this.outputOrderedInterceptors;
    }

    /**
     * @return the inputAdditionalConfiguration
     */
    public Map<String, ConfigurationExtensions> getInputAdditionalConfiguration() {
        return this.inputAdditionalConfiguration;
    }

    /**
     * @return the outputAdditionalConfiguration
     */
    public Map<String, ConfigurationExtensions> getOutputAdditionalConfiguration() {
        return this.outputAdditionalConfiguration;
    }

    /**
     * @return the inputResponseOrderedInterceptors
     */
    public Map<String, Interceptor> getInputResponseOrderedInterceptors() {
        return this.inputResponseOrderedInterceptors;
    }

    /**
     * @return the outputResponseOrderedInterceptors
     */
    public Map<String, Interceptor> getOutputResponseOrderedInterceptors() {
        return this.outputResponseOrderedInterceptors;
    }

    /**
     * @return the inputResponseAdditionalConfiguration
     */
    public Map<String, ConfigurationExtensions> getInputResponseAdditionalConfiguration() {
        return this.inputResponseAdditionalConfiguration;
    }

    /**
     * @return the outputResponseAdditionalConfiguration
     */
    public Map<String, ConfigurationExtensions> getOutputResponseAdditionalConfiguration() {
        return this.outputResponseAdditionalConfiguration;
    }
}
