/**
 * Copyright (c) 2007-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.listener;

import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;

import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.process.async.AsyncContext;

/**
 * This class is used when a JBI message is received on the Component from the JBI environment.
 * <p>
 * Service Engine case : <br>
 * The goal of this class is to process a service with the JBI message
 * information.<br>
 * For instance, a ServiceUnit has been deployed on the component, defining a
 * <code>provides</code> section.<br>
 * This section defines an Endpoint <em>"myService/jbiMyServiceEndpoint"</em>
 * and some property extensions.<br>
 * A <em>"myService/jbiMyServiceEndpoint"</em> has been activated on the JBI
 * environment.<br>
 * <br>
 * When a JBI consumer sends in the JBI environment a message to the
 * <em>"myService/jbiMyServiceEndpoint"</em> endpoint, this component received
 * the message, and calls the <code>onJBIMessage(exchange)</code> method.<br>
 * Then, the method has to process the message according to optional property
 * extensions (extensions of the provides that has activated the endpoint).
 * </p>
 * 
 * <p>
 * Binding Component case : <br>
 * The goal of this class is to call an external service, related to this JBI
 * message. The address associated with the requested endpoint is handled by the
 * Extensions object of the provides that has activated the endpoint.<br>
 * For instance, a ServiceUnit has been deployed on the component, defining a
 * <code>provides</code> section.<br>
 * This section defines a relation between the external address
 * <em>"http://www.enterprise.com/webservices/myService"</em> (an address
 * attribute of the provides extensions), and an Endpoint
 * <em>"myService/jbiMyServiceEndpoint"</em>.<br>
 * A <em>"myService/jbiMyServiceEndpoint"</em> has been activated on the JBI
 * environment.<br>
 * </p>
 * 
 * <p>
 * When a JBI consumer sends in the JBI environment a message to the
 * <em>"myService/jbiMyServiceEndpoint"</em> endpoint, this component received
 * the message, and calls the <code>onJBIMessage(exchange)</code> method.<br>
 * Then, the method has to send the message to
 * <em>"http://www.enterprise.com/webservices/myService"</em>.
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 */
public abstract class AbstractJBIListener extends AbstractListener {

    public AbstractJBIListener() {
        super();
    }

    /**
     * This method is invoked by the CDK when an asynchronous response is
     * received for the component. Its bound {@code AsyncContext} is set as
     * parameter.
     * <p />
     * This method MUST be overridden if asynchronous invocation are processed
     * by the component.
     * 
     * @param exchange
     * @param asyncContext
     */
    public boolean onAsyncJBIMessage(final Exchange asyncExchange, final AsyncContext asyncContext) {
        return false;
    }

    /**
     * This method is invoked by the CDK when an asynchronous context has
     * expires. Its bound {@code AsyncContext} is set as parameter.
     * <p />
     * This method MUST be overridden if asynchronous invocation are processed
     * by the component.
     * 
     * @param exchange
     * @param asyncContext
     */
    public boolean onExpiredAsyncJBIMessage(final Exchange originalExchange,
            final AsyncContext asyncContext) {
        // TODO: argument "originalExchange" is unneeded because included in "asyncContext"
        return false;
    }

    /**
     * <p>
     * This method is used to process a JBI message received on the Component
     * from the JBI environment.
     * </p>
     * 
     * <p>
     * This method can't throws Exceptions. You must managed Exception yourself
     * by using {@link Exchange} methods :<br/>
     * - setFault(Exception) : create a {@link Fault} with the given Exception
     * and set it in the {@link MessageExchange}<br/>
     * - setError(Exception) : set Error part of the {@link MessageExchange} and
     * set {@link MessageExchange} status to error. Exceptions handled by
     * {@link MessageExchange}s must be in the container classpath.
     * </p>
     * 
     * <p>
     * If the methods return {@code true}, according to the pattern use cases,
     * message can be terminated and sent back by the CDK :<br/>
     * - InOnly : DONE status is automatically set at the end of the message
     * exchange process.<br/>
     * - RobustInOnly : DONE status is automatically set if no Fault found in
     * exchange at the end of the message exchange process.<br/>
     * - InOut : DONE status is never set (forbidden).<br/>
     * - InOptOut : DONE status is automatically set if no Fault or Out message
     * found in exchange at the end of the message exchange process.
     * 
     * @param exchange
     *            the message exchange to process, wrapped to add some utility
     *            methods. Not null
     * 
     * @return {@code true} if the message must be terminated and sent back
     *         automatically by the CDK. The interceptors are activated.<br />
     *         {@code false} if you you want yourself terminate the process and
     *         not use the
     *         {@link org.ow2.petals.component.framework.process.MessageExchangeProcessor}
     * 
     * @throws RuntimeException
     *             RuntimeException will be handle by the MessageExchange as
     *             Error part. So handled {@link RuntimeException}s must be in
     *             the classpath of your Petals node(s)
     */
    public abstract boolean onJBIMessage(Exchange exchange);
    
}
