/**
 * Copyright (c) 2007-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.listener;

import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.transaction.NotSupportedException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.xml.namespace.QName;

import org.ow2.petals.commons.log.FlowAttributes;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.Message;
import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.ExchangeProperties;
import org.ow2.petals.component.framework.jbidescriptor.generated.MEPType;
import org.ow2.petals.component.framework.jbidescriptor.generated.MessageParam;
import org.ow2.petals.component.framework.jbidescriptor.generated.MessageProperties;
import org.ow2.petals.component.framework.jbidescriptor.generated.Param;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.logger.Utils;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.monitoring.Monitoring;
import org.ow2.petals.component.framework.monitoring.key.ServiceProviderKey;
import org.ow2.petals.component.framework.process.MessageInterceptorProcessor;
import org.ow2.petals.component.framework.process.async.AsyncContext;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.probes.api.key.ProbeKey;
import org.ow2.petals.probes.api.probes.KeyedCounterProbeWithExecutionStatus;
import org.ow2.petals.probes.api.probes.KeyedResponseTimeProbe;
import org.ow2.petals.probes.api.probes.KeyedStartDateItem;

/**
 * The abstract methods for a CDK listener. 
 * <p>
 * Both external and JBI listener extend it.
 * </p>
 * 
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Olivier Fabre - EBM WebSourcing
 * @author Adrien Louis - EBM WebSourcing
 */
public abstract class AbstractListener {

    /**
     * The JBI component itself
     */
    protected AbstractComponent component;

    /**
     * The messageExchange factory used to create MessageExchanges
     */
    protected MessageExchangeFactory factory;

    /**
     * The consumes for External Listeners
     */
    private Consumes consumes;

    /**
     * The provides for JBI listeners
     */
    private Provides provides;

    /**
     * The Consumes or Provides extensions
     */
    private ConfigurationExtensions extensions;

    /**
     * The current Transaction
     */
    private Transaction currentTransaction;

    /**
     * The message interceptor processor; Each listener have its own interceptor
     * processor in order to get better performances.
     */
    protected MessageInterceptorProcessor messageInterceptorProcessor;

    /**
     * The probe about response times of service providers
     */
    private KeyedResponseTimeProbe<ServiceProviderKey> probeServiceProviderResponseTimes;

    /**
     * The pending start date items of the probe about response times of service
     * providers
     */
    private ConcurrentHashMap<String, KeyedStartDateItem<ServiceProviderKey>> serviceProviderStartDateItems;

    /**
     * The probe counting invokations of service providers
     */
    private KeyedCounterProbeWithExecutionStatus<ProbeKey> probeServiceProviderInvokations;

    /**
     * The pending invokation keys of the probe counting service provider
     * invokations
     */
    private ConcurrentHashMap<String, ServiceProviderKey> serviceProviderInvokationKeys;

    /**
     * Creates a new instance of {@link AbstractListener}
     * 
     */
    public AbstractListener() {
        super();
    }

    /**
     * Create a message exchange with all the information set in the specified
     * CONSUME <br>
     * - address (interface, service, endpoint) <br>
     * - MEP <br>
     * - operation <br>
     * - PEtALS exchange' meta-data (org.objectweb.petals.xxx)
     * 
     * @return a message exchange set with Consumes attributes
     * @throws MessagingException
     *             The message exchange can't be created or no pattern specified
     *             in extensions
     * @throws PEtALSCDKException
     *             The current Transaction has failed to return its status
     */
    public Exchange createConsumeExchange(final Consumes consumes) throws MessagingException,
            PEtALSCDKException {
        final MEPConstants mep = this.retrievePattern(consumes);
        final Exchange exchange = this.createConsumeExchange(consumes, mep);
        return exchange;
    }

    /**
     * Create a message exchange with all the information set in the specified
     * CONSUME, with the given exchange pattern <br>
     * - address (interface, service, endpoint) <br>
     * - operation <br>
     * - PEtALS exchange' meta-data (org.objectweb.petals.xxx)
     * 
     * @return a message exchange set with Consumes attributes
     * @throws MessagingException
     *             The message exchange can't be created, or no pattern
     *             specified in extensions, or endpoint not registered.
     * @throws PEtALSCDKException
     *             The current Transaction has failed to return its status
     */
    public Exchange createConsumeExchange(final Consumes consumes, final MEPConstants mep)
            throws MessagingException, PEtALSCDKException {
        final Exchange exchange = this.createExchange(mep);
        this.setOperation(exchange, consumes);
        this.setProperties(exchange, consumes);
        this.setEndpoint(exchange, consumes);
        return exchange;
    }

    /**
     * Create an exchange implementation matching the given pattern.
     * 
     * @param mep
     *            a specific MEP pattern
     * @return an exchange implementation matching the given pattern
     * @throws MessagingException
     *             The message exchange can't be created or no pattern specified
     *             in extensions
     * @throws PEtALSCDKException
     *             The current Transaction has failed to return its status
     */
    public Exchange createExchange(final MEPConstants mep) throws MessagingException,
            PEtALSCDKException {
        MessageExchange exchange = null;
        if (mep == null) {
            throw new PEtALSCDKException("MEP cannot be null when creating an exchange");
        } else {
            exchange = this.factory.createExchange(mep.value());
        }
        // if there is a current transaction,
        try {
            if (this.currentTransaction != null
                    && this.currentTransaction.getStatus() == Status.STATUS_ACTIVE) {
                exchange.setProperty(MessageExchange.JTA_TRANSACTION_PROPERTY_NAME,
                        this.currentTransaction);
            }
        } catch (final SystemException e) {
            throw new PEtALSCDKException("Failed to get current transaction", e);
        }

        return new ExchangeImpl(exchange);
    }

    /**
     * Get the component.
     * 
     * @return
     */
    public final AbstractComponent getComponent() {
        return this.component;
    }

    /**
     * The direct usage of the Component context is not recommended in the CDK.
     * 
     * @return
     */
    @Deprecated
    public ComponentContext getComponentContext() {
        return this.component.getContext();
    }

    public Consumes getConsumes() {
        return this.consumes;
    }

    /**
     * Return the current Transaction.
     * 
     * @return The current Transaction or {@code null} if no Transaction Manager
     *         is implemented by the JBI container
     * 
     * @throws SystemException
     * @throws NotSupportedException
     */
    public Transaction getCurrentTransaction() throws PEtALSCDKException {
        if (this.currentTransaction == null) {
            final TransactionManager transactionManager = (TransactionManager) this.component
                    .getContext().getTransactionManager();
            if (transactionManager != null) {
                try {
                    transactionManager.begin();
                    this.currentTransaction = transactionManager.getTransaction();
                } catch (final NotSupportedException e) {
                    throw new PEtALSCDKException("Failed to create a Transaction", e);
                } catch (final SystemException e) {
                    throw new PEtALSCDKException("Failed to create a Transaction", e);
                }

            }
        }
        return this.currentTransaction;
    }

    /**
     * Get the service extensions.
     * 
     * @return Extensions, can be null if no extensions found
     */
    public ConfigurationExtensions getExtensions() {
        return this.extensions;
    }

    public String getInstallRoot() {
        final AbstractServiceUnitManager suManager = (AbstractServiceUnitManager) this
                .getComponent().getServiceUnitManager();
        String installRoot = null;
        if (this.consumes != null) {
            installRoot = suManager.getSUDataHandlerForConsumes(this.consumes).getInstallRoot();
        } else {
            installRoot = suManager.getSUDataHandlerForProvides(this.provides).getInstallRoot();
        }
        return installRoot;
    }

    /**
     * 
     * @return
     */
    public final Logger getLogger() {
        return this.component.getLogger();
    }

    /**
     * @return the messageInterceptorProcessor
     */
    public MessageInterceptorProcessor getMessageInterceptorProcessor() {
        return this.messageInterceptorProcessor;
    }

    public Provides getProvides() {
        return this.provides;
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.init(). This method is called after setting of all Listener
     * attributes
     * 
     */
    public void init() {
    }

    /**
     * Initialize the listener.
     * 
     * @param component
     */
    public final void init(final AbstractComponent component) {
        this.component = component;
        this.messageInterceptorProcessor = new MessageInterceptorProcessor(this.component,
                this.getLogger());
        this.factory = this.getComponent().getChannel().createExchangeFactory();

        final Monitoring monitoringMBean = component.getMonitoringBean();
        this.probeServiceProviderResponseTimes = monitoringMBean
                .getServiceProviderResponseTimesProbe();
        this.serviceProviderStartDateItems = monitoringMBean.getServiceProviderStartDateItems();
        this.probeServiceProviderInvokations = monitoringMBean.getServiceProviderInvokationsProbe();
        this.serviceProviderInvokationKeys = monitoringMBean.getServiceProviderInvokationKeys();

    }

    /**
     * Send a message to the endpoint specified in the MessageExchange in an
     * asynchronous way.<br/>
     * Process MessageExchange interceptors before sending message. <br/>
     * 
     * @param exchange
     *            message exchange to send; must be non-null
     * @throws MessagingException
     *             if the address has no endpoint associated to it, or if an
     *             error occurs while sending the message through the container
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public void send(Exchange exchange) throws MessagingException {
        this.component.getLogger().fine(
                "Send an asynchronous exchange with id : " + exchange.getExchangeId());
        boolean isProviderRole = exchange.isProviderRole();
        
        if (exchange.isOutMessage() || exchange.getFault() != null) {
            try {
                this.interceptAcceptResponse(exchange);
            } catch (final PEtALSCDKException e) {
                throw new MessagingException(e);
            }
        } else {
            try {
                this.interceptSend(exchange);
            } catch (final PEtALSCDKException e) {
                throw new MessagingException(e);
            }
        }

        // TODO pattern InOptionalOut, it is necessary to test that the incoming
        // message was not a fault (not if it is not a fault after processing
        // the message)
        if (isProviderRole) {

            // test if the flow instance id was initialized
            MessageExchange messageExchange = ((ExchangeImpl) exchange).getMessageExchange();
            FlowAttributes flowAttributes = messageExchange.getFlowAttributes();

            if (flowAttributes != null) {
                // necessary to put the flow instance id in the context for the
                // contextual file handler
                PetalsExecutionContext.putFlowAttributes(flowAttributes);
            }

            Utils.addMonitEndOrFailureTrace(this.component.getLogger(), exchange,
                    flowAttributes);

            Utils.updateStatsOnEndOfServiceProviderInvokation(exchange,
                    this.probeServiceProviderResponseTimes, this.serviceProviderStartDateItems,
                    this.probeServiceProviderInvokations, this.serviceProviderInvokationKeys,
                    this.getLogger());
        }

        this.getComponent().getChannel().send(((ExchangeImpl) exchange).getMessageExchange());
    }

    /**
     * Send a message to the endpoint specified in the MessageExchange in an
     * asynchronous way.<br/>
     * 
     * @param exchange
     *            message exchange to send; must be non-null
     * @throws MessagingException
     *             if the address has no endpoint associated to it, or if an
     *             error occurs while sending the message through the container
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public void sendAsync(final Exchange exchange, final AsyncContext asyncContext)
            throws MessagingException {
        this.component.getAsyncManager().referenceAsyncContext(exchange, asyncContext);
        try {
            this.send(exchange);
        } catch (final MessagingException e) {
            this.component.getAsyncManager().unreferenceAsyncContext(exchange.getExchangeId());
            throw e;
        }
    }

    /**
     * Call sendSync(MessageExchage exchange, long timeoutMS) with a timeout
     * value found in provides or consumes extensions or "0" if not found
     * 
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public boolean sendSync(final Exchange exchange) throws MessagingException {
        long timeoutMS = Constants.Component.DEFAULT_SEND_TIMEOUT;

        // set the timeout if it is configured
        if (this.consumes != null && this.consumes.getTimeout() != null) {
            timeoutMS = Long.valueOf(this.consumes.getTimeout());
        } else if (this.provides != null && this.provides.getTimeout() != null) {
            timeoutMS = Long.valueOf(this.provides.getTimeout());
        }

        return this.sendSync(exchange, timeoutMS);
    }

    /**
     * Send a message to the endpoint specified in the MessageExchange in an
     * synchronous way.<br/>
     * Process MessageExchange interceptors before sending message. <br/>
     * 
     * @param exchange
     *            message exchange to send; must be non-null
     * @param timeoutMS
     *            timeout period, in milliseconds (0 means no timeout); must be
     *            greater than or equal to zero
     * @return true if the message has been sent (if a timeout occurs, return
     *         false)
     * @throws MessagingException
     *             if the address has no endpoint associated to it, or if an
     *             error occurs while sending the message through the container
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public boolean sendSync(final Exchange exchange, final long timeoutMS)
            throws MessagingException {
        boolean bRet = false;

        try {
            this.interceptSend(exchange);
            this.component.getLogger().finest(
                    "Send a synchronous exchange with id : " + exchange.getExchangeId());
            bRet = this.getComponent().getChannel()
                    .sendSync(((ExchangeImpl) exchange).getMessageExchange(), timeoutMS);
            this.interceptSendResponse(exchange);
        } catch (final PEtALSCDKException e) {
            throw new MessagingException(e);
        } catch (final JBIException e) {
            throw new MessagingException(e);
        }

        return bRet;
    }

    /**
     * @param consumes
     *            the consumes to set
     */
    public final void setConsumes(final Consumes consumes) {
        this.consumes = consumes;
    }

    /**
     * 
     * @param transaction
     */
    public final void setCurrentTransaction(final Transaction transaction) {
        this.currentTransaction = transaction;
    }

    /**
     * @param consumes
     *            the consumes to set
     */
    public final void setExtensions(final ConfigurationExtensions extensions) {
        this.extensions = extensions;
    }

    /**
     * Set Provides section
     * 
     * @param provides
     *            the provides to set
     */
    public final void setProvides(final Provides provides) {
        this.provides = provides;
    }

    /**
     * Retrieve the MEP from the consumer node extensions.
     * 
     * @param consumes
     *            The consumer node of the SU JBI descriptor.
     * @return MEP specified in extensions, or <code>null</code> if no MEP
     *         pattern specified.
     * @throws MessagingException
     *             MEP pattern specified in consumes extension is not valid.
     */
    protected final MEPConstants retrievePattern(final Consumes consumes) throws MessagingException {
        MEPConstants mep = null;

        if (consumes.getMep() != null) {
            if (MEPType.IN_ONLY.equals(consumes.getMep())) {
                mep = MEPConstants.IN_ONLY_PATTERN;
            } else if (MEPType.ROBUST_IN_ONLY.equals(consumes.getMep())) {
                mep = MEPConstants.ROBUST_IN_ONLY_PATTERN;
            } else if (MEPType.IN_OUT.equals(consumes.getMep())) {
                mep = MEPConstants.IN_OUT_PATTERN;
            } else if (MEPType.IN_OPTIONAL_OUT.equals(consumes.getMep())) {
                mep = MEPConstants.IN_OPTIONAL_OUT_PATTERN;
            } else {
                throw new MessagingException("MEP pattern '" + consumes.getMep().value()
                        + "' is not valid");
            }
        }

        return mep;
    }

    /**
     * Call the Interceptors before sending the request exchange
     * 
     * @param exchange
     * @throws PEtALSCDKException
     */
    private void interceptSend(final Exchange exchange) throws PEtALSCDKException {
        if (this.consumes != null) {
            this.messageInterceptorProcessor.intercept(exchange, this.consumes,
                    Message.InterceptorPhase.SEND);
        }
    }

    /**
     * Call the Interceptors after a synchronous send.
     * 
     * @param exchange
     * @throws PEtALSCDKException
     */
    private void interceptSendResponse(final Exchange exchange) throws PEtALSCDKException {
        if (this.consumes != null) {
            this.messageInterceptorProcessor.intercept(exchange, this.consumes,
                    Message.InterceptorPhase.SEND_RESPONSE);
        }
    }

    /**
     * Call the Intercectors before an async message response
     * 
     * @param exchange
     * @throws PEtALSCDKException
     */
    private void interceptAcceptResponse(final Exchange exchange) throws PEtALSCDKException {
        final Provides provides = ((AbstractServiceUnitManager) this.getComponent()
                .getServiceUnitManager()).getProvidesFromEndpoint(exchange.getEndpoint());
        if (provides != null) {
            this.messageInterceptorProcessor.intercept(exchange, provides,
                    Message.InterceptorPhase.ACCEPT_RESPONSE);
        }
    }

    /**
     * Fill the exchange with the service information (itf, src and ep) from the
     * consume extension
     * 
     * @param exchange
     * @param consumes
     * 
     * @throws MessagingException
     *             The endpoint is not registered as an endpoint of the
     *             specified service.
     */
    private void setEndpoint(final Exchange exchange, final Consumes consumes)
            throws MessagingException {

        // Retrieve possible target endpoint
        final String endpointName = consumes.getEndpointName();
        final QName serviceName = consumes.getServiceName();
        ServiceEndpoint ep;
        if (endpointName != null && serviceName != null) {
            ep = this.getComponent().getContext().getEndpoint(serviceName, endpointName);
            if (ep == null) {
                // The asked endpoint is not registered
                throw new MessagingException("The endpoint '" + endpointName
                        + "' is not registered with the service name '" + serviceName.toString()
                        + "'.");
            }
        } else {
            ep = null;
        }

        // set exchange information
        exchange.setInterfaceName(consumes.getInterfaceName());
        exchange.setService(consumes.getServiceName());
        exchange.setEndpoint(ep);

        if (consumes.getLinkType() != null) {
            exchange.setProperty(Exchange.PROPERTY_PROVIDER_LINKTYPE, consumes.getLinkType()
                    .value());
        }
    }

    /**
     * Set operation to call if exists in extensions. If no extension can be
     * found or if operation value is "" or missing, operation isn't set.
     * 
     * @param exchange
     *            a {@link MessageExchange}, cannot be null
     * @param extensions
     */
    private final void setOperation(final Exchange exchange, final Consumes consumes) {

        if (consumes.getOperation() != null) {
            exchange.setOperation(consumes.getOperation());
        }
    }

    /**
     * Set exchange and message properties from extensions.
     * 
     * @param exchange
     *            a {@link Exchange}
     * @param extensions
     * @throws MessagingException
     */
    private final void setProperties(final Exchange exchange, final Consumes consumes)
            throws MessagingException {
        final ExchangeProperties exchangeProperties = consumes.getExchangeProperties();
        if (exchangeProperties != null) {
            for (final Param exchangeProperty : exchangeProperties.getExchangeProperty()) {
                exchange.setProperty(exchangeProperty.getName(), exchangeProperty.getValue());
            }
        }

        final MessageProperties messageProperties = consumes.getMessageProperties();
        if (messageProperties != null) {
            for (final MessageParam messageProperty : messageProperties.getMessageProperty()) {
                final NormalizedMessage message = ((ExchangeImpl) exchange).getMessageExchange()
                        .getMessage(messageProperty.getMessage().name());
                message.setProperty(messageProperty.getName(), messageProperty.getValue());
            }
        }
    }
}
