/**
 * Copyright (c) 2011-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.logger;

import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Logger;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange.Role;
import javax.jbi.messaging.MessagingException;

import org.ow2.petals.commons.log.FlowAttributes;
import org.ow2.petals.commons.log.Level;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.monitoring.key.ServiceProviderKey;
import org.ow2.petals.probes.api.enums.ExecutionStatus;
import org.ow2.petals.probes.api.exceptions.ProbeKeyMissingException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ResponseTimeCollectionStoppedException;
import org.ow2.petals.probes.api.exceptions.StartDateItemUnknownException;
import org.ow2.petals.probes.api.key.ProbeKey;
import org.ow2.petals.probes.api.probes.KeyedCounterProbeWithExecutionStatus;
import org.ow2.petals.probes.api.probes.KeyedResponseTimeProbe;
import org.ow2.petals.probes.api.probes.KeyedStartDateItem;

public class Utils {

    private static final String ERROR_MESSAGE_LINE_SEPARATOR = " - ";

    private static final String BUSINESS_ERROR_MESSAGE = "A business error occurs";

    private static final String UNKNOWN_ERROR_MESSAGE = "A unknown error occurs (%s)";

    private Utils() {
        // NOP
    }

    public static final void addMonitEndOrFailureTrace(Logger logger, Exchange exchange,
            FlowAttributes flowAttributes) {
        String flowInstanceId = null;
        String flowStepId = null;
        if (flowAttributes != null) {
            flowInstanceId = flowAttributes.getFlowInstanceId();
            flowStepId = flowAttributes.getFlowStepId();
        }

        try {
            AbstractFlowLogData flowLogData;
            Fault fault = exchange.getFault();
            if (!exchange.isErrorStatus() && fault == null) {
                if (exchange.isConsumerRole()) {
                    flowLogData = new ConsumeFlowStepEndLogData(flowInstanceId, flowStepId);
                } else {
                    flowLogData = new ProvideFlowStepEndLogData(flowInstanceId, flowStepId,
                            ((ExchangeImpl) exchange).getMessageExchange());
                }
            } else {
                final String failureMessage;
                if (exchange.isErrorStatus()) {
                    Exception error = exchange.getError();
                    assert error != null;
                    String message = error.getMessage();
                    if (message != null) {
                        failureMessage = message.replace("\n", ERROR_MESSAGE_LINE_SEPARATOR);
                    } else {
                        failureMessage = String.format(UNKNOWN_ERROR_MESSAGE,
                                error.getStackTrace()[0].toString());
                    }
                } else {
                    failureMessage = BUSINESS_ERROR_MESSAGE;
                }

                if (exchange.isConsumerRole()) {
                    flowLogData = new ConsumeFlowStepFailureLogData(flowInstanceId, flowStepId,
                            failureMessage);
                } else {
                    flowLogData = new ProvideFlowStepFailureLogData(flowInstanceId, flowStepId,
                            failureMessage, ((ExchangeImpl) exchange).getMessageExchange());
                }
            }

            logger.log(Level.MONIT, "", flowLogData);
        } catch (final MessagingException e) {
            logger.log(Level.WARNING, "Error logging business trace.", e);
        }
    }

    public static final void addMonitFailureTrace(Logger logger, FlowAttributes flowAttributes,
            String failureMessage, Role role) {
        failureMessage = failureMessage.replace("\n", ERROR_MESSAGE_LINE_SEPARATOR);

        String flowInstanceId = null;
        String flowStepId = null;
        if (flowAttributes != null) {
            flowInstanceId = flowAttributes.getFlowInstanceId();
            flowStepId = flowAttributes.getFlowStepId();
        }

        AbstractFlowLogData flowLogData;
        if (role == Role.CONSUMER) {
            flowLogData = new ConsumeFlowStepFailureLogData(flowInstanceId, flowStepId,
                    failureMessage);
        } else {
            flowLogData = new ProvideFlowStepFailureLogData(flowInstanceId, flowStepId,
                    failureMessage, null);
        }

        logger.log(Level.MONIT, "", flowLogData);
    }

    /**
     * Update monitoring probes when the service provider invokation ends.
     * 
     * @param exchange
     *            The message exchange associated to the service provider
     *            invokation.
     * @param probeServiceProviderResponseTimes
     *            The probe about response times of service providers
     * @param serviceProviderStartDateItems
     *            The pending start date items of the probe about response times
     *            of service providers
     * @param probeServiceProviderInvokations
     *            The probe counting invokations of service providers
     * @param serviceProviderInvokationKeys
     *            The pending invokation keys of the probe counting service
     *            provider invokations
     */
    public static final void updateStatsOnEndOfServiceProviderInvokation(final Exchange exchange,
            final KeyedResponseTimeProbe<ServiceProviderKey> probeServiceProviderResponseTimes,
            final ConcurrentHashMap<String, KeyedStartDateItem<ServiceProviderKey>> serviceProviderStartDateItems,
            final KeyedCounterProbeWithExecutionStatus<ProbeKey> probeServiceProviderInvokations,
            final ConcurrentHashMap<String, ServiceProviderKey> serviceProviderInvokationKeys,
            final Logger logger) {

        final ExecutionStatus executionStatus;
        if (exchange.getStatus() == ExchangeStatus.ERROR) {
            logger.fine("Ends a start date item in ERROR");
            executionStatus = ExecutionStatus.ERROR;
        } else if (exchange.getFault() != null) {
            executionStatus = ExecutionStatus.FAULT;
            logger.fine("Ends a start date item in FAULT");
        } else {
            executionStatus = ExecutionStatus.SUCCEEDED;
            logger.fine("Ends a start date item in SUCCESS");
        }

        updateStatsOnEndOfServiceProviderInvokation(exchange, executionStatus,
                probeServiceProviderResponseTimes, serviceProviderStartDateItems,
                probeServiceProviderInvokations, serviceProviderInvokationKeys, logger);

    }

    /**
     * Update monitoring probes when the service provider invokation ends, using
     * the execution status from a parameter value.
     * 
     * @param exchange
     *            The message exchange associated to the service provider
     *            invokation.
     * @param executionStatus
     *            The execution status
     * @param probeServiceProviderResponseTimes
     *            The probe about response times of service providers
     * @param serviceProviderStartDateItems
     *            The pending start date items of the probe about response times
     *            of service providers
     * @param probeServiceProviderInvokations
     *            The probe counting invokations of service providers
     * @param serviceProviderInvokationKeys
     *            The pending invokation keys of the probe counting service
     *            provider invokations
     */
    public static final void updateStatsOnEndOfServiceProviderInvokation(
            final Exchange exchange,
            final ExecutionStatus executionStatus,
            final KeyedResponseTimeProbe<ServiceProviderKey> probeServiceProviderResponseTimes,
            final ConcurrentHashMap<String, KeyedStartDateItem<ServiceProviderKey>> serviceProviderStartDateItems,
            final KeyedCounterProbeWithExecutionStatus<ProbeKey> probeServiceProviderInvokations,
            final ConcurrentHashMap<String, ServiceProviderKey> serviceProviderInvokationKeys,
            final Logger logger) {

        final String exchangeId = exchange.getExchangeId();

        try {
            probeServiceProviderResponseTimes.endsExecution(
                    serviceProviderStartDateItems.remove(exchangeId), executionStatus);
        } catch (final StartDateItemUnknownException e2) {
            logger.warning("Start date item not found into the probe about response times of service providers. Metric values can be uncorrect.");
        } catch (final ProbeNotStartedException e2) {
            logger.warning("The probe about response times of service providers is not started. Metric values can be uncorrect.");
        } catch (final ResponseTimeCollectionStoppedException e2) {
            logger.warning("The sample collecting of the probe about response times of service providers is stopped. Metric values can be uncorrect.");
        }

        try {
            probeServiceProviderInvokations.move(serviceProviderInvokationKeys.remove(exchangeId),
                    executionStatus);
        } catch (final ProbeKeyMissingException e) {
            logger.warning("The probe counting service provider invokations does not retrieve the key. Metric values can be uncorrect.");
        } catch (final ProbeNotStartedException e) {
            logger.warning("The probe counting service provider invokations is not started. Metric values can be uncorrect.");
        }

    }
}
