/**
 * Copyright (c) 2007-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.mbean;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.DynamicMBean;
import javax.management.IntrospectionException;
import javax.management.InvalidAttributeValueException;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanException;
import javax.management.ReflectionException;

import org.ow2.petals.component.framework.jbidescriptor.generated.Component;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.jbidescriptor.generated.Runtimepositivestrictint;
import org.ow2.petals.component.framework.jbidescriptor.generated.Runtimepositivestrictlong;
import org.ow2.petals.component.framework.jbidescriptor.generated.Settablepositivestrictlong;
import org.ow2.petals.component.framework.process.JBIProcessorManager;

/**
 * Abstract dynamic MBean used as base for configuration MBean.
 * 
 * @author Christophe DENEUX - Linagora
 */
public abstract class AbstractConfigurationMBean implements DynamicMBean {

    /**
     * Name of the attribute {@link Component}.messageProcessorMaxPoolSize. It
     * is check when loading the current class.
     * 
     */
    public final static String MESSAGE_PROCESSOR_MAX_POOL_SIZE_ATTR_NAME = "messageProcessorMaxPoolSize";

    /**
     * Name of the attribute {@link Component}.processorMaxPoolSize. It is check
     * when loading the current class.
     * 
     */
    public final static String PROCESSOR_MAX_POOL_SIZE_ATTR_NAME = "processorMaxPoolSize";

    /**
     * Name of the attribute {@link Component}.processorPoolSize. It is check
     * when loading the current class.
     * 
     */
    public final static String PROCESSOR_POOL_SIZE_ATTR_NAME = "processorPoolSize";

    protected final Logger logger;

    protected final Jbi jbiComponentConfiguration;

    static {
        try {
            // We check that the attribute names are the right ones.
            Component.class.getDeclaredField(MESSAGE_PROCESSOR_MAX_POOL_SIZE_ATTR_NAME);
            Component.class.getDeclaredField(PROCESSOR_MAX_POOL_SIZE_ATTR_NAME);
            Component.class.getDeclaredField(PROCESSOR_POOL_SIZE_ATTR_NAME);
        } catch (final Exception e) {
            throw new RuntimeException(e);
        }
    }

    public AbstractConfigurationMBean(final Jbi jbiComponentConfiguration, final Logger logger) {

        this.jbiComponentConfiguration = jbiComponentConfiguration;
        this.logger = logger;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.DynamicMBean#getAttributes(java.lang.String[])
     */
    public synchronized AttributeList getAttributes(final String[] attributes) {
        final AttributeList list = new AttributeList();

        for (final String attribute : attributes) {
            try {
                list.add(new Attribute(attribute, this.getAttribute(attribute)));
            } catch (final AttributeNotFoundException e) {
                this.logger.warning("Failed to get attribute '" + attribute
                        + "' in the configuration MBean : " + e.getClass().getSimpleName() + " : "
                        + e.getMessage());
            } catch (final MBeanException e) {
                this.logger.warning("Failed to get attribute '" + attribute
                        + "' in the configuration MBean : " + e.getClass().getSimpleName() + " : "
                        + e.getMessage());
            } catch (final ReflectionException e) {
                this.logger.warning("Failed to get attribute '" + attribute
                        + "' in the configuration MBean : " + e.getClass().getSimpleName() + " : "
                        + e.getMessage());
            }
        }

        return list;
    }

    private Field findInheritedField(final Class<?> clazz, final String fieldName)
            throws NoSuchFieldException {

        Class<?> currentClazz = clazz;

        while (currentClazz != null) {
            try {
                return currentClazz.getDeclaredField(fieldName);
            } catch (final NoSuchFieldException e) {
                currentClazz = currentClazz.getSuperclass();
                if (currentClazz == null) {
                    // No more superclass
                    throw e;
                }
            }
        }
        // Statement unreachable
        return null;

    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.DynamicMBean#setAttribute(javax.management.Attribute)
     */
    public synchronized void setAttribute(final Attribute attribute)
            throws AttributeNotFoundException, InvalidAttributeValueException, MBeanException,
            ReflectionException {

        try {
            final Field field = Component.class.getDeclaredField(attribute.getName());
            if (field.getType().equals(Runtimepositivestrictint.class)) {
                final Object processorPoolSizeAttr = this
                        .getAttribute(PROCESSOR_POOL_SIZE_ATTR_NAME);
                final int processorPoolSizeAttrValue = ((Integer) processorPoolSizeAttr).intValue();
                final Object processorMaxPoolSizeAttr = this
                        .getAttribute(PROCESSOR_MAX_POOL_SIZE_ATTR_NAME);
                final int processorMaxPoolSizeAttrValue = processorMaxPoolSizeAttr == null ? JBIProcessorManager.DEFAULT_PROCESSOR_POOL_SIZE_MAX
                        : ((Integer) processorMaxPoolSizeAttr).intValue();
                final Object messageProcessorMaxPoolSizeAttr = this
                        .getAttribute(MESSAGE_PROCESSOR_MAX_POOL_SIZE_ATTR_NAME);
                final int messageProcessorMaxPoolSizeAttrValue = messageProcessorMaxPoolSizeAttr == null ? processorMaxPoolSizeAttrValue
                        : ((Integer) messageProcessorMaxPoolSizeAttr).intValue();

                if (field.getName().equals(PROCESSOR_MAX_POOL_SIZE_ATTR_NAME)
                        && ((Integer) attribute.getValue()).intValue() <= processorPoolSizeAttrValue) {
                    throw new InvalidAttributeValueException(
                            "Invalid value for attribute '"
                                    + attribute.getName()
                                    + "': The value must be upper or equals to the processor core pool size.");
                } else if (field.getName().equals(PROCESSOR_MAX_POOL_SIZE_ATTR_NAME)
                        && ((Integer) attribute.getValue()).intValue() > messageProcessorMaxPoolSizeAttrValue) {
                    throw new InvalidAttributeValueException(
                            "Invalid value for attribute '"
                                    + attribute.getName()
                                    + "': The value must be lower or equals to the message processor max pool size.");
                } else if (field.getName().equals(PROCESSOR_POOL_SIZE_ATTR_NAME)
                        && ((Integer) attribute.getValue()).intValue() > processorMaxPoolSizeAttrValue) {
                    throw new InvalidAttributeValueException(
                            "Invalid value for attribute '"
                                    + attribute.getName()
                                    + "': The value must be lower or equals to the processor max pool size ("
                                    + processorMaxPoolSizeAttrValue + ").");
                } else if (field.getName().equals(MESSAGE_PROCESSOR_MAX_POOL_SIZE_ATTR_NAME)
                        && ((Integer) attribute.getValue()).intValue() < processorMaxPoolSizeAttrValue) {
                    throw new InvalidAttributeValueException(
                            "Invalid value for attribute '"
                                    + attribute.getName()
                                    + "': The value must be upper or equals to the processor max pool size ("
                                    + processorMaxPoolSizeAttrValue + ").");
                } else if (((Integer) attribute.getValue()).intValue() <= 0) {
                        throw new InvalidAttributeValueException("Invalid value for attribute '"
                                + attribute.getName()
                                + "': The value must be upper or equals to 1.");
                }
            }
            else if (field.getType().equals(Runtimepositivestrictlong.class)) {
                if (((Long) attribute.getValue()).intValue() <= 0) {
                    throw new InvalidAttributeValueException("Invalid value for attribute '"
                            + attribute.getName()
                            + "': The value must be upper or equals to 1.");
                }
            }
 else if (field.getType().equals(Settablepositivestrictlong.class)) {
                if (((Long) attribute.getValue()).longValue() <= 0) {
                    throw new InvalidAttributeValueException("Invalid value for attribute '"
                            + attribute.getName() + "': The value must be upper or equals to 1.");
                }
            }
            field.setAccessible(true);
            final Object settableObject = field.get(this.jbiComponentConfiguration.getComponent());
            
            // get the previous value
            String methodName = "getValue";
            if (field.getType().getSimpleName().endsWith("boolean")) {
                methodName = "isValue";
            }
            final Method getMethod = settableObject.getClass().getMethod(methodName);
            final Object oldValue = getMethod.invoke(settableObject);
            
            // set the value
            final Field primitiveField = this.findInheritedField(field.getType(), "value");
            final Method setMethod = settableObject.getClass().getMethod("setValue",
                    new Class[] { primitiveField.getType() });
            setMethod.invoke(settableObject, new Object[] { attribute.getValue() });
            
            this.saveConfiguration();
            this.notifyConfigurationChange(attribute.getName() + "modified", attribute.getName(),
                    oldValue, attribute.getValue());
            
        } catch (final SecurityException e) {
            throw new AttributeNotFoundException("Attribute '" + attribute.getName()
                    + "' is not reachable");
        } catch (final NoSuchFieldException e) {
            // try to set a custom attribute
            this.setCustomAttribute(attribute);
        } catch (final IllegalArgumentException e) {
            throw new ReflectionException(e);
        } catch (final IllegalAccessException e) {
            throw new ReflectionException(e);
        } catch (final NoSuchMethodException e) {
            throw new ReflectionException(e);
        } catch (final InvocationTargetException e) {
            throw new ReflectionException(e);
        }
    }
    
    protected void setCustomAttribute(final Attribute attribute) throws ReflectionException,
            AttributeNotFoundException, InvalidAttributeValueException {
        // NOP
    }
    
    protected abstract Object getCustomAttribute(final String attributeName) throws ReflectionException, AttributeNotFoundException;

    protected void saveConfiguration() {
        // NOP
    }
    
    protected void notifyConfigurationChange(final String msg, final String attrName,
            final Object oldVal, final Object newVal) {
        // NOP
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.DynamicMBean#getAttribute(java.lang.String)
     */
    public synchronized Object getAttribute(final String attribute)
    throws AttributeNotFoundException, MBeanException, ReflectionException {
        Object attributeObject = null;

        try {
            final Field field = Component.class.getDeclaredField(attribute);
            field.setAccessible(true);
            final Object gettableObject = field.get(this.jbiComponentConfiguration.getComponent());
            if (gettableObject != null) {
                String methodName = "getValue";
                if (field.getType().getSimpleName().endsWith("boolean")) {
                    methodName = "isValue";
                }
                final Method method = gettableObject.getClass().getMethod(methodName);
                attributeObject = method.invoke(gettableObject);
            }

        } catch (final SecurityException e) {
            throw new AttributeNotFoundException("Attribute '" + attribute + "' is not reachable");
        } catch (final NoSuchFieldException e) {
            // try to get a custom attribute
            attributeObject = this.getCustomAttribute(attribute);
        } catch (final IllegalArgumentException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final IllegalAccessException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final NoSuchMethodException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final InvocationTargetException e) {
            throw new ReflectionException(e, e.getMessage());
        }

        return attributeObject;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.DynamicMBean#setAttributes(javax.management.AttributeList
     * )
     */
    public synchronized AttributeList setAttributes(final AttributeList list) {
        final AttributeList attributeList = new AttributeList();

        for (final Object object : list) {
            final Attribute attribute = (Attribute) object;
            try {
                this.setAttribute((Attribute) attribute);
            } catch (AttributeNotFoundException e) {
                this.logger.warning("Failed to set attribute '" + attribute.getName()
                        + "' in the configuration MBean : " + e.getClass().getSimpleName() + " : "
                        + e.getMessage());
            } catch (InvalidAttributeValueException e) {
                this.logger.warning("Failed to set attribute '" + attribute.getName()
                        + "' in the configuration MBean : " + e.getClass().getSimpleName() + " : "
                        + e.getMessage());
            } catch (MBeanException e) {
                this.logger.warning("Failed to set attribute '" + attribute.getName()
                        + "' in the configuration MBean : " + e.getClass().getSimpleName() + " : "
                        + e.getMessage());
            } catch (ReflectionException e) {
                this.logger.warning("Failed to set attribute '" + attribute.getName()
                        + "' in the configuration MBean : " + e.getClass().getSimpleName() + " : "
                        + e.getMessage());
            } catch (final Throwable t) {
                this.logger.warning("Failed to set attribute '" + ((Attribute) attribute).getName()
                        + "' in the configuration MBean : " + t.getClass().getSimpleName() + " : "
                        + t.getMessage());
            }
        }

        return attributeList;
    }

    /**
     * Dynamically get attributes for the MBean. The attributes are the ones
     * from the properties file.
     * 
     * @return
     */
    protected MBeanAttributeInfo[] getAttributesInfos() {
        MBeanAttributeInfo[] attrs = null;
        final Field[] fields = Component.class.getDeclaredFields();
        final List<Field> settableFields = new ArrayList<Field>();
        final Component componentConfiguration = this.jbiComponentConfiguration.getComponent();

        try {
            // get the settable fields
            for (final Field field : fields) {
                if (this.isSettableField(field)) {
                    field.setAccessible(true);
                    if (field.get(componentConfiguration) != null) {
                        settableFields.add(field);
                    }
                }
            }

            // build the MBeanAttributeInfo
            int count = 0;
            attrs = new MBeanAttributeInfo[settableFields.size()];

            for (final Field field : settableFields) {
                final Object settableObject = field.get(componentConfiguration);
                Method getter;
                if (field.getType().getSimpleName().endsWith("boolean")) {
                    getter = settableObject.getClass().getMethod("isValue");
                } else {
                    getter = settableObject.getClass().getMethod("getValue");
                }
                final Field primitiveField = this.findInheritedField(field.getType(), "value");
                final Method setter = settableObject.getClass().getMethod("setValue",
                        new Class[] { primitiveField.getType() });

                attrs[count] = new MBeanAttributeInfo(field.getName(), "Property "
                        + field.getName(), getter, setter);
                count++;

            }

        } catch (final IllegalArgumentException e) {
            this.logger.warning("Failed to initialize fields of the configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final IllegalAccessException e) {
            this.logger.warning("Failed to initialize fields of the configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final SecurityException e) {
            this.logger.warning("Failed to initialize fields of the configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final NoSuchMethodException e) {
            this.logger.warning("Failed to initialize fields of the configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final IntrospectionException e) {
            this.logger.warning("Failed to initialize fields of the configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final NoSuchFieldException e) {
            this.logger.warning("Failed to initialize fields of the configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        }

        return attrs;
    }

    protected abstract boolean isSettableField(final Field field);

    @Override
    public Object invoke(String actionName, Object[] params, String[] signature)
            throws MBeanException, ReflectionException {
        // TODO Auto-generated method stub
        return null;
    }
}
