/**
 * Copyright (c) 2007-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.mbean;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.List;

import javax.management.Attribute;
import javax.management.AttributeNotFoundException;
import javax.management.DynamicMBean;
import javax.management.IntrospectionException;
import javax.management.InvalidAttributeValueException;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanOperationInfo;
import javax.management.ReflectionException;
import javax.management.RuntimeOperationsException;

import org.ow2.petals.component.framework.DefaultBootstrap;

/**
 * The dynamic MBean used to expose component attributes and operations via JMX.
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Roland Naudin - EBM WebSourcing
 */
public class ExtensionMBean extends AbstractConfigurationMBean implements DynamicMBean {

    private static class AttributeSetterGetter {

        public String attributeName;

        public Method getter;

        public Method setter;

        public AttributeSetterGetter(final String attributeName, final Method getter,
                final Method setter) {
            this.attributeName = attributeName;
            this.getter = getter;
            this.setter = setter;
        }
    }

    private final List<AttributeSetterGetter> attributeSetterGetterList = new ArrayList<AttributeSetterGetter>();

    private final static Hashtable<String, Class<?>> primitiveClasses = new Hashtable<String, Class<?>>(
            8);

    private final DefaultBootstrap bootstrap;
    {
        primitiveClasses.put(Boolean.TYPE.toString(), Boolean.TYPE);
        primitiveClasses.put(Character.TYPE.toString(), Character.TYPE);
        primitiveClasses.put(Byte.TYPE.toString(), Byte.TYPE);
        primitiveClasses.put(Short.TYPE.toString(), Short.TYPE);
        primitiveClasses.put(Integer.TYPE.toString(), Integer.TYPE);
        primitiveClasses.put(Long.TYPE.toString(), Long.TYPE);
        primitiveClasses.put(Float.TYPE.toString(), Float.TYPE);
        primitiveClasses.put(Double.TYPE.toString(), Double.TYPE);
    }

    /**
     * Creates a new instance of {@link ExtensionMBean}
     * 
     * @param bootstrap
     */
    public ExtensionMBean(final DefaultBootstrap bootstrap) {
        super(bootstrap.getJbiComponentConfiguration(), bootstrap.getLogger());
        this.bootstrap = bootstrap;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.DynamicMBean#invoke(java.lang.String,
     * java.lang.Object[], java.lang.String[])
     */
    public Object invoke(final String actionName, final Object[] params, final String[] signature)
    throws MBeanException, ReflectionException {
        Object result = null;
        Class<?>[] parameters = null;

        final Class<?> objClass = this.bootstrap.getClass();
        final ClassLoader loader = objClass.getClassLoader();
        if (signature != null) {
            parameters = this.findSignatureClasses(signature, loader);
        }
        // Query the metadata service to get the right method
        //
        Method method = null;
        try {
            method = objClass.getMethod(actionName, parameters);
        } catch (final Throwable t) {
            // Nothing to do
        }

        if (method == null) {
            throw new ReflectionException(new NoSuchMethodException(actionName),
                    "The operation with name " + actionName + " could not be found");
        }

        try {
            // invoke method
            result = method.invoke(this.bootstrap, params);
        } catch (final IllegalAccessException e) {
            throw new ReflectionException(e, "IllegalAccessException"
                    + " occured trying to invoke operation " + actionName);
        } catch (final RuntimeException e) {
            throw new RuntimeOperationsException(e, "RuntimeException"
                    + " occured trying to invoke operation " + actionName);
        } catch (final InvocationTargetException e) {
            throw new MBeanException(e, "Exception thrown in operation " + actionName);
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.DynamicMBean#getMBeanInfo()
     */
    public synchronized MBeanInfo getMBeanInfo() {
        // get the attributes
        final MBeanAttributeInfo[] attrs = this.getAttributesInfos();
        // get the operations
        final MBeanOperationInfo[] opps = this.getOperationsInfos();

        // create MBeanInfo with no constructors and no notifications
        return new MBeanInfo(this.getClass().getName(), "Property Manager MBean", attrs, null,
                opps, null);
    }
    /**
     * Dynamically get the operations for the MBean. The operations are defined
     * in the properties manager with a special prefix.
     * 
     * @return
     */
    protected MBeanOperationInfo[] getOperationsInfos() {
        final List<String> methods = this.bootstrap.getMethodList();
        final List<MBeanOperationInfo> operations = new ArrayList<MBeanOperationInfo>();
        final Method[] methodList = this.bootstrap.getClass().getMethods();

        for (final Method method : methodList) {
            final String name = method.getName();

            // add method if it has been defined by the user
            if (methods.contains(name)) {
                final MBeanOperationInfo oper = new MBeanOperationInfo("Operation description",
                        method);
                operations.add(oper);
            }
        }

        final MBeanOperationInfo[] result = new MBeanOperationInfo[operations.size()];
        operations.toArray(result);

        return result;
    }

    /**
     * Find the classes of the signature.
     * 
     * @param signature
     * @param loader
     * @return
     * @throws ReflectionException
     */
    private Class<?>[] findSignatureClasses(final String signature[], final ClassLoader loader)
    throws ReflectionException {
        if (signature == null) {
            return null;
        }

        final int length = signature.length;
        final Class<?> tab[] = new Class[length];

        if (length == 0) {
            return tab;
        }

        try {
            for (int i = 0; i < length; i++) {
                // try to get the primitive class
                final Class<?> primCla = primitiveClasses.get(signature[i]);
                if (primCla != null) {
                    tab[i] = primCla;
                } else {
                    // try to get the class if not primitive one
                    tab[i] = Class.forName(signature[i], false, loader);
                }
            }
        } catch (final ClassNotFoundException e) {
            throw new ReflectionException(e, "The parameter class could not be found");
        } catch (final RuntimeException e) {
            throw e;
        }
        return tab;
    }

    @Override
    protected void setCustomAttribute(final Attribute attribute) throws ReflectionException,
            AttributeNotFoundException, InvalidAttributeValueException {

        try {
            for (final AttributeSetterGetter attributeSetterGetter : this.attributeSetterGetterList) {
                if (attributeSetterGetter.attributeName.equals(attribute.getName())) {
                    attributeSetterGetter.setter.invoke(this.bootstrap, attribute.getValue());
                    return;
                }
            }
        } catch (final IllegalArgumentException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final IllegalAccessException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final InvocationTargetException e) {
            if (e.getCause() instanceof InvalidAttributeValueException) {
                throw (InvalidAttributeValueException) e.getCause();
            } else {
                throw new ReflectionException(e, e.getMessage());
            }
        }

        // Custom attribute not found
        throw new AttributeNotFoundException("Attribute '" + attribute + "' does not exist");

    }

    @Override
    protected Object getCustomAttribute(final String attributeName) throws ReflectionException,
            AttributeNotFoundException {
        try {
            for (final AttributeSetterGetter attributeSetterGetter : this.attributeSetterGetterList) {
                if (attributeSetterGetter.attributeName.equals(attributeName)) {
                    return attributeSetterGetter.getter.invoke(this.bootstrap);
                }
            }
        } catch (final IllegalArgumentException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final IllegalAccessException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final InvocationTargetException e) {
            throw new ReflectionException(e, e.getMessage());
        }

        // Custom attribute not found
        throw new AttributeNotFoundException("Attribute '" + attributeName + "' does not exist");
    }

    @Override
    protected boolean isSettableField(final Field field) {
        return field.getType().getSimpleName().startsWith("Runtime")
                || field.getType().getSimpleName().startsWith("Settable");
    }

    @Override
    protected MBeanAttributeInfo[] getAttributesInfos() {
        final MBeanAttributeInfo[] inheritedAttrs = super.getAttributesInfos();

        // get the custom attributes list
        final List<String> customAttrs = this.bootstrap.getAttributeList();

        // Add custom MBeanAttributeInfo
        final MBeanAttributeInfo[] attrs = Arrays.copyOf(inheritedAttrs, inheritedAttrs.length
                + customAttrs.size());
        int count = inheritedAttrs.length;

        try {
            final Method[] methodList = this.bootstrap.getClass().getMethods();
            for (final String attr : customAttrs) {
                Method setter = null;
                Method getter = null;
                for (final Method method : methodList) {
                    if (method.getName().equalsIgnoreCase("set" + attr)) {
                        setter = method;
                    } else if (method.getName().equalsIgnoreCase("get" + attr)
                            || method.getName().equalsIgnoreCase("is" + attr)) {
                        getter = method;
                    }
                }
                attrs[count] = new MBeanAttributeInfo(attr, "Property " + attr, getter, setter);
                this.attributeSetterGetterList.add(new AttributeSetterGetter(attr, getter, setter));
                count++;
            }

        } catch (final IllegalArgumentException e) {
            this.logger.warning("Failed to initialize fields of the configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final SecurityException e) {
            this.logger.warning("Failed to initialize fields of the configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final IntrospectionException e) {
            this.logger.warning("Failed to initialize fields of the configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        }

        return attrs;
    }

}
