/**
 * Copyright (c) 2008-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.mbean;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.management.MBeanException;
import javax.management.ObjectName;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.Interceptor;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.interceptor.MessageInterceptorManager;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorException;
import org.ow2.petals.component.framework.jbidescriptor.JBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.ComponentInterceptor;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.component.framework.su.ServiceUnitDataHandler;

/**
 * Add and engage interceptors at runtime.
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Roland Naudin - EBM WebSourcing
 */
public class InterceptorManager implements InterceptorManagerMBean {

    private AbstractComponent component;

    private Logger logger;

    private ObjectName mBeanName;

    private Jbi jbiComponentConfiguration;

    private File jbiComponentFile;

    /**
     * 
     * @param component
     */
    public InterceptorManager(AbstractComponent component, Jbi jbiComponentConfiguration,
            File jbiComponentFile) {
        this.component = component;
        this.jbiComponentConfiguration = jbiComponentConfiguration;
        this.jbiComponentFile = jbiComponentFile;
        this.logger = component.getLogger();
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.mbean.InterceptorManagerMBean#
     * engageInterceptorToComponent(java.lang.String, java.lang.String)
     */
    public void addInterceptor(String className, String interceptorName) throws MBeanException {
        this.logger.log(Level.INFO, "Add interceptor " + interceptorName);

        if (this.component.getInterceptor(interceptorName) != null) {
            throw new MBeanException(new Exception("This interceptor is already registered"));
        }

        final Interceptor itc;
        try {
            Class<?> clazz = Thread.currentThread().getContextClassLoader().loadClass(className);
            itc = (Interceptor) clazz.newInstance();
            itc.init(this.component, interceptorName, false, new HashMap<String, String>(),
                    this.logger);
        } catch (PEtALSCDKException e) {
            throw new MBeanException(e);
        } catch (InstantiationException e) {
            throw new MBeanException(e);
        } catch (IllegalAccessException e) {
            throw new MBeanException(e);
        } catch (ClassNotFoundException e) {
            throw new MBeanException(e);
        } catch (ClassCastException e) {
            throw new MBeanException(e);
        }
        ComponentInterceptor componentInterceptor = new ComponentInterceptor();
        componentInterceptor.setActive(false);
        componentInterceptor.setClazz(className);
        componentInterceptor.setName(interceptorName);

        Map<String, Interceptor> interceptors = this.component.getInterceptors();
        interceptors.put(interceptorName, itc);
        List<ComponentInterceptor> componentInterceptors = this.component
                .getComponentConfiguration().getComponentInterceptors().getInterceptor();
        componentInterceptors.add(componentInterceptor);
        this.save();
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.mbean.InterceptorManagerMBean#
     * engageInterceptor(java.lang.String)
     */
    public void engageInterceptor(String interceptorName) throws MBeanException {
        this.logger.log(Level.INFO, "Engage interceptor " + interceptorName);

        Interceptor interceptor = this.component.getInterceptor(interceptorName);
        if (interceptor == null) {
            throw new MBeanException(new Exception("Failed to retrieve interceptor named '"
                    + interceptorName + "'"));
        }

        if (!interceptor.isActive()) {
            AbstractServiceUnitManager suManager = (AbstractServiceUnitManager) this.component
                    .getServiceUnitManager();
            for (ServiceUnitDataHandler suDataHandler : suManager.getServiceUnitDataHandlers()
                    .values()) {
                for (MessageInterceptorManager manager : suDataHandler
                        .getMessageInterceptorManagersMap().values()) {
                    manager.addGlobalInterceptor(interceptor);
                }
            }

            interceptor.setActive(true);
            List<ComponentInterceptor> componentInterceptors = this.component
                    .getComponentConfiguration().getComponentInterceptors().getInterceptor();
            for (ComponentInterceptor componentInterceptor : componentInterceptors) {
                if (componentInterceptor.getName().equals(interceptor.getName())) {
                    componentInterceptor.setActive(true);
                    this.save();
                    break;
                }
            }
        }
    }

    /**
     * 
     * @param mbeanName
     */
    public void setMBeanName(ObjectName mbeanName) {
        this.mBeanName = mbeanName;
    }

    /**
     * 
     * @return
     */
    public ObjectName getMBeanName() {
        return this.mBeanName;
    }

    private void save() {
        try {
            FileOutputStream outputStream = new FileOutputStream(jbiComponentFile);
            try {
                JBIDescriptorBuilder.buildXmlJBIdescriptor(this.jbiComponentConfiguration,
                        outputStream);
            } finally {
                outputStream.close();
            }
        } catch (CDKJBIDescriptorException e) {
            this.logger
                    .warning("Failed to persist modification of the Interceptor manager MBean : "
                            + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (IOException e) {
            this.logger
                    .warning("Failed to persist modification of the Interceptor manager MBean : "
                            + e.getClass().getSimpleName() + " : " + e.getMessage());
        }
    }

}
