/**
 * Copyright (c) 2008-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.mbean;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.lang.reflect.Field;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.management.Attribute;
import javax.management.AttributeChangeNotification;
import javax.management.AttributeNotFoundException;
import javax.management.DynamicMBean;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanInfo;
import javax.management.MBeanNotificationInfo;
import javax.management.Notification;
import javax.management.NotificationBroadcasterSupport;
import javax.management.NotificationEmitter;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.management.ReflectionException;

import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorException;
import org.ow2.petals.component.framework.jbidescriptor.JBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;

/**
 * A runtime configuration MBean which sends notifications to subscribers. 
 * <p>
 * It is used to change configuration values at runtime.
 * </p>
 * 
 * @author Christophe Hamerling - EBM WebSourcing
 */
public class RuntimeConfigurationNotifier extends AbstractConfigurationMBean implements
        DynamicMBean, NotificationEmitter {

    private final NotificationBroadcasterSupport broadcasterSupport;

    private ObjectName mbeanName;

    private final AtomicLong seqNumber;

    private final File jbiComponentFile;

    /**
     * 
     * @param workspaceRoot
     * @param properties
     * @throws JBIException
     */
    public RuntimeConfigurationNotifier(final Jbi jbiComponentConfiguration,
            final File jbiComponentFile, final Logger logger) {
        super(jbiComponentConfiguration, logger);
        this.jbiComponentFile = jbiComponentFile;
        this.broadcasterSupport = new NotificationBroadcasterSupport();
        this.seqNumber = new AtomicLong(0);
    }
    

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.DynamicMBean#getMBeanInfo()
     */
    public synchronized MBeanInfo getMBeanInfo() {
        // get the attributes
        final MBeanAttributeInfo[] attrs = this.getAttributesInfos();

        final MBeanNotificationInfo notif = new MBeanNotificationInfo(null,
                "Runtime MBean notifier", "Runtime configuration modification notifier");
        final MBeanNotificationInfo[] notifs = new MBeanNotificationInfo[] { notif };

        // create MBeanInfo with no constructors and no operations
        return new MBeanInfo(this.getClass().getName(), "Property Manager MBean", attrs, null,
                null, notifs);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.NotificationBroadcaster#getNotificationInfo()
     */
    public MBeanNotificationInfo[] getNotificationInfo() {
        return new MBeanNotificationInfo[] { new MBeanNotificationInfo(
                new String[] { AttributeChangeNotification.ATTRIBUTE_CHANGE },
                AttributeChangeNotification.class.getName(),
                "Change runtime configuration attribute") };
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.NotificationBroadcaster#addNotificationListener(javax
     * .management.NotificationListener, javax.management.NotificationFilter,
     * java.lang.Object)
     */
    public void addNotificationListener(final NotificationListener listener,
            final NotificationFilter filter, final Object handback) {
        this.broadcasterSupport.addNotificationListener(listener, filter, handback);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.NotificationBroadcaster#removeNotificationListener(javax
     * .management.NotificationListener)
     */
    public void removeNotificationListener(final NotificationListener listener)
            throws ListenerNotFoundException {
        this.broadcasterSupport.removeNotificationListener(listener);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.NotificationEmitter#removeNotificationListener(javax
     * .management.NotificationListener, javax.management.NotificationFilter,
     * java.lang.Object)
     */
    public void removeNotificationListener(final NotificationListener listener,
            final NotificationFilter filter, final Object handback)
            throws ListenerNotFoundException {
        this.broadcasterSupport.removeNotificationListener(listener, filter, handback);
    }

    /**
     * Get the MBean Name
     * 
     * @return
     */
    public ObjectName getMBeanName() {
        return this.mbeanName;
    }

    /**
     * 
     * @param mbeanName
     */
    public void setMBeanName(final ObjectName mbeanName) {
        this.mbeanName = mbeanName;
    }

    @Override
    protected void setCustomAttribute(final Attribute attribute) throws ReflectionException,
            AttributeNotFoundException {
        throw new AttributeNotFoundException("Attribute '" + attribute.getName()
                + "' does not exist");
    }

    @Override
    protected void saveConfiguration() {

        try {
            final OutputStream jbiComponentFileAsOutputStream = new FileOutputStream(
                    this.jbiComponentFile);
            try {
                JBIDescriptorBuilder.buildXmlJBIdescriptor(this.jbiComponentConfiguration,
                        jbiComponentFileAsOutputStream);
            } finally {
                try {
                    jbiComponentFileAsOutputStream.close();
                } catch (final IOException e) {
                    this.logger.warning("Can't close jbi component file output stream : "
                            + e.getMessage());
                }
            }
        } catch (final FileNotFoundException e) {
            this.logger
                    .warning("Failed to persist modification of the runtime configuration MBean : "
                            + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final CDKJBIDescriptorException e) {
            this.logger
                    .warning("Failed to persist modification of the runtime configuration MBean : "
                            + e.getClass().getSimpleName() + " : " + e.getMessage());
        }
    }

    @Override
    protected void notifyConfigurationChange(final String msg, final String attrName,
            final Object oldVal, final Object newVal) {

        final Notification notif = new AttributeChangeNotification(this, this.seqNumber
                .getAndIncrement(), System.currentTimeMillis(), msg, attrName, String.class
                .getName(), oldVal, newVal);
        this.broadcasterSupport.sendNotification(notif);
    }

    @Override
    protected boolean isSettableField(final Field field) {
        return field.getType().getSimpleName().startsWith("Runtime");
    }

    @Override
    protected Object getCustomAttribute(final String attributeName) throws ReflectionException,
            AttributeNotFoundException {
        throw new AttributeNotFoundException("Attribute '" + attributeName + "' does not exist");
    }
}
