/**
 * Copyright (c) 2005-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.message;

import static javax.jbi.messaging.NormalizedMessageProperties.PROTOCOL_HEADERS;

import java.io.InputStream;
import java.net.URI;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.activation.DataHandler;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchange.Role;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.commons.exception.ExceptionUtil;
import org.ow2.petals.commons.log.FlowAttributes;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.util.SourceUtil;
import org.ow2.petals.component.framework.util.WSAHelper;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;

/**
 * This is a Message Exchange Wrapper.
 * <p>
 * The "in" and "out" normalized message are
 * cached into this class, authorizing the reset of these message. Can be use by
 * the Component Framework interceptors.
 * </p>
 * 
 * TODO: cache the fault
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 * @author Olivier Fabre - EBM WebSourcing
 * @author Christophe Hamerling - EBM WebSourcing
 */
public class ExchangeImpl implements Exchange {

    /**
     * The addressing information is cached on first getAddressing call
     */
    private Map<QName, String> addressingCache;

    private boolean checkRole;

    private final MessageExchange exchange;

    private NormalizedMessage inNormalizedExchange;

    private NormalizedMessage outNormalizedExchange;

    /**
     * 
     * Creates a new instance of {@link ExchangeImpl}
     * 
     * @param exchange
     */
    public ExchangeImpl(final MessageExchange exchange) {
        super();
        this.exchange = exchange;
        this.inNormalizedExchange = this.exchange.getMessage(IN_MESSAGE_NAME);
        this.outNormalizedExchange = this.exchange.getMessage(OUT_MESSAGE_NAME);
        this.checkRole = true;
    }

    /**
     * Copy attachment to a normalized message.
     * 
     * @param attachments
     *            Attachments to copy. Must be not-null.
     * @param normalizedMesg
     *            Normalized message in which attachments must be copied. Must
     *            be not-null.
     * @throws MessagingException
     *             An error occurs when adding an attachement to the normalized
     *             message.
     */
    final private static void copyAttachments(final Map<String, DataHandler> attachments,
            final NormalizedMessage normalizedMesg) throws MessagingException {
        for (final Entry<String, DataHandler> entry : attachments.entrySet()) {
            normalizedMesg.addAttachment(entry.getKey(), entry.getValue());
        }
    }

    /**
     * Copy properties to a normalized message.
     * 
     * @param properties
     *            Properties to copy. Must be not-null.
     * @param normalizedMesg
     *            Normalized message in which properties must be copied. Must be
     *            not-null.
     */
    final private static void copyMessageProperties(final Map<String, Object> properties,
            final NormalizedMessage normalizedMsg) {
        for (final Entry<String, Object> entry : properties.entrySet()) {
            normalizedMsg.setProperty(entry.getKey(), entry.getValue());
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * setInMessageAttachment(java.lang.String, javax.activation.DataHandler)
     */
    public void addInMessageAttachment(final String id, final DataHandler content)
            throws MessagingException {
        NormalizedMessage inMess = this.getInMessage();
        if (inMess == null) {
            inMess = this.exchange.createMessage();
        }
        inMess.addAttachment(id, content);
        this.setInMessage(inMess);
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * setOutMessageAttachment(java.lang.String, javax.activation.DataHandler)
     */
    public void addOutMessageAttachment(final String id, final DataHandler content)
            throws MessagingException {
        NormalizedMessage outMess = this.getOutMessage();
        if (outMess == null) {
            outMess = this.exchange.createMessage();
        }
        outMess.addAttachment(id, content);
        this.setOutMessage(outMess);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#createFault()
     */
    public Fault createFault() throws MessagingException {
        return this.exchange.createFault();
    }

    /**
     * Creates a message.
     * <p>
     * This method is not exposed by the {@code Exchange} interface.
     * 
     * @return
     * @throws MessagingException
     */
    public NormalizedMessage createMessage() throws MessagingException {
        return this.exchange.createMessage();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#getEndpoint()
     */
    public ServiceEndpoint getEndpoint() {
        return this.exchange.getEndpoint();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#getEndpointName()
     */
    public String getEndpointName() throws MessagingException {
        final ServiceEndpoint ep = this.exchange.getEndpoint();
        if (ep == null) {
            throw new MessagingException("You must specify an endpoint in the message exchange");
        }
        return ep.getEndpointName();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.api.message.Exchange#getError()
     */
    public Exception getError() {
        return this.exchange.getError();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#getExchangeId()
     */
    public String getExchangeId() {
        return this.exchange.getExchangeId();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#getExchangePattern
     * ()
     */
    public URI getExchangePattern() throws MessagingException {
        final URI pattern = this.exchange.getPattern();
        if (pattern == null) {
            throw new MessagingException(
                    "You must specify a message exchange pattern in the message exchange");
        }
        return pattern;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.api.message.Exchange#getFault()
     */
    public Fault getFault() {
        return this.exchange.getFault();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#getInAddressing()
     */
    public Map<QName, String> getInAddressing() {
        Map<QName, String> result = null;
        if (this.addressingCache == null) {
            if (this.inNormalizedExchange != null) {
                final Object protocolHeadersPropertyObject = this.inNormalizedExchange
                        .getProperty(PROTOCOL_HEADERS);
                if (protocolHeadersPropertyObject != null
                        && protocolHeadersPropertyObject instanceof Map) {
                    final Map<String, DocumentFragment> map = (Map<String, DocumentFragment>) protocolHeadersPropertyObject;
                    result = new HashMap<QName, String>();

                    final String wsaTo = WSAHelper.getTo(map);
                    if (wsaTo != null) {
                        result.put(Constants.WSStar.Addressing.TO_QNAME, wsaTo);
                    }
                    final String wsaFrom = WSAHelper.getFrom(map);
                    if (wsaFrom != null) {
                        result.put(Constants.WSStar.Addressing.FROM_QNAME, wsaFrom);
                    }
                    final String wsaReplyTo = WSAHelper.getReplyTo(map);
                    if (wsaReplyTo != null) {
                        result.put(Constants.WSStar.Addressing.REPLY_TO_QNAME, wsaReplyTo);
                    }
                }
            }
            this.addressingCache = result;
        } else {
            result = this.addressingCache;
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#getInMessage()
     */
    public NormalizedMessage getInMessage() {
        if (this.inNormalizedExchange == null) {
            if (this.exchange.getMessage(IN_MESSAGE_NAME) != null) {
                return this.exchange.getMessage(IN_MESSAGE_NAME);
            }
            try {
                this.inNormalizedExchange = this.exchange.createMessage();
                this.exchange.setMessage(this.inNormalizedExchange, IN_MESSAGE_NAME);
            } catch (final MessagingException e) {
                return null;
            }
        }
        return this.inNormalizedExchange;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * getInMessageAttachment(java.lang.String)
     */
    public DataHandler getInMessageAttachment(final String name) throws MessagingException {

        final NormalizedMessage inMessage = this.getInMessage();
        if (inMessage == null) {
            throw new MessagingException("In message cannot be null");
        }

        return inMessage.getAttachment(name);
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * getInMessageAttachmentNames()
     */
    @SuppressWarnings("unchecked")
    public Set<String> getInMessageAttachmentNames() throws MessagingException {

        final NormalizedMessage inMessage = this.getInMessage();
        if (inMessage == null) {
            throw new MessagingException("In message cannot be null");
        }

        return (Set<String>) inMessage.getAttachmentNames();
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * getInMessageAttachments()
     */
    public Set<DataHandler> getInMessageAttachments() throws MessagingException {
        final Set<DataHandler> attachments = new HashSet<DataHandler>();

        final NormalizedMessage inMessage = this.getInMessage();
        if (inMessage == null) {
            throw new MessagingException("In message cannot be null");
        }

        final Set<?> names = inMessage.getAttachmentNames();
        for (final Object attachId : names) {
            final DataHandler dh = inMessage.getAttachment((String) attachId);
            if (dh != null) {
                attachments.add(dh);
            }
        }
        return attachments;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * getInMessageContentAsDocument()
     */
    public Document getInMessageContentAsDocument() throws MessagingException {
        return this.getInMessageContentAsDocument(false);
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * getInMessageContentAsDocument(boolean)
     */
    public Document getInMessageContentAsDocument(final boolean forkSource)
            throws MessagingException {
        final NormalizedMessage inMessage = this.getInMessage();
        Document result = null;
        final Source source = inMessage.getContent();
        try {
            if (source != null) {
                result = SourceUtil.createDocument(source, forkSource);
            }
        } catch (PEtALSCDKException e) {
            throw new MessagingException(e);
        }

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * getInMessageContentAsSource()
     */
    public Source getInMessageContentAsSource() throws MessagingException {
        final NormalizedMessage inMessage = this.getInMessage();
        if (inMessage == null) {
            throw new MessagingException("In message cannot be null");
        }
        return inMessage.getContent();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#getInMessageProperty
     * (java.lang.String)
     */
    public Object getInMessageProperty(final String name) throws MessagingException {
        final NormalizedMessage nm = this.getInMessage();
        if (nm == null) {
            throw new MessagingException("In message cannot be null");
        }
        return nm.getProperty(name);
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * getInMessagePropertyNames()
     */
    @SuppressWarnings("unchecked")
    public Set<String> getInMessagePropertyNames() throws MessagingException {
        final NormalizedMessage nm = this.getInMessage();
        if (nm == null) {
            throw new MessagingException("In message cannot be null");
        }
        return (Set<String>) nm.getPropertyNames();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#getInSubject()
     */
    public Subject getInSubject() throws MessagingException {
        final NormalizedMessage inMessage = this.getInMessage();
        if (inMessage == null) {
            throw new MessagingException("In message cannot be null");
        }

        return inMessage.getSecuritySubject();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#getInterfaceName
     * ()
     */
    public QName getInterfaceName() {
        return this.exchange.getInterfaceName();
    }

    /**
     * Return the wrapped Message Exchange.
     * <p>
     * This method is not exposed by the {@code Exchange} interface.
     * 
     * @return the wrapped Message Exchange
     * @throws MessagingException
     */
    public MessageExchange getMessageExchange() throws MessagingException {
        if (this.inNormalizedExchange != null && this.exchange.getMessage(IN_MESSAGE_NAME) == null) {
            this.exchange.setMessage(this.inNormalizedExchange, IN_MESSAGE_NAME);
        }

        if (this.outNormalizedExchange != null
                && this.exchange.getMessage(OUT_MESSAGE_NAME) == null) {
            this.exchange.setMessage(this.outNormalizedExchange, OUT_MESSAGE_NAME);
        }

        return this.exchange;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#getOperation()
     */
    public QName getOperation() {
        return this.exchange.getOperation();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#getOperationName
     * ()
     */
    public String getOperationName() {
        final QName operation = this.exchange.getOperation();
        if (operation == null) {
            return null;
        }
        return operation.getLocalPart();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#getOutMessage()
     */
    public NormalizedMessage getOutMessage() {
        if (this.outNormalizedExchange == null) {
            if (this.exchange.getMessage(OUT_MESSAGE_NAME) != null) {
                return this.exchange.getMessage(OUT_MESSAGE_NAME);
            } else {
                try {
                    this.outNormalizedExchange = this.exchange.createMessage();
                } catch (final MessagingException e) {
                    return null;
                }
            }
        }
        return this.outNormalizedExchange;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * getOutMessageAttachment(java.lang.String)
     */
    public DataHandler getOutMessageAttachment(final String name) throws MessagingException {

        final NormalizedMessage outMessage = this.getOutMessage();
        if (outMessage == null) {
            throw new MessagingException("Out message cannot be null");
        }

        return outMessage.getAttachment(name);
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * getOutMessageAttachmentNames()
     */
    @SuppressWarnings("unchecked")
    public Set<String> getOutMessageAttachmentNames() throws MessagingException {

        final NormalizedMessage outMessage = this.getOutMessage();
        if (outMessage == null) {
            throw new MessagingException("Out message cannot be null");
        }

        return (Set<String>) outMessage.getAttachmentNames();
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * getOutMessageAttachments()
     */
    public Set<DataHandler> getOutMessageAttachments() throws MessagingException {
        final Set<DataHandler> attachments = new HashSet<DataHandler>();

        final NormalizedMessage outMessage = this.getOutMessage();
        if (outMessage == null) {
            throw new MessagingException("Out message cannot be null");
        }

        final Set<?> names = outMessage.getAttachmentNames();
        for (final Object attachId : names) {
            final DataHandler dh = outMessage.getAttachment((String) attachId);
            if (dh != null) {
                attachments.add(dh);
            }
        }
        return attachments;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * getOutMessageContentAsDocument()
     */
    public Document getOutMessageContentAsDocument() throws MessagingException {
        return this.getOutMessageContentAsDocument(true);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.api.message.Exchange#
     * getOutMessageContentAsDocument(boolean)
     */
    public Document getOutMessageContentAsDocument(final boolean forkSource)
            throws MessagingException {
        final NormalizedMessage outMessage = this.getOutMessage();
        Document result = null;
        final Source source = outMessage.getContent();
        try {
            if (source != null) {
                result = SourceUtil.createDocument(source, forkSource);
            }
        } catch (PEtALSCDKException e) {
            throw new MessagingException(e);
        }

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * getOutMessageContentAsSource()
     */
    public Source getOutMessageContentAsSource() throws MessagingException {
        final NormalizedMessage outMessage = this.getOutMessage();
        if (outMessage == null) {
            throw new MessagingException("Out message cannot be null");
        }
        return outMessage.getContent();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#getOutMessageProperty
     * (java.lang.String)
     */
    public Object getOutMessageProperty(final String name) throws MessagingException {
        final NormalizedMessage nm = this.getOutMessage();
        if (nm == null) {
            throw new MessagingException("Out message cannot be null");
        }
        return nm.getProperty(name);
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * getOutMessagePropertyNames()
     */
    public Set<?> getOutMessagePropertyNames() throws MessagingException {
        final NormalizedMessage nm = this.getOutMessage();
        if (nm == null) {
            throw new MessagingException("Out message cannot be null");
        }
        return nm.getPropertyNames();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#getInSubject()
     */
    public Subject getOutSubject() throws MessagingException {
        final NormalizedMessage outMessage = this.getOutMessage();
        if (outMessage == null) {
            throw new MessagingException("Out message cannot be null");
        }

        return outMessage.getSecuritySubject();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.api.message.Exchange#getPattern()
     */
    public URI getPattern() {
        return this.exchange.getPattern();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#getProperty(java
     * .lang.String)
     */
    public Object getProperty(final String name) {
        return this.exchange.getProperty(name);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#getPropertyNames
     * ()
     */
    public Set<?> getPropertyNames() {
        return this.exchange.getPropertyNames();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.api.message.Exchange#getRole()
     */
    public MessageExchange.Role getRole() {
        return this.exchange.getRole();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.api.message.Exchange#getService()
     */
    public QName getService() {
        return this.exchange.getService();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.api.message.Exchange#getStatus()
     */
    public ExchangeStatus getStatus() {
        return this.exchange.getStatus();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#isActiveStatus()
     */
    public boolean isActiveStatus() {
        return ExchangeStatus.ACTIVE.equals(this.getStatus());
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#isConsumerRole()
     */
    public boolean isConsumerRole() {
        return MessageExchange.Role.CONSUMER.equals(this.getRole());
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#isDoneStatus()
     */
    public boolean isDoneStatus() {
        return ExchangeStatus.DONE.equals(this.getStatus());
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#isErrorStatus()
     */
    public boolean isErrorStatus() {
        return ExchangeStatus.ERROR.equals(this.getStatus());
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#isInMessage()
     */
    public boolean isInMessage() {
        return (this.exchange.getMessage(IN_MESSAGE_NAME) != null && this.exchange.getMessage(
                IN_MESSAGE_NAME).getContent() != null)
                || (this.inNormalizedExchange != null && this.inNormalizedExchange.getContent() != null);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#isInOnlyPattern()
     */
    public boolean isInOnlyPattern() {
        return MEPConstants.IN_ONLY_PATTERN.equals(this.exchange.getPattern());
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * isInOptionalOutPattern()
     */
    public boolean isInOptionalOutPattern() {
        return MEPConstants.IN_OPTIONAL_OUT_PATTERN.equals(this.exchange.getPattern());
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#isInOutPattern()
     */
    public boolean isInOutPattern() {
        return MEPConstants.IN_OUT_PATTERN.equals(this.exchange.getPattern());
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#isOutMessage()
     */
    public boolean isOutMessage() {
        return (this.exchange.getMessage(OUT_MESSAGE_NAME) != null && this.exchange.getMessage(
                OUT_MESSAGE_NAME).getContent() != null)
                || (this.outNormalizedExchange != null && this.outNormalizedExchange.getContent() != null);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#isProviderRole()
     */
    public boolean isProviderRole() {
        return MessageExchange.Role.PROVIDER.equals(this.getRole());
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#isRobustInOnlyPattern
     * ()
     */
    public boolean isRobustInOnlyPattern() {
        return MEPConstants.ROBUST_IN_ONLY_PATTERN.equals(this.exchange.getPattern());
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#isTransacted()
     */
    public boolean isTransacted() {
        return this.exchange.isTransacted();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setActiveStatus()
     */
    public void setActiveStatus() throws MessagingException {
        this.setStatus(ExchangeStatus.ACTIVE);
    }

    public void setCheckRole(final boolean checkRole) {
        this.checkRole = checkRole;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setDoneStatus()
     */
    public void setDoneStatus() throws MessagingException {
        this.setStatus(ExchangeStatus.DONE);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setEndpoint(javax
     * .jbi.servicedesc.ServiceEndpoint)
     */
    public void setEndpoint(final ServiceEndpoint endpoint) {
        this.exchange.setEndpoint(endpoint);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setError(java
     * .lang.Exception)
     */
    public void setError(final Exception exception) {
        final String packageName = exception.getClass().getPackage().getName();
        if (exception instanceof MessagingException || packageName.startsWith("java")
                || packageName.startsWith("org.w3c") || packageName.startsWith("org.xml")) {
            this.exchange.setError(exception);
        } else {
            final MessagingException msgException = new MessagingException(
                    ExceptionUtil.getExtendedMessage(exception));
            this.exchange.setError(msgException);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setErrorStatus()
     */
    public void setErrorStatus() throws MessagingException {
        this.setStatus(ExchangeStatus.ERROR);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setFault(javax
     * .jbi.messaging.Fault)
     */
    public void setFault(final Fault fault) throws MessagingException {
        this.exchange.setFault(fault);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setFault(java
     * .lang.Exception)
     */
    public void setFault(final Throwable e) throws MessagingException {
        final Fault fault = this.exchange.createFault();
        final Document faultDocument = SourceUtil.createSoapFaultDocument(e,
                Exchange.DEFAULT_FAULT_QNAME);
        if (SourceUtil.isDOM3(faultDocument)) {
            fault.setContent(SourceUtil.createDOMSource(faultDocument));
        } else {
            try {
                fault.setContent(SourceUtil.createStreamSource(faultDocument));
            } catch (final PEtALSCDKException e1) {
                throw new MessagingException(e1);
            }
        }
        this.exchange.setFault(fault);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setInAddressing
     * (java.util.Map)
     */
    public void setInAddressing(final Map<QName, String> addressing) {
        // get the protocol headers map
        final NormalizedMessage message = this.getInMessage();
        final Object o = message.getProperty(PROTOCOL_HEADERS);

        // clone: The constructor HashMap(Map) clones each element
        final Map<QName, String> remains = new HashMap<QName, String>(addressing);

        Map<String, DocumentFragment> headers = null;
        // replace data from existing map
        if (o != null && o instanceof Map) {
            headers = (Map<String, DocumentFragment>) o;
            // get the addressing Document Fragment and replace things if
            // exists, else create new fragment and insert...
            for (final Map.Entry<QName, String> entry : addressing.entrySet()) {
                final QName key = entry.getKey();
                final DocumentFragment destFrag = headers.get(key.toString());
                if (destFrag != null && destFrag.getFirstChild() != null
                        && destFrag.getFirstChild() instanceof Element) {
                    destFrag.getFirstChild().setTextContent(entry.getValue());
                    remains.remove(key);
                }
            }
        } else {
            // create map
            headers = new HashMap<String, DocumentFragment>(remains.size());
        }

        // fill map with remaining data
        for (final Map.Entry<QName, String> entry : remains.entrySet()) {
            final QName key = entry.getKey();
            final DocumentFragment df = WSAHelper.getElement(key, entry.getValue());
            if (df != null) {
                headers.put(key.toString(), df);
            }
        }

        // fill property
        message.setProperty(PROTOCOL_HEADERS, headers);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setInMessage(
     * javax.jbi.messaging.NormalizedMessage)
     */
    public void setInMessage(final NormalizedMessage msg) throws MessagingException {
        this.checkPatternMatching(IN_MESSAGE_NAME);
        this.checkRoleMatching(IN_MESSAGE_NAME);
        this.inNormalizedExchange = msg;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * setInMessageAttachments(java.util.Map)
     */
    public void setInMessageAttachment(String name, DataHandler attachment)
            throws MessagingException {
        NormalizedMessage inMess = this.getInMessage();
        if (inMess == null) {
            inMess = this.exchange.createMessage();
        }

        inMess.addAttachment(name, attachment);
        this.setInMessage(inMess);
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * setInMessageAttachments(java.util.Map)
     */
    public void setInMessageAttachments(final Map<String, DataHandler> attachments)
            throws MessagingException {
        NormalizedMessage inMess = this.getInMessage();
        if (inMess == null) {
            inMess = this.exchange.createMessage();
        }
        copyAttachments(attachments, inMess);
        this.setInMessage(inMess);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setInMessageContent
     * (org.w3c.dom.Document)
     */
    public void setInMessageContent(final Document inContent) throws MessagingException {
        NormalizedMessage inMess = this.getInMessage();
        if (inMess == null) {
            inMess = this.exchange.createMessage();
        }

        if (SourceUtil.isDOM3(inContent)) {
            inMess.setContent(SourceUtil.createDOMSource(inContent));
        } else {
            try {
                inMess.setContent(SourceUtil.createStreamSource(inContent));
            } catch (final PEtALSCDKException e) {
                throw new MessagingException(e);
            }
        }

        this.setInMessage(inMess);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setInMessageContent
     * (java.io.InputStream)
     */
    public void setInMessageContent(InputStream inContent) throws MessagingException {
        NormalizedMessage inMess = this.getInMessage();
        if (inMess == null) {
            inMess = this.exchange.createMessage();
        }

        inMess.setContent(new StreamSource(inContent));
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setInMessageContent
     * (javax.xml.transform.Source)
     */
    public void setInMessageContent(final Source inContent) throws MessagingException {
        NormalizedMessage inMess = this.getInMessage();
        if (inMess == null) {
            inMess = this.exchange.createMessage();
        }
        inMess.setContent(inContent);
        this.setInMessage(inMess);
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * setInMessageProperties(java.util.Map)
     */
    public void setInMessageProperties(final Map<String, Object> properties)
            throws MessagingException {
        NormalizedMessage inMess = this.getInMessage();
        if (inMess == null) {
            inMess = this.exchange.createMessage();
        }
        copyMessageProperties(properties, inMess);
        this.setInMessage(inMess);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setInMessageProperty
     * (java.lang.String, java.lang.Object)
     */
    public void setInMessageProperty(final String name, final Object value)
            throws MessagingException {
        NormalizedMessage inMess = this.getInMessage();
        if (inMess == null) {
            inMess = this.exchange.createMessage();
        }
        inMess.setProperty(name, value);
        this.setInMessage(inMess);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setInSubject(
     * javax.security.auth.Subject)
     */
    public void setInSubject(final Subject subject) {
        this.getInMessage().setSecuritySubject(subject);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setInterfaceName
     * (javax.xml.namespace.QName)
     */
    public void setInterfaceName(final QName interfaceName) {
        this.exchange.setInterfaceName(interfaceName);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setOperation(
     * javax.xml.namespace.QName)
     */
    public void setOperation(final QName name) {
        this.exchange.setOperation(name);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setOutMessage
     * (javax.jbi.messaging.NormalizedMessage)
     */
    public void setOutMessage(final NormalizedMessage msg) throws MessagingException {
        this.checkPatternMatching(OUT_MESSAGE_NAME);
        this.checkRoleMatching(OUT_MESSAGE_NAME);
        this.outNormalizedExchange = msg;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * setOutMessageAttachment(java.lang.String, javax.activation.DataHandler)
     */
    public void setOutMessageAttachment(String name, DataHandler attachment)
            throws MessagingException {
        NormalizedMessage outMess = this.getOutMessage();
        if (outMess == null) {
            outMess = this.exchange.createMessage();
        }

        outMess.addAttachment(name, attachment);
        this.setOutMessage(outMess);
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * setOutMessageAttachments(java.util.Map)
     */
    public void setOutMessageAttachments(final Map<String, DataHandler> attachments)
            throws MessagingException {
        NormalizedMessage outMess = this.getOutMessage();
        if (outMess == null) {
            outMess = this.exchange.createMessage();
        }
        copyAttachments(attachments, outMess);
        this.setOutMessage(outMess);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setOutMessageContent
     * (org.w3c.dom.Document)
     */
    public void setOutMessageContent(final Document outContent) throws MessagingException {
        NormalizedMessage outMess = this.getOutMessage();
        if (outMess == null) {
            outMess = this.exchange.createMessage();
        }
        if (SourceUtil.isDOM3(outContent)) {
            outMess.setContent(SourceUtil.createDOMSource(outContent));
        } else {
            try {
                outMess.setContent(SourceUtil.createStreamSource(outContent));
            } catch (final PEtALSCDKException e) {
                throw new MessagingException(e);
            }
        }

        this.setOutMessage(outMess);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setOutMessageContent
     * (java.io.InputStream)
     */
    public void setOutMessageContent(InputStream outContent) throws MessagingException {
        NormalizedMessage outMess = this.getOutMessage();
        if (outMess == null) {
            outMess = this.exchange.createMessage();
        }

        outMess.setContent(new StreamSource(outContent));
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setOutMessageContent
     * (javax.xml.transform.Source)
     */
    public void setOutMessageContent(final Source outContent) throws MessagingException {
        NormalizedMessage outMess = this.getOutMessage();
        if (outMess == null) {
            outMess = this.exchange.createMessage();
        }
        outMess.setContent(outContent);
        this.setOutMessage(outMess);
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.api.message.Exchange#
     * setOutMessageProperties(java.util.Map)
     */
    public void setOutMessageProperties(final Map<String, Object> properties)
            throws MessagingException {
        NormalizedMessage outMess = this.getOutMessage();
        if (outMess == null) {
            outMess = this.exchange.createMessage();
        }
        copyMessageProperties(properties, outMess);
        this.setOutMessage(outMess);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setOutMessageProperty
     * (java.lang.String, java.lang.Object)
     */
    public void setOutMessageProperty(final String name, final Object value)
            throws MessagingException {
        NormalizedMessage outMess = this.getOutMessage();
        if (outMess == null) {
            outMess = this.exchange.createMessage();
        }
        outMess.setProperty(name, value);
        this.setOutMessage(outMess);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setInSubject(
     * javax.security.auth.Subject)
     */
    public void setOutSubject(final Subject subject) {
        this.getOutMessage().setSecuritySubject(subject);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setProperty(java
     * .lang.String, java.lang.Object)
     */
    public void setProperty(final String name, final Object obj) {
        this.exchange.setProperty(name, obj);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setService(javax
     * .xml.namespace.QName)
     */
    public void setService(final QName service) {
        this.exchange.setService(service);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.message.Exchange#setStatus(javax
     * .jbi.messaging.ExchangeStatus)
     */
    public void setStatus(final ExchangeStatus status) throws MessagingException {
        this.checkStatus(status);

        this.exchange.setStatus(status);
    }

    /**
     * Check that the exchange MEP allows the specified reference. This only
     * works for the JBI predefined MEPs. It check that reference is IN or OUT.
     * Other reference or MEP will cause an exception
     * 
     * @param name
     * @throws MessagingException
     *             the MEP does not allow the reference
     */
    protected void checkPatternMatching(final String name) throws MessagingException {
        if (name == null) {
            throw new MessagingException("The reference name is not defined.");
        }

        if (OUT_MESSAGE_NAME.equalsIgnoreCase(name)
                && (MEPConstants.IN_ONLY_PATTERN.equals(this.exchange.getPattern()) || MEPConstants.ROBUST_IN_ONLY_PATTERN
                        .equals(this.exchange.getPattern()))) {
            throw new MessagingException("the MessageExchange state does not allow this operation.");
        }
    }

    /**
     * Check that the Role allows the specified reference. This only works for
     * the JBI predefined MEPs. It check that reference is IN or OUT and the
     * Role is CONSUMER or PROVIDER. Other reference will cause an exception
     * 
     * @param name
     * @throws MessagingException
     *             the Role does not allow the reference
     */
    protected void checkRoleMatching(final String name) throws MessagingException {
        if (name == null) {
            throw new MessagingException("The reference name is not defined.");
        }

        // The check on the role can be unactivated for CDK intercepting purpose
        if (!this.checkRole) {
            return;
        }

        if (MessageExchange.Role.CONSUMER.equals(this.exchange.getRole())) {
            if (IN_MESSAGE_NAME.equalsIgnoreCase(name)) {
                return;
            }
        } else if (MessageExchange.Role.PROVIDER.equals(this.exchange.getRole())
                && OUT_MESSAGE_NAME.equalsIgnoreCase(name)) {
            return;
        }
        throw new MessagingException("The Role does not allow this operation.");
    }

    /**
     * Check the status against the cached Normalized messages.
     * 
     * @param status
     * @throws MessagingException
     */
    private void checkStatus(final ExchangeStatus status) throws MessagingException {

        // check that DONE can be set.
        if (ExchangeStatus.DONE.equals(status)) {
            if (Role.CONSUMER.equals(this.exchange.getRole())) {
                if (MEPConstants.IN_OUT_PATTERN.equals(this.exchange.getPattern())
                        && this.getFault() == null
                        && this.exchange.getMessage(OUT_MESSAGE_NAME) == null) {
                    throw new MessagingException(
                            "The MessageExchange state does not allow this operation.");
                }
            } else {
                if (MEPConstants.ROBUST_IN_ONLY_PATTERN.equals(this.exchange.getPattern())
                        && this.getFault() != null) {
                    throw new MessagingException(
                            "The MessageExchange state does not allow this operation.");
                } else if (MEPConstants.IN_OUT_PATTERN.equals(this.exchange.getPattern())) {
                    throw new MessagingException(
                            "The MessageExchange state does not allow this operation.");
                } else if (MEPConstants.IN_OPTIONAL_OUT_PATTERN.equals(this.exchange.getPattern())) {
                    if (this.exchange.getFault() != null
                            && this.exchange.getMessage(OUT_MESSAGE_NAME) == null) {
                        throw new MessagingException(
                                "The MessageExchange state does not allow this operation.");
                    } else if (this.exchange.getFault() == null
                            && this.exchange.getMessage(OUT_MESSAGE_NAME) != null) {
                        throw new MessagingException(
                                "The MessageExchange state does not allow this operation.");
                    }
                }
            }
        }
    }

    @Override
    public FlowAttributes getFlowAttributes() {
        return this.exchange.getFlowAttributes();
    }
}
