/**
 * Copyright (c) 2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.monitoring;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;

import javax.management.MBeanNotificationInfo;
import javax.management.NotificationBroadcasterSupport;

import org.ow2.petals.component.framework.monitoring.defect.JmxDefectCreator;
import org.ow2.petals.component.framework.monitoring.defect.MessageExchangeProcessorObjectPoolDefectCreator;
import org.ow2.petals.component.framework.monitoring.defect.MessageExchangeProcessorThreadPoolDefectCreator;
import org.ow2.petals.component.framework.monitoring.key.ServiceProviderKey;
import org.ow2.petals.component.framework.monitoring.probe.MessageExchangeAcceptorPoolProbe;
import org.ow2.petals.component.framework.monitoring.probe.impl.MessageExchangeAcceptorPoolProbeImpl;
import org.ow2.petals.probes.api.KeyedProbesFactory;
import org.ow2.petals.probes.api.KeyedProbesFactoryBuilder;
import org.ow2.petals.probes.api.MacroProbesFactory;
import org.ow2.petals.probes.api.MacroProbesFactoryBuilder;
import org.ow2.petals.probes.api.ProbesFactory;
import org.ow2.petals.probes.api.ProbesFactoryBuilder;
import org.ow2.petals.probes.api.enums.ExecutionStatus;
import org.ow2.petals.probes.api.exceptions.MultipleProbesFactoriesFoundException;
import org.ow2.petals.probes.api.exceptions.NoProbesFactoryFoundException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.key.ExecutionStatusProbeKey;
import org.ow2.petals.probes.api.key.ProbeKey;
import org.ow2.petals.probes.api.probes.CounterProbe;
import org.ow2.petals.probes.api.probes.DurationProbe;
import org.ow2.petals.probes.api.probes.KeyedCounterProbeWithExecutionStatus;
import org.ow2.petals.probes.api.probes.KeyedResponseTimeProbe;
import org.ow2.petals.probes.api.probes.KeyedStartDateItem;
import org.ow2.petals.probes.api.probes.ResponseTimeRelativeValues;
import org.ow2.petals.probes.api.probes.macro.ObjectPoolProbe;
import org.ow2.petals.probes.api.probes.macro.ThreadPoolProbe;

/**
 * The monitoring MBean of the component
 * 
 * @author Christophe DENEUX - Linagora
 */
public class Monitoring extends NotificationBroadcasterSupport implements MonitoringMBean {

    /**
     * Defect notification sequence number
     */
    private final AtomicLong notificationSeqNum = new AtomicLong(0);

    // --- Probes --- //

    /**
     * The probe gauging the message exchange acceptors.
     */
    private final MessageExchangeAcceptorPoolProbe probeMessageExchangeAcceptorPool;

    /**
     * The macro probe about the message exchange processor object pool.
     */
    private final ObjectPoolProbe probeMessageExchangeProcessorObjectPool;

    /**
     * The probe counting the number of exhaustion of the message exchange
     * processor object pool
     */
    private final CounterProbe probeMessageExchangeProcessorObjectPoolExhaustion;

    /**
     * The macro probe about the message exchange processor thread pool.
     */
    private final ThreadPoolProbe probeMessageExchangeProcessorThreadPool;

    /**
     * The probe about the processing duration of message exchange processors
     */
    private final DurationProbe probeMessageExchangeProcessorDuration;

    /**
     * The probe about response times of service providers
     */
    private final KeyedResponseTimeProbe<ServiceProviderKey> probeServiceProviderResponseTimes;

    /**
     * The pending start date items of the probe about response times of service
     * providers
     */
    private final ConcurrentHashMap<String, KeyedStartDateItem<ServiceProviderKey>> serviceProviderStartDateItems = new ConcurrentHashMap<String, KeyedStartDateItem<ServiceProviderKey>>();

    /**
     * The probe counting invokations of service providers
     */
    private final KeyedCounterProbeWithExecutionStatus<ProbeKey> probeServiceProviderInvokations;

    /**
     * The pending invokation keys of the probe counting service provider
     * invokations
     */
    private final ConcurrentHashMap<String, ServiceProviderKey> serviceProviderInvokationKeys = new ConcurrentHashMap<String, ServiceProviderKey>();

    // --- Defect creators --- //

    /**
     * The defect creator associated to the defect 'exhaustion of the message
     * exchange processors object pool'.
     */
    private JmxDefectCreator messageExchangeProcessorObjectPoolDefectCreator;

    /**
     * The defect creator associated to the defect 'exhaustion of the message
     * exchange processors thread pool'.
     */
    private JmxDefectCreator messageExchangeProcessorThreadPoolDefectCreator;

    /**
     * Creates the monitoring MBean
     * 
     * @param responseTimeProbesTimer
     *            The timer used as sampler by response time probes. <b>The
     *            caller is responsible to cancel the timer to free resources
     *            attached</b>.
     * @param samplePeriod
     *            The duration of a sample, in milliseconds.
     * @throws NoProbesFactoryFoundException
     *             No probe implementation found in the classloader
     * @throws MultipleProbesFactoriesFoundException
     *             Several probe implementations found in the classloader
     */
    public Monitoring(final Timer responseTimeProbesTimer, final long samplePeriod)
            throws MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final MacroProbesFactoryBuilder macroProbesFactoryBuilder = new MacroProbesFactoryBuilder();
        final MacroProbesFactory macroProbesFactory = macroProbesFactoryBuilder
                .getMacroProbesFactory();

        this.messageExchangeProcessorObjectPoolDefectCreator = new MessageExchangeProcessorObjectPoolDefectCreator(
                this);
        this.probeMessageExchangeProcessorObjectPool = macroProbesFactory
                .createGenericObjectPoolProbe(this.messageExchangeProcessorObjectPoolDefectCreator);

        this.messageExchangeProcessorThreadPoolDefectCreator = new MessageExchangeProcessorThreadPoolDefectCreator(
                this);
        this.probeMessageExchangeProcessorThreadPool = macroProbesFactory
                .createThreadPoolProbe(this.messageExchangeProcessorThreadPoolDefectCreator);

        final ProbesFactoryBuilder<Long, Long> longProbesFactoryBuilder = new ProbesFactoryBuilder<Long, Long>();
        final ProbesFactory<Long, Long> longProbesFactory = longProbesFactoryBuilder.getProbesFactory();
        this.probeMessageExchangeProcessorObjectPoolExhaustion = longProbesFactory
                .createCounterProbe();
        this.probeMessageExchangeProcessorDuration = longProbesFactory.createDurationProbe(
                responseTimeProbesTimer, samplePeriod);
        
        this.probeMessageExchangeAcceptorPool = new MessageExchangeAcceptorPoolProbeImpl(this);

        // Keyed probe for service providers
        final KeyedProbesFactoryBuilder<ServiceProviderKey, Long, Long> serviceProvidersProbesFactoryBuilder = new KeyedProbesFactoryBuilder<ServiceProviderKey, Long, Long>();
        final KeyedProbesFactory<ServiceProviderKey, Long, Long> serviceProvidersProbesFactory = serviceProvidersProbesFactoryBuilder
                .getKeyedProbesFactory();
        this.probeServiceProviderResponseTimes = serviceProvidersProbesFactory
                .createKeyedResponseTimeProbe(responseTimeProbesTimer, samplePeriod);
        this.probeServiceProviderInvokations = serviceProvidersProbesFactory
                .createKeyedCounterProbeWithExecutionStatus();

    }

    /**
     * @return A new sequence number, used to send JMX notification
     */
    public long getNotificationSeqNum() {
        return this.notificationSeqNum.getAndIncrement();
    }

    /**
     * @return The probe of the message exchange acceptor pool
     */
    public MessageExchangeAcceptorPoolProbe getMessageExchangeAcceptorPoolProbe() {
        return this.probeMessageExchangeAcceptorPool;
    }

    /**
     * @return The probe of the message exchange processor object pool
     */
    public ObjectPoolProbe getMessageExchangeProcessorObjectPoolProbe() {
        return this.probeMessageExchangeProcessorObjectPool;
    }

    /**
     * @return The probe counting exhaustion of the message exchange processor
     *         object pool
     */
    public CounterProbe getMessageExchangeProcessorObjectPoolExhaustionProbe() {
        return this.probeMessageExchangeProcessorObjectPoolExhaustion;
    }

    /**
     * @return The probe of the message exchange processor thread pool
     */
    public ThreadPoolProbe getMessageExchangeProcessorThreadPoolProbe() {
        return this.probeMessageExchangeProcessorThreadPool;
    }

    /**
     * @return The probe about the processing duration of message exchange
     *         processors
     */
    public DurationProbe getMessageExchangeProcessorDurationProbe() {
        return this.probeMessageExchangeProcessorDuration;
    }

    /**
     * @return The probe about the response times of service providers
     */
    public KeyedResponseTimeProbe<ServiceProviderKey> getServiceProviderResponseTimesProbe() {
        return this.probeServiceProviderResponseTimes;
    }

    /**
     * @return The pending start date items of the probe about response times of
     *         service providers
     */
    public ConcurrentHashMap<String, KeyedStartDateItem<ServiceProviderKey>> getServiceProviderStartDateItems() {
        return this.serviceProviderStartDateItems;
    }

    /**
     * @return The probe counting the invokations of service providers
     */
    public KeyedCounterProbeWithExecutionStatus<ProbeKey> getServiceProviderInvokationsProbe() {
        return this.probeServiceProviderInvokations;
    }

    /**
     * @return The pending invokation keys of the probe counting service
     *         provider invokations
     */
    public ConcurrentHashMap<String, ServiceProviderKey> getServiceProviderInvokationKeys() {
        return this.serviceProviderInvokationKeys;
    }

    /**
     * Initialize all probes of the monitoring MBean
     * 
     * @throws ProbeInitializedException
     *             A probe is already initialized
     * @throws ProbeStartedException
     *             A probe is started
     * @throws ProbeInitializationException
     *             An error occurs during a probe initialization
     */
    public void init() throws ProbeInitializedException, ProbeStartedException,
            ProbeInitializationException {

        this.probeMessageExchangeAcceptorPool.init();
        this.probeMessageExchangeProcessorObjectPool.init();
        this.probeMessageExchangeProcessorThreadPool.init();
        this.probeMessageExchangeProcessorObjectPoolExhaustion.init();
        this.probeMessageExchangeProcessorDuration.init();
        this.probeServiceProviderResponseTimes.init();
        this.probeServiceProviderInvokations.init();

        this.doInit();
    }

    /**
     * Add here the initialization of all your probes
     * 
     * @throws ProbeInitializedException
     *             A probe is already initialized
     * @throws ProbeStartedException
     *             A probe is started
     * @throws ProbeInitializationException
     *             An error occurs during a probe initialization
     * 
     */
    public void doInit() throws ProbeInitializedException, ProbeStartedException,
            ProbeInitializationException {
        // NOP
    }

    /**
     * Start all probes of the monitoring MBean
     * 
     * @throws ProbeStartupException
     *             An error occurs starting a probe
     * @throws ProbeStartedException
     *             A probe is already started
     * @throws ProbeNotInitializedException
     *             A probe is not initialized
     */
    public void start() throws ProbeNotInitializedException, ProbeStartedException,
            ProbeStartupException {
        this.probeMessageExchangeAcceptorPool.start();
        this.probeMessageExchangeProcessorObjectPool.start();
        this.probeMessageExchangeProcessorThreadPool.start();
        this.probeMessageExchangeProcessorObjectPoolExhaustion.start();
        this.probeMessageExchangeProcessorDuration.start();
        this.probeServiceProviderResponseTimes.start();
        this.probeServiceProviderInvokations.start();

        this.doStart();
    }

    /**
     * Add here the start of your own probes
     * 
     * @throws ProbeNotInitializedException
     *             An error occurs starting a probe
     * @throws ProbeStartedException
     *             A probe is already started
     * @throws ProbeStartupException
     *             A probe is not initialized
     */
    protected void doStart() throws ProbeNotInitializedException, ProbeStartedException,
            ProbeStartupException {
        // NOP
    }

    /**
     * Stop all probes of the monitoring MBean
     * 
     * @throws ProbeStopException
     *             An error occurs during the stop of a probe
     * @throws ProbeNotStartedException
     *             A probe is not started
     * @throws ProbeNotInitializedException
     *             A probe is not initialized
     * 
     */
    public void stop() throws ProbeNotInitializedException, ProbeNotStartedException,
            ProbeStopException {
        this.probeMessageExchangeAcceptorPool.stop();
        this.probeMessageExchangeProcessorObjectPool.stop();
        this.probeMessageExchangeProcessorThreadPool.stop();
        this.probeMessageExchangeProcessorObjectPoolExhaustion.stop();
        this.probeMessageExchangeProcessorDuration.stop();
        this.probeServiceProviderResponseTimes.stop();
        this.probeServiceProviderInvokations.stop();

        this.doStop();
    }

    /**
     * Add here the stop of all your probes
     * 
     * @throws ProbeStopException
     *             An error occurs during the stop of a probe
     * @throws ProbeNotStartedException
     *             A probe is not started
     * @throws ProbeNotInitializedException
     *             A probe is not initialized
     * 
     */
    public void doStop() throws ProbeNotInitializedException, ProbeNotStartedException,
            ProbeStopException {
        // NOP
    }

    /**
     * Shutdown all probes of the monitoring MBean
     * 
     * @throws ProbeNotInitializedException
     *             A probe is not initialized
     * @throws ProbeStartedException
     *             A probe is started
     * @throws ProbeShutdownException
     *             An error occurs during the shutdown of a probe
     * 
     */
    public void shutdown() throws ProbeShutdownException, ProbeStartedException,
            ProbeNotInitializedException {
        this.probeMessageExchangeAcceptorPool.shutdown();
        this.probeMessageExchangeProcessorObjectPool.shutdown();
        this.probeMessageExchangeProcessorThreadPool.shutdown();
        this.probeMessageExchangeProcessorObjectPoolExhaustion.shutdown();
        this.probeMessageExchangeProcessorDuration.shutdown();
        this.probeServiceProviderResponseTimes.shutdown();
        this.probeServiceProviderInvokations.shutdown();

        this.doShutdown();
    }

    /**
     * Add here the shutdown of all you probes
     * 
     * @throws ProbeNotInitializedException
     *             A probe is not initialized
     * @throws ProbeStartedException
     *             A probe is started
     * @throws ProbeShutdownException
     *             An error occurs during the shutdown of a probe
     * 
     */
    public void doShutdown() throws ProbeShutdownException, ProbeStartedException,
            ProbeNotInitializedException {
        // NOP
    }

    @Override
    public MBeanNotificationInfo[] getNotificationInfo() {
        final List<MBeanNotificationInfo> notifInfos = new ArrayList<MBeanNotificationInfo>();
        notifInfos
                .addAll(Arrays.asList(this.probeMessageExchangeAcceptorPool.getNotificationInfo()));
        notifInfos.add(this.messageExchangeProcessorObjectPoolDefectCreator.getNotificationInfo());
        notifInfos.add(this.messageExchangeProcessorThreadPoolDefectCreator.getNotificationInfo());

        notifInfos.addAll(this.addNotificationInfo());

        return notifInfos.toArray(new MBeanNotificationInfo[notifInfos.size()]);
    }

    /**
     * Add here all information about dedicated notification of the component
     */
    public List<MBeanNotificationInfo> addNotificationInfo() {
        return Collections.EMPTY_LIST;
    }

    // Probe accessors //

    //
    // --- Message exchange acceptor thread pool metrics
    //

    @Override
    public int getMessageExchangeAcceptorThreadPoolMaxSize() throws ProbeNotInitializedException {
        return this.probeMessageExchangeAcceptorPool.getMaxValue();
    }

    @Override
    public int getMessageExchangeAcceptorThreadPoolCurrentSize() throws ProbeNotStartedException {
        return this.probeMessageExchangeAcceptorPool.getInstantValue();
    }

    //
    // --- Message exchange processor thread pool metrics
    //

    @Override
    public long getMessageExchangeProcessorThreadPoolMaxSize() {
        return this.probeMessageExchangeProcessorThreadPool.getThreadPoolMaxSize();
    }

    @Override
    public long getMessageExchangeProcessorThreadPoolMinSize() {
        return this.probeMessageExchangeProcessorThreadPool.getThreadPoolMinSize();
    }

    @Override
    public long getMessageExchangeProcessorThreadPoolAllocatedThreadsMax()
            throws ProbeNotInitializedException {
        return this.probeMessageExchangeProcessorThreadPool.getThreadPoolAllocatedThreadsMax();
    }

    @Override
    public long getMessageExchangeProcessorThreadPoolAllocatedThreadsCurrent()
            throws ProbeNotStartedException {
        return this.probeMessageExchangeProcessorThreadPool.getThreadPoolAllocatedThreadsCurrent();
    }

    @Override
    public long getMessageExchangeProcessorThreadPoolIdleThreadsMax()
            throws ProbeNotInitializedException {
        return this.probeMessageExchangeProcessorThreadPool.getThreadPoolIdleThreadsMax();
    }

    @Override
    public long getMessageExchangeProcessorThreadPoolIdleThreadsCurrent()
            throws ProbeNotStartedException {
        return this.probeMessageExchangeProcessorThreadPool.getThreadPoolIdleThreadsCurrent();
    }

    @Override
    public long getMessageExchangeProcessorThreadPoolQueuedRequestsMax()
            throws ProbeNotInitializedException {
        return this.probeMessageExchangeProcessorThreadPool.getThreadPoolQueuedRequestsMax();
    }

    @Override
    public long getMessageExchangeProcessorThreadPoolQueuedRequestsCurrent()
            throws ProbeNotStartedException {
        return this.probeMessageExchangeProcessorThreadPool.getThreadPoolQueuedRequestsCurrent();
    }

    //
    // --- Message exchange processor object pool metrics
    //

    @Override
    public long getMessageExchangeProcessorObjectPoolMaxSize() {
        return this.probeMessageExchangeProcessorObjectPool.getObjectPoolMaxSize();
    }

    @Override
    public long getMessageExchangeProcessorObjectPoolMinIdleSize() {
        return this.probeMessageExchangeProcessorObjectPool.getObjectPoolMinIdleSize();
    }

    @Override
    public long getMessageExchangeProcessorObjectPoolBorrowedObjectsMax()
            throws ProbeNotInitializedException {
        return this.probeMessageExchangeProcessorObjectPool.getObjectPoolBorrowedObjectsMax();
    }

    @Override
    public long getMessageExchangeProcessorObjectPoolBorrowedObjectsCurrent()
            throws ProbeNotStartedException {
        return this.probeMessageExchangeProcessorObjectPool.getObjectPoolBorrowedObjectsCurrent();
    }

    @Override
    public long getMessageExchangeProcessorObjectPoolIdleObjectsMax()
            throws ProbeNotInitializedException {
        return this.probeMessageExchangeProcessorObjectPool.getObjectPoolIdleObjectsMax();
    }

    @Override
    public long getMessageExchangeProcessorObjectPoolIdleObjectsCurrent()
            throws ProbeNotStartedException {
        return this.probeMessageExchangeProcessorObjectPool.getObjectPoolIdleObjectsCurrent();
    }

    @Override
    public long getMessageExchangeProcessorObjectPoolExhaustion()
            throws ProbeNotInitializedException {
        return this.probeMessageExchangeProcessorObjectPoolExhaustion.getValue();
    }

    //
    // --- Message exchange processors (processing duration)
    //

    @Override
    public Long[] getMessageExchangeProcessorAbsoluteDurations()
            throws ProbeNotInitializedException {
        final Long[] values = this.probeMessageExchangeProcessorDuration
                .getConvertedAbsoluteResponseTimeValues().get(ExecutionStatus.SUCCEEDED.name());
        if (values == null) {
            return new Long[] { 0L, 0L, 0L };
        } else {
            return values;
        }
    }

    @Override
    public Long[] getMessageExchangeProcessorRelativeDurations()
            throws ProbeNotStartedException {
        final Long[] values = this.probeMessageExchangeProcessorDuration
                .getConvertedRelativeResponseTimeValues().get(ExecutionStatus.SUCCEEDED.name());
        if (values == null) {
            return new Long[] { 0L, 0L, 0L, 0L, 0L, 0L };
        } else {
            return values;
        }
    }

    //
    // --- Response times of service providers
    //

    @Override
    public Map<String[], Long[]> getServiceProviderInvokationsResponseTimeAbs()
            throws ProbeNotInitializedException {
        return this.probeServiceProviderResponseTimes.getConvertedAbsoluteResponseTimeValues();
    }

    @Override
    public Map<String[], Long[]> getServiceProviderInvokationsResponseTimeRel()
            throws ProbeNotStartedException {
        return this.probeServiceProviderResponseTimes.getConvertedRelativeResponseTimeValues();
    }

    @Override
    public Map<String[], Long[]> getServiceProviderInvokationsResponseTimeRelGroupedByItfSvcOpAndExecStatus()
            throws ProbeNotStartedException {

        final Map<String[], Long[]> agregatedResults = new HashMap<String[], Long[]>();

        final Map<ExecutionStatusProbeKey<ServiceProviderKey>, ResponseTimeRelativeValues> responseTimeValues = this.probeServiceProviderResponseTimes
                .getRelativeResponseTimeValues();
        final Iterator<Map.Entry<ExecutionStatusProbeKey<ServiceProviderKey>, ResponseTimeRelativeValues>> itResponseTimeValues = responseTimeValues
                .entrySet().iterator();
        while (itResponseTimeValues.hasNext()) {
            final Map.Entry<ExecutionStatusProbeKey<ServiceProviderKey>, ResponseTimeRelativeValues> responseTimeValue = itResponseTimeValues
                    .next();
            final ExecutionStatusProbeKey<ServiceProviderKey> key = responseTimeValue.getKey();
            final ServiceProviderKey originalKey = key.getOriginalKey();
            final String[] newKey = new String[] { originalKey.getInterfaceName().toString(),
                    originalKey.getServiceName().toString(), originalKey.getOperation().toString(),
                    key.getExecutionStatus().toString() };
            if (agregatedResults.containsKey(newKey)) {
                // A new mep found --> Aggregation
                final Long[] newValues = agregatedResults.get(newKey);
                final ResponseTimeRelativeValues values = responseTimeValue.getValue();
                newValues[0] = Math.max(values.getMax(), newValues[0]);
                newValues[1] = Math.min(values.getMin(), newValues[1]);
                newValues[2] += values.getResponseTimesSum();
                newValues[3] += values.getResponseTimesNb();
                // TODO: How to aggregate 2 values of percentile ?
                newValues[3] = values.getPercent10();
                newValues[4] = values.getPercent50();
                newValues[5] = values.getPercent90();
                agregatedResults.put(newKey, newValues);

            } else {
                final ResponseTimeRelativeValues values = responseTimeValue.getValue();
                final Long[] newValues = new Long[7];
                newValues[0] = values.getMax();
                newValues[1] = values.getMin();
                newValues[2] = values.getResponseTimesSum();
                newValues[3] = values.getResponseTimesNb();
                newValues[4] = values.getPercent10();
                newValues[5] = values.getPercent50();
                newValues[6] = values.getPercent90();
                agregatedResults.put(newKey, newValues);
            }
        }

        final Map<String[], Long[]> results = new HashMap<String[], Long[]>();
        for (final Map.Entry<String[], Long[]> entry : agregatedResults.entrySet()) {
            final Long[] values = entry.getValue();
            final Long[] newValues = new Long[6];
            newValues[0] = values[0];
            newValues[1] = values[2] / values[3];
            newValues[2] = values[1];
            newValues[3] = values[4];
            newValues[4] = values[5];
            newValues[5] = values[6];

            results.put(entry.getKey(), newValues);
        }

        return results;
    }

    //
    // --- Counters of service providers
    //
    @Override
    public Map<String[], Long> getServiceProviderInvokations() throws ProbeNotInitializedException {
        return this.probeServiceProviderInvokations.getConvertedValues();
    }

}
