/**
 * Copyright (c) 2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */

package org.ow2.petals.component.framework.monitoring;

import java.util.Map;

import org.ow2.petals.probes.api.enums.ExecutionStatus;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;

/**
 * Interface of the MBean monitoring of the component
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public interface MonitoringMBean {

    //
    // --- Message exchange acceptor thread pool metrics
    //

    /**
     * @return The max size of the message exchange acceptor thread pool.
     * @throws ProbeNotInitializedException
     *             The probe is not initialized or shutdown.
     */
    public int getMessageExchangeAcceptorThreadPoolMaxSize() throws ProbeNotInitializedException;

    /**
     * @return The current size of the message exchange acceptor thread pool.
     * @throws ProbeNotStartedException
     *             The probe is not started or stopped.
     */
    public int getMessageExchangeAcceptorThreadPoolCurrentSize() throws ProbeNotStartedException;

    //
    // --- Message exchange processor thread pool metrics
    //

    /**
     * @return The current max size of the message exchange processor thread
     *         pool.
     */
    public long getMessageExchangeProcessorThreadPoolMaxSize();

    /**
     * @return The current min size of the message exchange processor thread
     *         pool.
     */
    public long getMessageExchangeProcessorThreadPoolMinSize();

    /**
     * @return The max number of allocated threads of the message exchange
     *         processor thread pool.
     * @throws ProbeNotInitializedException
     *             The probe is not initialized or shutdown.
     */
    public long getMessageExchangeProcessorThreadPoolAllocatedThreadsMax()
            throws ProbeNotInitializedException;

    /**
     * @return The current number of allocated threads of the message exchange
     *         processor thread pool.
     * @throws ProbeNotStartedException
     *             The probe is not started or stopped.
     */
    public long getMessageExchangeProcessorThreadPoolAllocatedThreadsCurrent()
            throws ProbeNotStartedException;

    /**
     * @return The max number of idle threads of the message exchange processor
     *         thread pool.
     * @throws ProbeNotInitializedException
     *             The probe is not initialized or shutdown.
     */
    public long getMessageExchangeProcessorThreadPoolIdleThreadsMax()
            throws ProbeNotInitializedException;

    /**
     * @return The current number of idle threads of the message exchange
     *         processor thread pool.
     * @throws ProbeNotStartedException
     *             The probe is not started or stopped.
     */
    public long getMessageExchangeProcessorThreadPoolIdleThreadsCurrent()
            throws ProbeNotStartedException;

    /**
     * @return The max number of enqueued tasks into the message exchange
     *         processor thread pool.
     * @throws ProbeNotInitializedException
     *             The probe is not initialized or shutdown.
     */
    public long getMessageExchangeProcessorThreadPoolQueuedRequestsMax()
            throws ProbeNotInitializedException;

    /**
     * @return The current number of enqueued tasks into the message exchange
     *         processor thread pool.
     * @throws ProbeNotStartedException
     *             The probe is not started or stopped.
     */
    public long getMessageExchangeProcessorThreadPoolQueuedRequestsCurrent()
            throws ProbeNotStartedException;

    //
    // --- Message exchange processor object pool metrics
    //

    /**
     * @return The current max size of objects that the message exchange
     *         processor object pool can accept.
     */
    public long getMessageExchangeProcessorObjectPoolMaxSize();

    /**
     * @return The current min size of idle objects under which the message
     *         exchange processor pool will spawn new ones.
     */
    public long getMessageExchangeProcessorObjectPoolMinIdleSize();

    /**
     * @return The max number of borrowed objects of the message exchange
     *         processor object pool.
     * @throws ProbeNotInitializedException
     *             The probe is not initialized or shutdown.
     */
    public long getMessageExchangeProcessorObjectPoolBorrowedObjectsMax()
            throws ProbeNotInitializedException;

    /**
     * @return The current number of borrowed objects of the message exchange
     *         processor object pool.
     * @throws ProbeNotStartedException
     *             The probe is not started or stopped.
     */
    public long getMessageExchangeProcessorObjectPoolBorrowedObjectsCurrent()
            throws ProbeNotStartedException;

    /**
     * @return The max number of idle objects of the message exchange processor
     *         object pool.
     * @throws ProbeNotInitializedException
     *             The probe is not initialized or shutdown.
     */
    public long getMessageExchangeProcessorObjectPoolIdleObjectsMax()
            throws ProbeNotInitializedException;

    /**
     * @return The current number of idle objects of the message exchange
     *         processor object pool.
     * @throws ProbeNotStartedException
     *             The probe is not started or stopped.
     */
    public long getMessageExchangeProcessorObjectPoolIdleObjectsCurrent()
            throws ProbeNotStartedException;

    /**
     * @return The exhaustion number of the message exchange processor object
     *         pool since the last start of the component.
     * @throws ProbeNotInitializedException
     *             The probe is not initialized or shutdown.
     */
    public long getMessageExchangeProcessorObjectPoolExhaustion()
            throws ProbeNotInitializedException;

    //
    // --- Message exchange processors (processing duration)
    //

    /**
     * @return The absolute durations of the message exchange processors:
     *         <ol>
     *         <li>the first long value is the max response time,</li>
     *         <li>the second long value is the average of response times,</li>
     *         <li>the third long value is the min response time.</li>
     *         </ol>
     * @throws ProbeNotInitializedException
     *             The probe is not initialized or shutdown.
     */
    public Long[] getMessageExchangeProcessorAbsoluteDurations()
            throws ProbeNotInitializedException;

    /**
     * @return The relative durations of the message exchange processors:
     *         <ol>
     *         <li>the first long value is the max response time,</li>
     *         <li>the second long value is the average of response times,</li>
     *         <li>the third long value is the min response time,</li>
     *         <li>the fourth long value is the 10-percentile response time,</li>
     *         <li>the fifth long value is the 50-percentile response time,</li>
     *         <li>the sixth long value is the 90-percentile response time.</li>
     *         </ol>
     * @throws ProbeNotStartedException
     *             The probe is not started or stopped.
     */
    public Long[] getMessageExchangeProcessorRelativeDurations() throws ProbeNotStartedException;

    //
    // --- Response times of service providers
    //

    /**
     * <p>
     * Returns the response time values of service providers since the last
     * start.
     * </p>
     * <p>
     * The map key is structured as a {@link String} array where:
     * <ul>
     * <li>1st: the interface name (as qualified name) of the invoked service,</li>
     * <li>2nd: the service name (as qualified name) of the invoked service,</li>
     * <li>3rd: the operation invoked (as qualified name),</li>
     * <li>4th: the message exchange pattern</li>
     * <li>5th: the execution status (see {@link ExecutionStatus} ).</li>
     * </ul>
     * Each value of the map entry is an array of response times where:
     * <ul>
     * <li>the first value is the max response time,</li>
     * <li>the second value is the average,</li>
     * <li>the third value is the min value.</li>
     * </ul>
     * </p>
     * 
     * @return The absolute response time values.
     * @throws ProbeNotStartedException
     *             The probe is not started or stopped.
     */
    public Map<String[], Long[]> getServiceProviderInvokationsResponseTimeAbs()
            throws ProbeNotInitializedException;

    /**
     * <p>
     * Returns the response time values of service providers on a time sample.
     * </p>
     * <p>
     * The map key is an array containing:
     * <ul>
     * <li>1st: the interface name (as qualified name) of the invoked service,</li>
     * <li>2nd: the service name (as qualified name) of the invoked service,</li>
     * <li>3rd: the operation invoked (as qualified name),</li>
     * <li>4th: the message exchange pattern</li>
     * <li>5th: the execution status (see {@link ExecutionStatus} ).</li>
     * </ul>
     * </p>
     * <p>
     * Each value of the map entry is an array of response times where:
     * <ul>
     * <li>the first value is the max response time on the time sample,</li>
     * <li>the second value is the average,</li>
     * <li>the third value is the min value,</li>
     * <li>the fourth value is the 10-percentile,</li>
     * <li>the fifth value is the 50-percentile,</li>
     * <li>the sixth value is the 90-percentile.</li>
     * </ul>
     * </p>
     * 
     * @return The response time values on a time sample.
     * @throws ProbeNotStartedException
     *             The probe is not started or stopped.
     */
    public Map<String[], Long[]> getServiceProviderInvokationsResponseTimeRel()
            throws ProbeNotStartedException;

    /**
     * <p>
     * Returns the response time values of service providers on a time sample
     * group by interface, service,operation and execution status
     * </p>
     * <p>
     * The map key is an array containing:
     * <ul>
     * <li>1st: the interface name (as qualified name) of the invoked service,</li>
     * <li>2nd: the service name (as qualified name) of the invoked service,</li>
     * <li>3rd: the operation invoked (as qualified name),</li>
     * <li>4th: the execution status (see {@link ExecutionStatus} ).</li>
     * </ul>
     * </p>
     * <p>
     * Each value of the map entry is an array of response times where:
     * <ul>
     * <li>the first value is the max response time on the time sample,</li>
     * <li>the second value is the average,</li>
     * <li>the third value is the min value,</li>
     * <li>the fourth value is the 10-percentile,</li>
     * <li>the fifth value is the 50-percentile,</li>
     * <li>the sixth value is the 90-percentile.</li>
     * </ul>
     * </p>
     * 
     * @return The response time values. All execution status are contained in
     *         the map, even if set to 0.
     * 
     * @throws ProbeNotStartedException
     */
    public Map<String[], Long[]> getServiceProviderInvokationsResponseTimeRelGroupedByItfSvcOpAndExecStatus()
            throws ProbeNotStartedException;

    //
    // --- Counters of service providers
    //

    /**
     * <p>
     * Returns the invokation number of service providers.
     * </p>
     * <p>
     * The map key is an array containing:
     * <ul>
     * <li>1st: the interface name (as qualified name) of the invoked service,</li>
     * <li>2nd: the service name (as qualified name) of the invoked service,</li>
     * <li>3rd: the operation invoked (as qualified name),</li>
     * <li>4th: the message exchange pattern</li>
     * <li>5th: the execution status (see {@link ExecutionStatus} ).</li>
     * </ul>
     * </p>
     * <p>
     * Each value of the map entry is the number of invokations </ul>
     * </p>
     * 
     * @return The invokation number of service providers.
     * @throws ProbeNotInitializedException
     *             The probe is not initialized or shutdown.
     */
    public Map<String[], Long> getServiceProviderInvokations() throws ProbeNotInitializedException;
}
