/**
 * Copyright (c) 2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.monitoring.key;

import java.net.URI;

import javax.xml.namespace.QName;

import org.ow2.petals.probes.api.key.ProbeKey;

/**
 * <p>
 * A probe key identifying a service provider:
 * <ol>
 * <li>an interface name,</li>
 * <li>a service name,</li>
 * <li>invoked operation,</li>
 * <li>message exchange pattern used.</li>
 * </ol>
 * </p>
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ServiceProviderKey implements ProbeKey {

    /**
     * Interface name of the service provider
     */
    private final QName interfaceName;

	    /**
     * Service name of the service provider
     */
    private final QName serviceName;

    /**
     * Invoked operation
     */
    private final QName operation;

	/**
	 * MEP used
	 */
	private final URI mep;

	/**
	 * For optimization reason, the string representation of the
	 * ServiceClientKey is computed at constructor level as it is immutable, and
	 * not each time it is needed.
	 */
	private final String[] keyAsStringArray;

	/**
	 * Creates a new instance of ServiceClientKey
	 * 
	 * @param address
	 *            URL of the external web-service
	 * @param operation
	 *            Invoked operation
	 * @param mep
	 *            MEP used
	 */
    public ServiceProviderKey(final QName interfaceName, final QName serviceName,
            final QName operation,
			final URI mep) {
        this.interfaceName = interfaceName;
        this.serviceName = serviceName;
        this.operation = operation;
		this.mep = mep;
        this.keyAsStringArray = new String[] { this.interfaceName.toString(),
                this.serviceName.toString(), this.operation.toString(),
				this.mep.toString() };
	}

	@Override
	public String[] toStringArray() {
		return this.keyAsStringArray;
	}

    @Override
    public String toReadableString() {
        return new StringBuilder(this.interfaceName.toString()).append(", ")
                .append(this.serviceName.toString()).append(", ").append(this.mep.toASCIIString())
                .append(", ").append(this.operation).toString();
    }

    // Generated by Eclipse
    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#hashCode()
     */
    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((interfaceName == null) ? 0 : interfaceName.hashCode());
        result = prime * result + ((mep == null) ? 0 : mep.hashCode());
        result = prime * result + ((operation == null) ? 0 : operation.hashCode());
        result = prime * result + ((serviceName == null) ? 0 : serviceName.hashCode());
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (!(obj instanceof ServiceProviderKey)) {
            return false;
        }
        ServiceProviderKey other = (ServiceProviderKey) obj;
        if (interfaceName == null) {
            if (other.interfaceName != null) {
                return false;
            }
        } else if (!interfaceName.equals(other.interfaceName)) {
            return false;
        }
        if (mep == null) {
            if (other.mep != null) {
                return false;
            }
        } else if (!mep.equals(other.mep)) {
            return false;
        }
        if (operation == null) {
            if (other.operation != null) {
                return false;
            }
        } else if (!operation.equals(other.operation)) {
            return false;
        }
        if (serviceName == null) {
            if (other.serviceName != null) {
                return false;
            }
        } else if (!serviceName.equals(other.serviceName)) {
            return false;
        }
        return true;
    }

    public QName getInterfaceName() {
        return this.interfaceName;
    }

    public QName getServiceName() {
        return this.serviceName;
    }

    public QName getOperation() {
        return this.operation;
    }

    public URI getMep() {
        return this.mep;
    }
}