/**
 * Copyright (c) 2007-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.listener.AbstractExternalListener;

import com.ebmwebsourcing.easycommons.lang.StringHelper;

/**
 * This manager is used to enable a binding component to handle external service call. 
 * <p>
 * It manages listeners for external services (start and stop).
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 */
public class ExternalListenerManager {

    /**
     * A map of external listeners
     */
    private Map<String, List<AbstractExternalListener>> externalListeners;

    /**
     * The JBI component
     */
    private AbstractComponent component;

    /**
     * Creates a new instance of {@link ExternalListenerManager}
     * 
     * @param componentConfiguration
     * @param component
     * @param componentContext
     * @param deliveryChannel
     * @param logger
     */
    public ExternalListenerManager(AbstractComponent component) {
        super();
        this.component = component;
        this.externalListeners = new Hashtable<String, List<AbstractExternalListener>>();
    }

    public void createListeners(String serviceUnitName, List<Consumes> consumesList)
            throws PEtALSCDKException {
        for (Consumes consumes : consumesList) {
            AbstractExternalListener externalListener = createAndInitExternalListener(consumes);
            if (externalListener != null) {
                addExternalListerner(serviceUnitName, externalListener);
            } else {
                throw new PEtALSCDKException(
                        "No ExternalListener can be created. Check if you have "
                                + "provided an ExternalListener class "
                                + "name in your component descriptor (jbi.xml)");
            }
        }
    }

    /**
     * Start a listener for each given consumes.
     * 
     * @param serviceUnitName
     *            name of the service unit to start
     * @param consumes
     *            the consumes nodes defining external services to listen. Not
     *            null
     * @throws PEtALSCDKException
     *             if an error occurs during listener starting
     */
    public void startListening(String serviceUnitName) throws PEtALSCDKException {
        if (this.externalListeners.containsKey(serviceUnitName)) {
            for (AbstractExternalListener externalListener : this.externalListeners
                    .get(serviceUnitName)) {
                externalListener.start();
            }
        }
    }

    /**
     * Stop all external listeners related to the given SU.
     * 
     * @param serviceUnitName
     *            name of the service unit to start
     * @throws PEtALSCDKException
     *             if an error occurs during listener starting
     */
    public void stopListening(String serviceUnitName) throws PEtALSCDKException {
        if (this.externalListeners.containsKey(serviceUnitName)) {
            for (AbstractExternalListener externalListener : this.externalListeners
                    .get(serviceUnitName)) {
                externalListener.stop();
            }
        }
    }

    /**
     * Remove the SU bound listeners.
     * 
     */
    public void removeListeners(String serviceUnitName) {
        this.externalListeners.remove(serviceUnitName);
    }

    /**
     * Create and initialize an external listener for the given SU
     * configuration.
     * 
     * @param consumes
     * @return
     * @throws PEtALSCDKException
     */
    private AbstractExternalListener createAndInitExternalListener(Consumes consumes)
            throws PEtALSCDKException {
        AbstractExternalListener externalListener = null;
        final String externalListenerClassNameValue = component.getComponentConfiguration()
                .getExternalListenerClassName();
        final String externalListenerClassName = externalListenerClassNameValue.replaceAll("\\s", "");
        if (!StringHelper.isNullOrEmpty(externalListenerClassName)) {
            try {
                ClassLoader componentClassLoader = this.getClass().getClassLoader();
                Class<?> clazz = componentClassLoader.loadClass(externalListenerClassName);
                externalListener = (AbstractExternalListener) clazz.newInstance();
            } catch (InstantiationException e) {
                throw new PEtALSCDKException("Can't instanciate ExternaListener", e);
            } catch (IllegalAccessException e) {
                throw new PEtALSCDKException("Can't instanciate ExternaListener", e);
            } catch (ClassNotFoundException e) {
                throw new PEtALSCDKException("Can't find ExternaListener class : "
                        + externalListenerClassName, e);
            }
            externalListener.init(component);
            externalListener.setConsumes(consumes);
            ConfigurationExtensions extensions = new ConfigurationExtensions(consumes.getAny());
            externalListener.setExtensions(extensions);
            externalListener.init();
        }
        return externalListener;
    }

    /**
     * Add the external listener to the listener map.
     * 
     * @param serviceUnitName
     * @param externalListener
     */
    private void addExternalListerner(String serviceUnitName,
            AbstractExternalListener externalListener) {
        List<AbstractExternalListener> externalListenersForSU = this.externalListeners
                .get(serviceUnitName);
        if (externalListenersForSU == null) {
            externalListenersForSU = new ArrayList<AbstractExternalListener>();
            this.externalListeners.put(serviceUnitName, externalListenersForSU);
        }

        externalListenersForSU.add(externalListener);

    }

}
