/**
 * Copyright (c) 2006-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;
import javax.management.AttributeChangeNotification;
import javax.management.ListenerNotFoundException;
import javax.management.Notification;
import javax.management.NotificationListener;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.jbidescriptor.generated.Runtimelong;
import org.ow2.petals.component.framework.mbean.RuntimeConfigurationNotifier;
import org.ow2.petals.component.framework.monitoring.probe.MessageExchangeAcceptorPoolProbe;

/**
 * This class helps a Component to listen to {@link MessageExchange}s on the
 * DeliveryChannel (accept() method).
 * <p>
 * It starts or stops a pool of thread that listen to new incoming JBI from the
 * bus. {@link MessageExchange}s.
 * </p>
 * 
 * @author Adrien Louis - EBM WebSourcing
 * @author Olivier Fabre - EBM WebSourcing
 * @author Christophe DENEUX - EBM WebSourcing
 * @since 3.0
 */
public class JBIAcceptorManager extends AbstractMessageManager implements NotificationListener {

    /**
     * Max duration to wait when stopping an acceptor before to interrupt it, by
     * default.
     */
    protected static final long DEFAULT_ACCEPTOR_STOP_MAX_WAIT = 500;

    private final List<MessageExchangeAcceptor> acceptorPool = new ArrayList<MessageExchangeAcceptor>();

    private final JBIAcceptorThreadFactory acceptorFactory;

    private final JBIProcessorManager jbiProcessorManager;

    private final RuntimeConfigurationNotifier runtimeConfiguration;
    
    /**
     * The message exchange acceptor pool probe
     */
    final MessageExchangeAcceptorPoolProbe messageExchangeAcceptorPoolProbe;

    /**
     * Creates a new instance of {@link JBIAcceptorManager}
     * 
     * @param component
     * @param jbiProcessorManager
     * @param runtimeConfiguration
     * @param logger
     */
    public JBIAcceptorManager(final AbstractComponent component,
            final JBIProcessorManager jbiProcessorManager,
            final RuntimeConfigurationNotifier runtimeConfiguration,
            final Logger logger) {
        super(component, logger);
        this.jbiProcessorManager = jbiProcessorManager;
        this.runtimeConfiguration = runtimeConfiguration;
        this.acceptorFactory = new JBIAcceptorThreadFactory(component, this.jbiProcessorManager,
                this.logger);
        this.messageExchangeAcceptorPoolProbe = component.getMonitoringBean()
                .getMessageExchangeAcceptorPoolProbe();
        this.messageExchangeAcceptorPoolProbe.setAcceptorPool(this.acceptorPool);
    }

    private long getAcceptorStopMaxWait() {
        final Runtimelong acceptorStopMaxWait = this.component.getComponentConfiguration()
                .getAcceptorStopMaxWait();
        long acceptorStopMaxWaitValue;
        if (acceptorStopMaxWait == null) {
            acceptorStopMaxWaitValue = DEFAULT_ACCEPTOR_STOP_MAX_WAIT;
        } else {
            acceptorStopMaxWaitValue = acceptorStopMaxWait.getValue();
            if (acceptorStopMaxWaitValue <= 0) {
                acceptorStopMaxWaitValue = DEFAULT_ACCEPTOR_STOP_MAX_WAIT;
                this.logger
                        .warning("Invalid value for attribute 'acceptor-stop-max-wait': The value must be upper or equals to 1. Default value used ("
                                + DEFAULT_ACCEPTOR_STOP_MAX_WAIT + ").");
            }
        }

        return acceptorStopMaxWaitValue;
    }

    public void init() throws JBIException {
        this.runtimeConfiguration.addNotificationListener(this, null, null);
    }

    /**
     * Initialize and start threads which listen to {@link MessageExchange}s
     * 
     * @throws JBIException
     * 
     */
    public void start() throws JBIException {
        this.logger.fine("Creating JBI Acceptor pool:");

        final int corePoolSize = this.component.getComponentConfiguration().getAcceptorPoolSize()
                .getValue();
        final int retryMaxTryValue = this.getRetryMaxTry();
        final long retryBaseDurationValue = this.getRetryBaseDuration();

        this.logger.fine("\t- core pool size: " + corePoolSize);
        this.logger.fine("\t- retry policy : max try number: " + retryMaxTryValue);
        this.logger.fine("\t- retry policy : base duration: " + retryBaseDurationValue);

        for (int i = 0; i < corePoolSize; i++) {
            final MessageExchangeAcceptor acceptorThread = this.acceptorFactory.newThread(
                    retryMaxTryValue, retryBaseDurationValue);
            acceptorThread.start();
            this.acceptorPool.add(acceptorThread);
        }
        this.messageExchangeAcceptorPoolProbe.setAcceptorPoolSize(this.acceptorPool.size());
    }

    /**
     * Stop threads that listen to {@link MessageExchange}s
     * 
     */
    public void stop() {
        for (final MessageExchangeAcceptor acceptorThread : this.acceptorPool) {
            acceptorThread.interrupt();
        }
        final Iterator<MessageExchangeAcceptor> itAcceptorPool = this.acceptorPool.iterator();
        while (itAcceptorPool.hasNext()) {
            final MessageExchangeAcceptor acceptorThread = itAcceptorPool.next();
            itAcceptorPool.remove();
            this.stopAcceptorThread(acceptorThread);
        }
        this.messageExchangeAcceptorPoolProbe.setAcceptorPoolSize(this.acceptorPool.size());
    }

    public void shutdown() throws JBIException {
        try {
            this.runtimeConfiguration.removeNotificationListener(this, null, null);
        } catch (final ListenerNotFoundException e) {
            throw new JBIException("Error unregistering the runtime configuration listener", e);
        }
    }

    private void stopAcceptorThread(final MessageExchangeAcceptor acceptorThread) {
        acceptorThread.ends();
        try {
            // Wait the end of the pending task
            acceptorThread.join(this.getAcceptorStopMaxWait());

            if (!acceptorThread.isInterrupted()) {
                acceptorThread.interrupt();
            }

        } catch (InterruptedException e) {
            this.logger.log(Level.WARNING,
                    "The stop one acceptor of the acceptor thread pool was interrupted", e);
        }
    }

    /**
     * Reset the core Pool size.
     * 
     * @param size
     */
    private void setCorePoolSize(final int size) {
        if (this.acceptorPool != null) {
            this.logger.info("Reset the acceptor thread pool size to " + size);
            if (size > this.acceptorPool.size()) {
                final int nbThreadToCreate = size - this.acceptorPool.size();
                for (int i = 0; i < nbThreadToCreate; i++) {
                    final MessageExchangeAcceptor acceptorThread = this.acceptorFactory.newThread(
                            this.getRetryMaxTry(), this.getRetryBaseDuration());
                    acceptorThread.start();
                    this.acceptorPool.add(acceptorThread);
                    this.messageExchangeAcceptorPoolProbe.setAcceptorPoolSize(this.acceptorPool
                            .size());
                }
            } else if (size < this.acceptorPool.size()) {
                int nbThreadToStop = this.acceptorPool.size() - size;
                final Iterator<MessageExchangeAcceptor> itAcceptorPool = this.acceptorPool
                        .iterator();
                while (itAcceptorPool.hasNext() && nbThreadToStop > 0) {
                    final MessageExchangeAcceptor acceptorThread = itAcceptorPool.next();
                    itAcceptorPool.remove();
                    this.messageExchangeAcceptorPoolProbe.setAcceptorPoolSize(this.acceptorPool
                            .size());
                    nbThreadToStop--;
                    this.stopAcceptorThread(acceptorThread);
                }
            } else {
                this.logger.fine("Unchange value, so no acceptor thread created or stopped !");
            }
        }
    }

    ///////////////////////////////////////////////////////
    // Implementation of methods of NotificationListener //
    ///////////////////////////////////////////////////////

    public void handleNotification(final Notification notification, final Object obj) {
        if (notification instanceof AttributeChangeNotification) {
            final AttributeChangeNotification attrNotif = (AttributeChangeNotification) notification;
            final String attrName = attrNotif.getAttributeName();
            if (this.logger.isLoggable(Level.FINE)) {
                this.logger.log(Level.FINE,
                        "Notification received by " + this.getClass().getName() + " for attribute : " + attrName);
            }
            if (attrName.equals("acceptorPoolSize")) {
                final Integer newVal = (Integer) attrNotif.getNewValue();
                this.setCorePoolSize(newVal.intValue());
            }
            else {
                this.logger.log(Level.FINE,
                        "Notification ignored by " + this.getClass().getName() + " for attribute : " + attrName);
            }
        }
    }
}
