/**
 * Copyright (c) 2008-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.util.NoSuchElementException;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.management.AttributeChangeNotification;
import javax.management.ListenerNotFoundException;
import javax.management.Notification;
import javax.management.NotificationListener;

import org.apache.commons.pool.PoolUtils;
import org.apache.commons.pool.impl.GenericObjectPool;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.mbean.RuntimeConfigurationNotifier;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.monitoring.Monitoring;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.probes.CounterProbe;
import org.ow2.petals.probes.api.probes.macro.ObjectPoolProbe;
import org.ow2.petals.probes.api.probes.macro.ThreadPoolProbe;


/**
 * The JBI processor manager is in charge of JBI Message processor threads and instances.
 * @author Christophe Hamerling - EBM WebSourcing
 * @since 3.0
 */
public class JBIProcessorManager extends JBIProcessorManagerObservable implements
        NotificationListener {

    /**
     * The number of message exchange processor, by default.
     */
    public static final int DEFAULT_PROCESSOR_POOL_SIZE_MAX = 50;

    /**
     * Keep alive time (in seconds) of processor pool, by default.
     */
    protected static final int DEFAULT_PROCESSOR_POOL_SIZE_KEEP_ALIVE_TIME = 300;

    /**
     * Max duration (in milliseconds) to wait when stopping a processor before
     * to interrupt it, by default.
     */
    protected static final int DEFAULT_PROCESSOR_STOP_MAX_WAIT = 15000;

    /**
     * The message exchange processor thread pool
     */
    private ThreadPoolExecutor processorThreadpool = null;

    /**
     * The probe attached to the message exchange processor thread pool
     */
    private final ThreadPoolProbe processorThreadPoolProbe;
    
    /**
     * The message exchange processor thread pool
     */
    private GenericObjectPool processorObjectPool = null;

    /**
     * The probe attached to the message exchange processor object pool
     */
    private final ObjectPoolProbe processorObjectPoolProbe;

    /**
     * The probe counting the number of exhaustion of the message exchange
     * processor object pool
     */
    private final CounterProbe processorObjectPoolExhaustionProbe;

    private final AbstractComponent component;

    private final RuntimeConfigurationNotifier runtimeConfiguration;

    private final Logger logger;

    /**
     * Creates a new instance of {@link JBIProcessorManager}
     * 
     * @param component
     * @param runtimeConfiguration
     * @param logger
     */
    public JBIProcessorManager(final AbstractComponent component,
            final RuntimeConfigurationNotifier runtimeConfiguration,
            final Logger logger) {
        this.component = component;
        this.runtimeConfiguration = runtimeConfiguration;

        final Monitoring monitoringBean = component.getMonitoringBean();
        this.processorThreadPoolProbe = monitoringBean.getMessageExchangeProcessorThreadPoolProbe();
        this.processorObjectPoolProbe = monitoringBean.getMessageExchangeProcessorObjectPoolProbe();
        this.processorObjectPoolExhaustionProbe = monitoringBean
                .getMessageExchangeProcessorObjectPoolExhaustionProbe();
        this.logger = logger;
    }

    public void init() throws JBIException {
        this.runtimeConfiguration.addNotificationListener(this, null, null);
    }

    public void shutdown() throws JBIException {
        try {
            this.runtimeConfiguration.removeNotificationListener(this, null, null);
        } catch (final ListenerNotFoundException e) {
            throw new JBIException("Error unregistering the runtime configuration listener", e);
        }
    }

    /**
     * Start the Processor Thread Pool.
     */
    public void start() throws JBIException {
        this.logger.fine("Creating JBI Processor pool:");
        
        // Core pool size
        final int coreProcessorPoolSize = this.component.getComponentConfiguration()
                .getProcessorPoolSize().getValue();

        // Max pool size
        int maxProcessorPoolSize;
        if (this.component.getComponentConfiguration().getProcessorMaxPoolSize() != null) {
            maxProcessorPoolSize = this.component.getComponentConfiguration()
                    .getProcessorMaxPoolSize().getValue();
        }
        else {
            maxProcessorPoolSize = DEFAULT_PROCESSOR_POOL_SIZE_MAX;
        }

        // Compare core pool size to the max pool size. The max pool size MUST
        // be upper or equals to the core pool size.
        if (coreProcessorPoolSize > maxProcessorPoolSize) {
            // Incoherent values -> max pool size is increased to match the
            // value of core pool size
            final int previousMax = maxProcessorPoolSize;
            maxProcessorPoolSize = coreProcessorPoolSize;
            this.logger
                    .warning("The value of 'processor-pool-size' ("
                            + coreProcessorPoolSize
                            + ") is invalid, greater than 'processor-max-pool-size' ("
                            + previousMax
                            + "). The value of 'processor-max-pool-size' is set to the value of 'processor-pool-size'");
        }

        // Keep alive time
        int processorPoolKeepAliveTime;
        if (this.component.getComponentConfiguration().getProcessorKeepAliveTime() != null) {
            processorPoolKeepAliveTime = this.component.getComponentConfiguration()
                    .getProcessorKeepAliveTime().getValue();
            if (processorPoolKeepAliveTime <= 0) {
                // The keep alive time MUST be greater than 0
                this.logger.warning("The value of 'processor-keep-alive-time' ("
                        + processorPoolKeepAliveTime
                        + ") is invalid. The value must be strictly upper to 0.");
                processorPoolKeepAliveTime = DEFAULT_PROCESSOR_POOL_SIZE_KEEP_ALIVE_TIME;
            }
        } else {
            processorPoolKeepAliveTime = DEFAULT_PROCESSOR_POOL_SIZE_KEEP_ALIVE_TIME;
        }

        // Message processor pool max size
        int msgProcessorMaxPoolSize;
        if (this.component.getComponentConfiguration().getMessageProcessorMaxPoolSize() != null) {
            msgProcessorMaxPoolSize = this.component.getComponentConfiguration()
                    .getMessageProcessorMaxPoolSize().getValue();
            if (msgProcessorMaxPoolSize <= 0) {
                // The message processor max pool size MUST be greater than 0
                this.logger.warning("The value of 'message-processor-max-pool-size' ("
                        + msgProcessorMaxPoolSize
                        + ") is invalid. The value must be strictly upper to 0.");
                msgProcessorMaxPoolSize = maxProcessorPoolSize;
            }
        } else {
            msgProcessorMaxPoolSize = maxProcessorPoolSize;
        }

        // Compare message processor max pool size to the max size of the thread
        // pool. The message processor max pool size MUST
        // be upper or equals to the max size of the thread.
        if (msgProcessorMaxPoolSize < maxProcessorPoolSize) {
            // Incoherent values -> max pool size is increased to match the
            // value of the max size of the thread pool
            final int previousMax = msgProcessorMaxPoolSize;
            msgProcessorMaxPoolSize = maxProcessorPoolSize;
            this.logger
                    .warning("The value of 'message-processor-max-pool-size' ("
                            + previousMax
                            + ") is invalid, lower than 'processor-max-pool-size' ("
                            + maxProcessorPoolSize
                            + "). The value of 'message-processor-max-pool-size' is set to the value of 'processor-max-pool-size'");
        }

        this.logger.fine("\t- core pool size: " + coreProcessorPoolSize);
        this.logger.fine("\t- max core pool size: " + maxProcessorPoolSize);
        this.logger.fine("\t- keep alive time: " + processorPoolKeepAliveTime);
        this.logger.fine("\t- message processoir max pool size: " + msgProcessorMaxPoolSize);
        
        this.processorObjectPool = new GenericObjectPool(
                // MessageExchange processor object factory
                new MessageExchangeProcessorObjectFactory(this.component, this.logger),
                // Max pool size
                msgProcessorMaxPoolSize,
                // When exhausted
                GenericObjectPool.WHEN_EXHAUSTED_FAIL,
                // Max wait - Has no sens because of failure on exhaustion
                GenericObjectPool.DEFAULT_MAX_WAIT,
                // Max idle
                GenericObjectPool.DEFAULT_MAX_IDLE,
                // Min idle
                coreProcessorPoolSize,
                // Test on borrow
                GenericObjectPool.DEFAULT_TEST_ON_BORROW,
                // Test on return
                GenericObjectPool.DEFAULT_TEST_ON_RETURN,
                // Time between eviction runs (10min)
                600000l,
                // Number of tests per eviction run: 33% of objects are tested per run
                -33,
                // Min evictable idle time
                processorPoolKeepAliveTime,
                // Test while idle
                GenericObjectPool.DEFAULT_TEST_WHILE_IDLE);

        // initialize the pool with N objects
        try {
            PoolUtils.prefill(this.processorObjectPool, coreProcessorPoolSize);
        } catch (final IllegalArgumentException e) {
            throw new JBIException(e);
        } catch (final Exception e) {
            throw new JBIException(e);
        }

        this.processorThreadpool = new MessageExchangeProcessorThreadPoolExecutor(
                coreProcessorPoolSize, maxProcessorPoolSize, processorPoolKeepAliveTime,
                TimeUnit.SECONDS, new SynchronousQueue<Runnable>(), new JBIProcessorThreadFactory(
                        this.component, this.logger), this.processorObjectPool,
                this.processorObjectPoolProbe, this.jbiProccesorManagerObservers, this.logger);

        this.processorThreadpool.prestartAllCoreThreads();

        this.processorThreadPoolProbe.setThreadPool(this.processorThreadpool);
        this.processorObjectPoolProbe.setObjectPool(this.processorObjectPool);
    }

    /**
     * Stop the Processor Thread Pool.
     * 
     */
    public void stop() {
        // try to let the executing tasks finish
        this.processorThreadpool.shutdown();
        try {
            long processorStopMaxWait;
            if (this.component.getComponentConfiguration().getProcessorStopMaxWait() != null) {
                processorStopMaxWait = this.component.getComponentConfiguration()
                        .getProcessorStopMaxWait().getValue();
                if (processorStopMaxWait <= 0) {
                    // The duration MUST be greater than 0
                    this.logger.warning("The value of 'processor-stop-max-time' ("
                            + processorStopMaxWait
                            + ") is invalid. The value must be strictly upper to 0.");
                    processorStopMaxWait = DEFAULT_PROCESSOR_STOP_MAX_WAIT;
                }
            }
            else {
                processorStopMaxWait = DEFAULT_PROCESSOR_STOP_MAX_WAIT;
            }

            this.processorThreadpool.awaitTermination(processorStopMaxWait, TimeUnit.MILLISECONDS);
        } catch (final InterruptedException e) {
            this.logger.log(Level.SEVERE, "Workers pool interrupted : " + e.getMessage());
        }
        // interrupt the remaining executing tasks
        this.processorThreadpool.shutdownNow();

        try {
            this.processorObjectPool.close();
        } catch (final Exception e) {
            this.logger.log(Level.WARNING,
                    "Failed to close the Messsage Exchange Processor Object pool properly", e);
        }
    }

    /**
     * Get the message exchange processor thread pool.
     */
    public ThreadPoolExecutor getProcessorThreadPool() {
        return this.processorThreadpool;
    }

    /**
     * Get the message exchange processor (object-)pool.
     */
    public GenericObjectPool getProcessorObjectPool() {
        return this.processorObjectPool;
    }

    /**
     * Reset the core Pool size.
     * 
     * @param size
     */
    protected void setCorePoolSize(final int size) {
        this.logger.info("Reset the processor thread pool size to " + size);
        try {
            this.processorThreadpool.setCorePoolSize(size);
            this.processorThreadpool.prestartAllCoreThreads();

            this.processorObjectPool.setMinIdle(size);
            // Note: If the new size is bigger than the current one, we should
            // add new object instance into the pool. Because of exceptions to
            // manage, new objects will be added to the pool on demand.

        } catch (final IllegalArgumentException e) {
            String warningMessage = String
                    .format("Invalid value used (less than or equal to zero) to configure the processor thread pool core size: %d",
                            size);
            this.logger.warning(warningMessage);
        }
    }

    /**
     * Reset the max core Pool size.
     * 
     * @param size
     */
    protected void setMaxCorePoolSize(final int size) {
        this.logger.info("Reset the maximum processor thread pool size to " + size);
        try {
            this.processorThreadpool.setMaximumPoolSize(size);
            this.processorObjectPool.setMaxActive(size);
        } catch (final IllegalArgumentException e) {
            String warningMessage = String
                    .format("Invalid value used (less than or equal to zero, or less than the core pool size) to configure the processor thread pool max size: %d",
                            size);
            this.logger.warning(warningMessage);
        }
    }

    /**
     * Reset the keep alive time of the processor pool
     * 
     * @param size
     */
    protected void setPoolKeepAliveTime(final int time) {
        this.logger.info("Reset the processor thread pool keep alive time to " + time);
        try {
            this.processorThreadpool.setKeepAliveTime(time, TimeUnit.SECONDS);
            this.processorObjectPool.setMinEvictableIdleTimeMillis(time * 1000);
        } catch (final IllegalArgumentException e) {
            String warningMessage = String
                    .format("Invalid value used (less than or equal to zero) to configure the processor thread pool keep alive time: %d",
                            time);
            this.logger.warning(warningMessage);
        }
    }

    /**
     * Reset the message processor max pool size
     * 
     * @param size
     */
    protected void setMessageProcessorMaxPoolSize(final int size) {
        this.logger.info("Reset the message processor max pool size to " + size);
        if (size <= 0) {
            String warningMessage = String
                    .format("Invalid value used (less than or equal to zero) to configure the message processor max pool size: %d",
                            size);
            this.logger.warning(warningMessage);
        } else if (size < this.processorThreadpool.getMaximumPoolSize()) {
            String warningMessage = String
                    .format("Invalid value used (lower than the processor thread pool max size: %d) to configure the message processor max pool size: %d",
                            this.processorThreadpool.getMaximumPoolSize(), size);
            this.logger.warning(warningMessage);
        } else {
            this.processorObjectPool.setMaxActive(size);
        }
    }

    public void process(final ExchangeImpl exchange, final int nbMaxTry, final long baseWaitDuration)
            throws InterruptedException {
        int nbTry = 0;
        long waitDuration = 0;

        for (int i = 0; i < nbMaxTry; i++) {
            try {
                MessageExchangeProcessor messageExchangeProcessor;
                try {
                    messageExchangeProcessor = (MessageExchangeProcessor) this.processorObjectPool
                            .borrowObject();
                    try {
                        this.processorObjectPoolProbe.pick();
                    } catch (final ProbeNotStartedException e) {
                        this.logger
                                .warning("The message exchange processor object pool probe is not started. Metric values will not be valid.");
                    }
                } catch (final Exception e) {
                    if (e instanceof NoSuchElementException) {
                        // No message exchange processor is available in the object pool
                        try {
                            this.processorObjectPoolExhaustionProbe.inc();
                        } catch (final ProbeNotStartedException e1) {
                            this.logger
                                    .warning("The message exchange processor object pool exhaustion probe is not started. Metric values will not be valid.");
                        }
                        throw (NoSuchElementException) e;
                    } else {
                        if (nbTry == nbMaxTry) {
                            this.logger
                                    .log(Level.WARNING,
                                            "Try #"
                                                    + i
                                                    + ". An unknown error occurs borrowing a message exchange processor from the pool. The message exchange is lost",
                                            e);
                            return;
                        } else {
                            waitDuration = waitDuration + baseWaitDuration * (i + 1)
                                    + (long) (10 * Math.random());
                            this.logger
                                    .log(Level.WARNING,
                                            "Try #"
                                                    + i
                                                    + ". An unknown error occurs borrowing a message exchange processor from the pool. Wait "
                                                    + waitDuration + "ms before next try.", e);
                            continue;
                        }
                    }
                }
                if (messageExchangeProcessor == null) {
                    throw new NoSuchElementException(
                            "No more JBI message exchange processor is available in the pool.");
                } else {
                    messageExchangeProcessor.setMessageExchange(exchange);

                    try {
                        this.processorThreadpool.execute(messageExchangeProcessor);
                        try {
                            this.processorThreadPoolProbe.pick();
                        } catch (final ProbeNotStartedException e) {
                            this.logger
                                    .warning("The message exchange processor thread pool probe is not started. Metric values will not be valid.");
                        }
                        break;
                    } catch (final RejectedExecutionException e) {
                        messageExchangeProcessor.clean();
                        try {
                            this.processorObjectPool.returnObject(messageExchangeProcessor);
                            try {
                                this.processorObjectPoolProbe.pick();
                            } catch (final ProbeNotStartedException e1) {
                                this.logger
                                        .warning("The message exchange processor thread pool probe is not started. Metric values will not be valid.");
                            }
                        } catch (final Exception e1) {
                            this.logger
                                    .log(Level.WARNING,
                                            "A error occurs returning the message exchange processor to the pool. It is ignored.",
                                            e1);
                        }
                        throw e;
                    }
                }
            } catch (final NoSuchElementException e) {
                if (nbTry == nbMaxTry) {
                    this.logger
                            .warning("Try #"
                                    + i
                                    + ". No more JBI message exchange processor is available in the pool, and the try number is reached. Message discarded.");
                } else {
                    waitDuration = waitDuration + baseWaitDuration * (i + 1)
                            + (long) (10 * Math.random());
                    this.logger.warning("Try #" + i
                            + ". No JBI message exchange processor is available in the pool. Wait "
                            + waitDuration + "ms before next try.");
                    Thread.sleep(waitDuration);
                }
            } catch (final RejectedExecutionException e) {
                if (nbTry == nbMaxTry) {
                    this.logger
                            .warning("Try #"
                                    + i
                                    + ". No more thread is available to execute a JBI message exchange processor, and the try number is reached. Message discarded.");
                } else {
                    waitDuration = waitDuration + baseWaitDuration * (i + 1)
                            + (long) (10 * Math.random());
                    this.logger
                            .warning("Try #"
                                    + i
                                    + ". No thread is available to execute a JBI message exchange processor. Wait "
                                    + waitDuration + "ms before next try.");
                    Thread.sleep(waitDuration);
                }
            }
        }
    }

    // /////////////////////////////////////////////////////
    // Implementation of methods of NotificationListener //
    // /////////////////////////////////////////////////////

    public void handleNotification(final Notification notification, final Object obj) {
        if (notification instanceof AttributeChangeNotification) {
            final AttributeChangeNotification attrNotif = (AttributeChangeNotification) notification;
            final String attrName = attrNotif.getAttributeName();
            if (this.logger.isLoggable(Level.FINE)) {
                this.logger.log(Level.FINE, "Notification received by " + this.getClass().getName()
                        + " for attribute : " + attrName);
            }
            if (attrName.equals("processorPoolSize")) {
                final Integer newVal = (Integer) attrNotif.getNewValue();
                this.setCorePoolSize(newVal.intValue());
            } else if (attrName.equals("processorMaxPoolSize")) {
                final Integer newVal = (Integer) attrNotif.getNewValue();
                this.setMaxCorePoolSize(newVal.intValue());
            } else if (attrNotif.getAttributeName().equals("processorKeepAliveTime")) {
                final Integer newVal = (Integer) attrNotif.getNewValue();
                this.setPoolKeepAliveTime(newVal.intValue());
            } else if (attrNotif.getAttributeName().equals("messageProcessorMaxPoolSize")) {
                final Integer newVal = (Integer) attrNotif.getNewValue();
                this.setMessageProcessorMaxPoolSize(newVal.intValue());
            } else {
                this.logger.log(Level.FINE, "Notification ignored by " + this.getClass().getName()
                        + " for attribute : " + attrName);
            }
        }
    }
}
