/**
 * Copyright (c) 2008-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;

import org.ow2.petals.commons.log.Level;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.message.ExchangeImpl;


/**
 * The JBI Message Exchange Acceptor. 
 * <p>
 * Dispatch the message to a JBI worker which
 * will use the {@link AbstractJBIListener} component implementation.
 * </p>
 * 
 * @author Christophe Hamerling - EBM WebSourcing
 * @since 3.0
 */
public class MessageExchangeAcceptor extends Thread {

    /**
     * The component's logger
     */
    private final Logger logger;

    private final DeliveryChannel deliveryChannel;

    private volatile boolean running = false;

    private final Object runningLock = new Object();

    private final JBIProcessorManager jbiProcessorManager;

    /**
     * Number of tries to submit a message exchange processor for processing
     * before to declare that it cannot be processed.
     */
    private final int nbMaxTry;

    /**
     * Base duration, in milliseconds, to wait between two processing submission
     * tries.
     **/
    private final long baseWaitDuration;

    /**
     * 
     * @param component
     * @param runtimeConfiguration
     * @param jbiPool
     */
    public MessageExchangeAcceptor(AbstractComponent component,
            JBIProcessorManager jbiProcessorManager,
            Logger logger, ThreadGroup threadGroup, String threadName, final int nbMaxTry,
            final long baseWaitDuration) {
        super(threadGroup, threadName);

        this.jbiProcessorManager = jbiProcessorManager;
        this.deliveryChannel = component.getChannel();
        this.nbMaxTry = nbMaxTry;
        this.baseWaitDuration = baseWaitDuration;
        this.logger = logger;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Thread#run()
     */
    @Override
    public void run() {
        this.logger.log(Level.FINE, "Starting MessageExchangeAcceptor Thread " + this.getName());
        this.running = true;

        while (this.running) {
            try {
                // receiving a message
                final ExchangeImpl exchange = new ExchangeImpl(this.deliveryChannel.accept());

                this.logger.fine("Accepting a JBI message with Id : " + exchange.getExchangeId());
                
                this.jbiProcessorManager.process(exchange, this.nbMaxTry, this.baseWaitDuration);

            } catch (final Throwable e) {
                // search for an interrupted exception
                Throwable cause = e;
                synchronized (this.runningLock) {
                    while (cause != null && this.running) {
                        if (cause instanceof InterruptedException) {
                            this.running = false;
                            this.logger.fine("Thread '" + this.getName()
                                    + "' interrupted, stop accepting JBI messages");
                        }
                        cause = cause.getCause();
                    }
                    if (this.running) {
                        this.logger.log(Level.WARNING,
                                "An unexpected exception occurs during the processing of the JBI acceptor '"
                                        + this.getName() + "'", e);
                    }
                }
            }
        }
    }

    public void ends() {
        synchronized (this.runningLock) {
            if (this.running) {
                this.running = false;
                this.logger.fine("Thread '" + this.getName() + "' will stop to accept JBI messages.");
            }
        }
    }
}
