/**
 * Copyright (c) 2006-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import static org.ow2.petals.commons.log.PetalsExecutionContext.FLOW_INSTANCE_ID_PROPERTY_NAME;
import static org.ow2.petals.commons.log.PetalsExecutionContext.FLOW_PREVIOUS_STEP_ID_PROPERTY_NAME;
import static org.ow2.petals.commons.log.PetalsExecutionContext.FLOW_STEP_ID_PROPERTY_NAME;

import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.xml.namespace.QName;

import org.ow2.petals.commons.exception.ExceptionUtil;
import org.ow2.petals.commons.log.FlowAttributes;
import org.ow2.petals.commons.log.Level;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.Message;
import org.ow2.petals.component.framework.api.Message.InterceptorPhase;
import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.exception.SOAP11FaultServerException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.ExchangeProperties;
import org.ow2.petals.component.framework.jbidescriptor.generated.MessageParam;
import org.ow2.petals.component.framework.jbidescriptor.generated.MessageProperties;
import org.ow2.petals.component.framework.jbidescriptor.generated.Param;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.logger.ProvideFlowStepBeginLogData;
import org.ow2.petals.component.framework.logger.Utils;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.monitoring.Monitoring;
import org.ow2.petals.component.framework.monitoring.key.ServiceProviderKey;
import org.ow2.petals.component.framework.process.async.AsyncContext;
import org.ow2.petals.component.framework.process.async.AsyncMessageManager;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.component.framework.su.ServiceUnitDataHandler;
import org.ow2.petals.probes.api.enums.ExecutionStatus;
import org.ow2.petals.probes.api.exceptions.ProbeException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ResponseTimeCollectionStoppedException;
import org.ow2.petals.probes.api.exceptions.StartDateItemLostException;
import org.ow2.petals.probes.api.key.ProbeKey;
import org.ow2.petals.probes.api.probes.DurationProbe;
import org.ow2.petals.probes.api.probes.KeyedCounterProbeWithExecutionStatus;
import org.ow2.petals.probes.api.probes.KeyedResponseTimeProbe;
import org.ow2.petals.probes.api.probes.KeyedStartDateItem;
import org.ow2.petals.probes.api.probes.StartDateItem;

import com.ebmwebsourcing.easycommons.lang.StringHelper;
import com.ebmwebsourcing.easycommons.thread.ExecutionContext;
import com.ebmwebsourcing.easycommons.uuid.QualifiedUUIDGenerator;

/**
 * Processor for received MessageExchanges. 
 * <p>
 * This processor can be overridden to
 * define a specific strategy to accept messages from the DeliveryChannel. See
 * <code>acceptMessageExchange()</code>. <br>
 * Override the <code>createMessageExchangeProcessor</code> method in the
 * <code>Component</code> class to set your processor.
 * </p>
 * 
 * @author Adrien Louis - EBM WebSourcing 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Christophe Hamerling - EBM WebSourcing
 */
public class MessageExchangeProcessor implements Runnable {

    /**
     * The component logger;
     */
    private final AbstractComponent component;

    /**
     * The component logger
     */
    private final Logger logger;

    /**
     * The component delivery channel
     */
    private final DeliveryChannel deliveryChannel;

    /**
     * The component transaction manager
     */
    private final Object transactionManager;

    /**
     * The component implementation of the message processor
     */
    private final AbstractJBIListener jbiListener;

    /**
     * The component service unit manager
     */
    private final AbstractServiceUnitManager suManager;

    /**
     * The component asynchronous message manager
     */
    private final AsyncMessageManager asyncManager;

    /**
     * The component name
     */
    private final String componentName;

    /**
     * The exchange, during one message processing
     */
    private ExchangeImpl exchange = null;

    /**
     * The probe about the processing duration
     */
    private final DurationProbe mexProcessingDurationProbe;

    /**
     * The current duration item used by {@link #mexProcessingDurationProbe}.
     */
    private StartDateItem mexProcessingDurationItem;

    /**
     * The probe about response times of service providers
     */
    private final KeyedResponseTimeProbe<ServiceProviderKey> probeServiceProviderResponseTimes;

    /**
     * The pending start date items of the probe about response times of service
     * providers
     */
    private final ConcurrentHashMap<String, KeyedStartDateItem<ServiceProviderKey>> serviceProviderStartDateItems;

    /**
     * The probe counting invokations of service providers
     */
    private final KeyedCounterProbeWithExecutionStatus<ProbeKey> probeServiceProviderInvokations;

    /**
     * The pending invokation keys of the probe counting service provider
     * invokations
     */
    private final ConcurrentHashMap<String, ServiceProviderKey> serviceProviderInvokationKeys;

    /**
     * Creates a processor that listen to new {@link MessageExchange} on the
     * specified channel. Exchanges are ignored depending on the specified
     * exchange-status to ignore.
     * 
     * @param component
     * @param jbiListener
     * @param logger
     */
    public MessageExchangeProcessor(final AbstractComponent component,
            final AbstractJBIListener jbiListener, final Logger logger) {
        this.component = component;
        this.jbiListener = jbiListener;

        final Monitoring monitoringMbean = component.getMonitoringBean();
        this.mexProcessingDurationProbe = monitoringMbean
                .getMessageExchangeProcessorDurationProbe();
        this.probeServiceProviderResponseTimes = monitoringMbean
                .getServiceProviderResponseTimesProbe();
        this.serviceProviderStartDateItems = monitoringMbean.getServiceProviderStartDateItems();
        this.probeServiceProviderInvokations = monitoringMbean.getServiceProviderInvokationsProbe();
        this.serviceProviderInvokationKeys = monitoringMbean.getServiceProviderInvokationKeys();
        this.logger = logger;

        // get a direct reference to the required component attributes
        this.deliveryChannel = component.getChannel();
        this.suManager = (AbstractServiceUnitManager) component.getServiceUnitManager();
        this.asyncManager = component.getAsyncManager();
        // FIXME: Bug as the generated class of the Fractal binding to the
        // transaction
        // manager is not in the classloader of the component when this latest
        // is isolated. It is not possible to cast it in the component!
        this.transactionManager = component.getContext().getTransactionManager();
        this.componentName = component.getContext().getComponentName();
        // this.ignoreStatus =
        // component.getComponentConfiguration().getIgnoredStatus();
    }

    /**
     * Set the current exchange to process.
     * 
     * @param exchange
     */
    public void setMessageExchange(final ExchangeImpl exchange) {
        this.exchange = exchange;

    }

    /**
     * Get the current processing exchange.
     * 
     * @param exchange
     */
    public ExchangeImpl getMessageExchange() {
        return this.exchange;
    }

    /**
     * Invoke just before to execute the message exchange processor (ie.
     * invoking {@link #run()}.
     */
    public void beforeExecute() {
        try {
            this.mexProcessingDurationItem = this.mexProcessingDurationProbe.newExecution();
        } catch (final ProbeException e) {
            this.logger
                    .warning("An error occurs in the the processing duration probe of message exchange processors: "
                            + e.getMessage() + ". Metric values will not be valid.");
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Runnable#run()
     */
    @Override
    public void run() {
        try {
            this.process(this.exchange);
        } catch (final Exception e) {
            this.logger.log(Level.SEVERE, e.getMessage(), e);
            try {
                // Return an error acknowledgment only if the exchange is
                // not terminated
                if (this.exchange.isActiveStatus()) {
                    // The exception must be managed according to the
                    // MEP
                    if (MEPConstants.IN_OUT_PATTERN.equals(this.exchange.getPattern())) {
                        // a fault must be returned
                        this.exchange.setFault(e);
                    } else {
                        // For the other MEPs, an error could be
                        // returned
                        this.exchange.setError(new MessagingException(ExceptionUtil
                                .getExtendedMessage(e)));
                    }

                    this.deliveryChannel.send(this.exchange.getMessageExchange());
                }

                // test if the flow instance id was initialized
                MessageExchange messageExchange = exchange.getMessageExchange();
                FlowAttributes flowAttributes = messageExchange.getFlowAttributes();

                Utils.addMonitEndOrFailureTrace(logger, exchange, flowAttributes);
            } catch (final MessagingException me) {
                this.logger.log(Level.SEVERE,
                        "Cannot send back an error acknowledgement or a Fault", me);
            }
        } finally {
            ExecutionContext.getProperties().clear();
        }
    }

    /**
     * Invoke just after the message exchange processor execution (ie. invoking
     * {@link #run()}.
     */
    public void afterExecute() {
        try {
            this.mexProcessingDurationProbe.endsExecution(this.mexProcessingDurationItem,
                    ExecutionStatus.SUCCEEDED);
            this.mexProcessingDurationItem = null;
        } catch (final ProbeException e) {
            this.logger
                    .warning("An error occurs in the the processing duration probe of message exchange processors: "
                            + e.getMessage() + ". Metric values will not be valid.");
        }
    }

    /**
     * Clean the Message Exchange Processor. It cleans all the resources used
     * for one Message Processing.
     */
    protected final void clean() {
        // reset the JBI Listener instance
        this.jbiListener.setConsumes(null);
        this.jbiListener.setProvides(null);
        this.jbiListener.setExtensions(null);
        this.jbiListener.setCurrentTransaction(null);
        this.exchange = null;
    }

    /**
     * Launch the component processing of the given message exchange. If the
     * endpoint of the exchange is directly managed by the targeted component,<br />
     * no Provides exist: the component fully manages its kind of message.<br />
     * So the MessageInterceptorManager isn't called and no ExchangeProperties
     * are set.
     * 
     * @param exchange
     *            The message exchange to process. Must be non null
     * @throws InterruptedException
     * @throws MessagingException
     *             A problem occurred during manipulation of the exchange
     */
    private void process(final Exchange exchange) throws MessagingException {
        this.setCurrentTransaction();
        // Asynchronous expired message must be process as consumer instead of
        // as provider
        if (exchange.isProviderRole()
                && exchange.getProperty(AsyncMessageManager.ASYNC_EXPIRED_MESSAGE_PROPERTY) == null) {
            try {
                this.processAsProvider(exchange);
            } catch (RuntimeException e) {
                if (exchange.isProviderRole()) {
                    throw e;
                }
            } catch (MessagingException e) {
                if (exchange.isProviderRole()) {
                    throw e;
                }
            }
        } else {
            try {
                this.processAsConsumer(exchange);
            } catch (RuntimeException e) {
                if (exchange.isConsumerRole()) {
                    throw e;
                }
            } catch (MessagingException e) {
                if (exchange.isConsumerRole()) {
                    throw e;
                }
            }
        }
    }

    /**
     * Process the exchange as service consumer
     * 
     * @param exchange
     * @throws MessagingException
     */
    private final void processAsConsumer(final Exchange exchange) throws MessagingException {
        AsyncContext asyncContext = null;

        this.logger.finest("Process an exchange as CONSUMER with id : " + exchange.getExchangeId());
        final Consumes consumes = this.getConsumesFromExchange(exchange);
        if (consumes != null) {
            this.jbiListener.setConsumes(consumes);
            this.jbiListener.setExtensions(this.getExtensions(consumes));
            // TODO: intercept even the asynchronous exchanges?
            try {
                this.jbiListener.getMessageInterceptorProcessor().intercept(exchange, consumes,
                        InterceptorPhase.SEND_RESPONSE);
            } catch (final PEtALSCDKException e) {
                throw new MessagingException(e);
            }
        } else {
            this.logger
                    .fine("The exchange '"
                            + exchange.getExchangeId()
                            + "' refers to an unknown consumer on the component : "
                            + this.componentName
                            + ". The listener will not have the consumes and the extensions for this exchange");
        }

        final boolean isAnAsynchronousExchange;
        if (this.exchange.getProperty(AsyncMessageManager.ASYNC_MESSAGE_PROPERTY) != null) {
            // A message previously sent asynchronously is returned
            isAnAsynchronousExchange = true;
            asyncContext = this.asyncManager.unreferenceAsyncContext(this.exchange.getExchangeId());
        } else {
            isAnAsynchronousExchange = false;
        }

        if (exchange.isInOnlyPattern()) {
            this.processInOnlyAsConsumer(exchange, asyncContext);
        } else if (exchange.isRobustInOnlyPattern()) {
            this.processRobustInOnlyAsConsumer(exchange, asyncContext, isAnAsynchronousExchange);
        } else if (exchange.isInOutPattern()) {
            this.processInOutAsConsumer(exchange, asyncContext, isAnAsynchronousExchange);
        } else if (exchange.isInOptionalOutPattern()) {
            this.processInOptionalOutAsConsumer(exchange, asyncContext, isAnAsynchronousExchange);
        } else {
            throw new MessagingException("Message Exchange Pattern not recognized :"
                    + exchange.getPattern());
        }
    }

    /**
     * Process the exchange as service provider
     * 
     * @param exchange
     *            The ongoing exchange
     * @param asyncContext
     *            The asynchronous context
     * @throws MessagingException
     */
    private final void processAsProvider(final Exchange exchange) throws MessagingException {

        // Because probes are managed in methods "processXXXXAsProvider", the
        // following flag is required to manage probes on a MessagingException
        // occurring in this method.
        boolean probesToManageInThisMethod = true;
        boolean isAnActiveExchange = false;
        try {
            // necessary to test also that it is not a fault for the pattern
            // InOptionalOut
            isAnActiveExchange = exchange.isActiveStatus();
            final boolean exchangeIsActiveStatusAndIsNotAFault = isAnActiveExchange
                    && exchange.getFault() == null;
            if (exchangeIsActiveStatusAndIsNotAFault) {

                probesToManageInThisMethod = true;
                this.updateStatsOnStartOfServiceProviderInvokation(exchange,
                        this.probeServiceProviderResponseTimes, this.serviceProviderStartDateItems,
                        this.probeServiceProviderInvokations, this.serviceProviderInvokationKeys);

                // only increment step id when a new processing as provider is
                // requested ;
                // we should *NOT* increment stepid on acknowledgements or
                // errors.
                MessageExchange messageExchange = ((ExchangeImpl) exchange).getMessageExchange();
                FlowAttributes flowAttributes = messageExchange.getFlowAttributes();
                if (flowAttributes != null) {
                    PetalsExecutionContext.putFlowAttributes(flowAttributes);
                    PetalsExecutionContext.nextFlowStepId();
                    FlowAttributes newFlowAttributes = PetalsExecutionContext.getFlowAttributes();
                    messageExchange.setFlowAttributes(newFlowAttributes);
                }

                this.logger.log(Level.MONIT, "", newProvideFlowStepBeginLogData(exchange));
            }

            final Provides provides = this.getProvidesFromExchange(exchange);
            if (!isOperationAvailableForExchangeEndpoint(exchange)) {
                throw new MessagingException("Operation '" + exchange.getOperation().toString()
                        + "' not found in endpoint '" + exchange.getEndpointName()
                        + "' description");
            }
            if (provides == null) {
                if (this.component.getServiceDescription(exchange.getEndpoint()) == null) {
                    throw new MessagingException("The endpoint '" + exchange.getEndpoint()
                            + "' is unknown on the component : " + this.componentName);
                }
                this.logger.fine("Process an exchange managed directly by the component");
            } else {
                this.jbiListener.setProvides(provides);
                this.jbiListener.setExtensions(this.getExtensions(provides));
                try {
                    this.jbiListener.getMessageInterceptorProcessor().intercept(exchange, provides,
                            InterceptorPhase.ACCEPT);
                } catch (final PEtALSCDKException e) {
                    throw new MessagingException(e);
                }
            }

            // TODO: Check that CORRELATION_ID is not the same thing than
            // process instance id
            if (exchange.getProperty(Message.CORRELATION_ID_KEY) == null) {
                // add a correlation id
                final String correlationId = new QualifiedUUIDGenerator(
                        org.ow2.petals.commons.Constants.UUID_DOMAIN).getNewID();
                exchange.setProperty(Message.CORRELATION_ID_KEY, correlationId);
            }

            probesToManageInThisMethod = false;

            if (exchange.isInOnlyPattern()) {
                this.processInOnlyAsProvider(exchange, provides);
            } else if (exchange.isRobustInOnlyPattern()) {
                this.processRobustInOnlyAsProvider(exchange, provides);
            } else if (exchange.isInOutPattern()) {
                this.processInOutAsProvider(exchange, provides);
            } else if (exchange.isInOptionalOutPattern()) {
                this.processInOptionalOutAsProvider(exchange, provides);
            } else {
                probesToManageInThisMethod = true;
                throw new MessagingException("Message Exchange Pattern not recognized :"
                        + exchange.getPattern());
            }
        } catch (final MessagingException e) {

            // If the exchange was received with a not active status, probes
            // has not to be managed.
            if (probesToManageInThisMethod && isAnActiveExchange) {
                Utils.updateStatsOnEndOfServiceProviderInvokation(exchange, ExecutionStatus.ERROR,
                        this.probeServiceProviderResponseTimes, this.serviceProviderStartDateItems,
                        this.probeServiceProviderInvokations, this.serviceProviderInvokationKeys,
                        this.logger);
            }

            throw e;
        }
    }

    private boolean isOperationAvailableForExchangeEndpoint(Exchange exchange)
            throws MessagingException {
        boolean operationExists = true;
        ServiceUnitDataHandler suDataHandler = suManager.getSUDataHandlerForEndpoint(exchange
                .getEndpoint());
        // a null suDataHandler means an autogenerated ep, cant check it here
        // TODO find a best way to implement check for autogenerated ep
        if (suDataHandler != null) {
            List<QName> epOperations = suDataHandler.getEndpointOperationsList(exchange
                    .getEndpoint());
            // if operation list is empty this means that endpoint is described
            // by a fake generated wsdl, so skip operation check phase
            if (epOperations != null && !epOperations.isEmpty()) {
                operationExists = epOperations.contains(exchange.getOperation());
            }
        }
        return operationExists;
    }

    protected final ProvideFlowStepBeginLogData newProvideFlowStepBeginLogData(Exchange exchange)
            throws MessagingException {
        String edptName = "";
        QName serviceName = null;
        if (exchange.getEndpoint() != null) {
            edptName = exchange.getEndpoint().getEndpointName();
            serviceName = exchange.getEndpoint().getServiceName();
        }
        final Properties mdcProperties = ExecutionContext.getProperties();
        return new ProvideFlowStepBeginLogData(
                mdcProperties.getProperty(FLOW_INSTANCE_ID_PROPERTY_NAME),
                mdcProperties.getProperty(FLOW_STEP_ID_PROPERTY_NAME),
                StringHelper.nonNullValue(exchange.getInterfaceName()),
                StringHelper.nonNullValue(serviceName), StringHelper.nonNullValue(exchange
                        .getOperation()), edptName,
                mdcProperties.getProperty(FLOW_PREVIOUS_STEP_ID_PROPERTY_NAME),
                ((ExchangeImpl) exchange).getMessageExchange());
    }

    /**
     * Process "in-only" MessageExchange as Consumer
     * 
     * @param exchange
     * @param asyncContext
     *            The asynchronous context
     * @throws MessagingException
     */
    private final void processInOnlyAsConsumer(final Exchange exchange,
            final AsyncContext asyncContext) throws MessagingException {
        if (asyncContext == null) {
            this.jbiListener.onJBIMessage(exchange);
        } else {
            if (exchange.getProperty(AsyncMessageManager.ASYNC_EXPIRED_MESSAGE_PROPERTY) != null) {
                // the asynchronous message previously sent has expired
                this.jbiListener.onExpiredAsyncJBIMessage(
                        asyncContext.getOriginalExchange(),
                        asyncContext);
            } else {
                // the exchange is a response of the asynchronous message
                // previously sent
                this.jbiListener.onAsyncJBIMessage(exchange, asyncContext);
            }
        }
    }

    /**
     * Process "in-only" MessageExchange As Provider
     * 
     * @param exchange
     *            the MessageExchange
     * @param provides
     *            The provides
     * @throws MessagingException
     *             When errors occur during processing
     */
    private final void processInOnlyAsProvider(final Exchange exchange, final Provides provides)
            throws MessagingException {

        boolean activeStatusReceived = exchange.isActiveStatus();
        boolean faultReceived = exchange.getFault() != null;

        final boolean reply = this.jbiListener.onJBIMessage(exchange);

        if (reply && activeStatusReceived) {
            try {
                if (exchange.isActiveStatus()) {
                    exchange.setDoneStatus();
                }

                interceptAndSend(exchange, provides, faultReceived);
            } finally {
                Utils.updateStatsOnEndOfServiceProviderInvokation(exchange,
                        this.probeServiceProviderResponseTimes, this.serviceProviderStartDateItems,
                        this.probeServiceProviderInvokations, this.serviceProviderInvokationKeys,
                        this.logger);
            }
        }
    }

    /**
     * Process "in-optional-out" MessageExchange Consumer.
     * 
     * @param exchange
     * @throws MessagingException
     */
    private final void processInOptionalOutAsConsumer(final Exchange exchange,
            final AsyncContext asyncContext, final boolean isAnAsynchronousExchange) throws MessagingException {
        boolean reply = false;
        boolean doneOrErrorStatusReceived = false;
        if (exchange.isDoneStatus() || exchange.isErrorStatus()) {
            doneOrErrorStatusReceived = true;
        }
        if (asyncContext == null) {
            if (isAnAsynchronousExchange && exchange.isActiveStatus()) {
                this.logger.warning("A response message do not match any Asynchronous Context. Its bound Asynchronous Context has probably already expired!");
                exchange.setError(new Exception(
                        "A response message do not match any Asynchronous Context. Its bound Asynchronous Context has probably already expired!"));
                exchange.setProperty(AsyncMessageManager.ASYNC_EXPIRED_MESSAGE_PROPERTY, null);
                this.deliveryChannel.send(((ExchangeImpl) exchange).getMessageExchange());
            }
            else {
                reply = this.jbiListener.onJBIMessage(exchange);
            }
        } else {
            if (exchange.getProperty(AsyncMessageManager.ASYNC_EXPIRED_MESSAGE_PROPERTY) != null) {
                // the asynchronous message previously sent has expired
                reply = this.jbiListener.onExpiredAsyncJBIMessage(asyncContext.getOriginalExchange(),
                        asyncContext);
            } else {
                // the exchange is a response of the asynchronous message
                // previously sent
                reply = this.jbiListener.onAsyncJBIMessage(exchange, asyncContext);
            }
        }
        if (reply && !doneOrErrorStatusReceived) {
            if (exchange.getFault() == null && exchange.isActiveStatus()) {
                // Set Status DONE if no Fault and ACTIVE Status
                exchange.setDoneStatus();
            }
            this.deliveryChannel.send(((ExchangeImpl) exchange).getMessageExchange());
        }
    }

    /**
     * Process "in-optional-out" MessageExchange as Provider *
     * 
     * @param exchange
     *            the MessageExchange
     * @param provides
     *            The provides
     * @throws MessagingException
     *             When errors occur during processing
     */
    private final void processInOptionalOutAsProvider(final Exchange exchange,
            final Provides provides) throws MessagingException {

        boolean activeStatusReceived = exchange.isActiveStatus();
        boolean faultReceived = exchange.getFault() != null;

        final boolean reply = this.jbiListener.onJBIMessage(exchange);

        if (provides != null && exchange.isOutMessage()) {
            forwardMessageData(exchange, provides);
        }

        // Set Status DONE if no Fault or Out messages and ACTIVE Status
        if (reply & activeStatusReceived) {
            try {
                if (exchange.isActiveStatus()
                        && (faultReceived || exchange.getFault() == null
                                && !((ExchangeImpl) exchange).isOutMessage())) {
                    exchange.setDoneStatus();
                }

                interceptAndSend(exchange, provides, faultReceived);
            } finally {
                Utils.updateStatsOnEndOfServiceProviderInvokation(exchange,
                        this.probeServiceProviderResponseTimes, this.serviceProviderStartDateItems,
                        this.probeServiceProviderInvokations, this.serviceProviderInvokationKeys,
                        this.logger);
            }
        }
    }

    /**
     * Process "in-out" MessageExchange as Consumer.
     * 
     * @param exchange
     * @throws MessagingException
     */
    private final void processInOutAsConsumer(final Exchange exchange,
            final AsyncContext asyncContext, final boolean isAnAsynchronousExchange) throws MessagingException {
        boolean reply = false;
        if (asyncContext == null) {
            if (isAnAsynchronousExchange && exchange.isActiveStatus()) {
                this.logger.warning("A response message do not match any Asynchronous Context. Its bound Asynchronous Context has probably already expired!");
                exchange.setError(new Exception(
                        "A response message do not match any Asynchronous Context. Its bound Asynchronous Context has probably already expired!"));
                exchange.setProperty(AsyncMessageManager.ASYNC_EXPIRED_MESSAGE_PROPERTY, null);
                this.deliveryChannel.send(((ExchangeImpl) exchange).getMessageExchange());
            }
            else {
                reply = this.jbiListener.onJBIMessage(exchange);
            }
        } else {
            if (exchange.getProperty(AsyncMessageManager.ASYNC_EXPIRED_MESSAGE_PROPERTY) != null) {
                // the asynchronous message previously sent has expired
                reply = this.jbiListener.onExpiredAsyncJBIMessage(
                        asyncContext.getOriginalExchange(), asyncContext);
            } else {
                // the exchange is a response of the asynchronous message
                // previously sent
                reply = this.jbiListener.onAsyncJBIMessage(exchange, asyncContext);
            }
        }

        if (reply) {
            if (exchange.isActiveStatus()) {
                exchange.setDoneStatus();
            }
            this.deliveryChannel.send(((ExchangeImpl) exchange).getMessageExchange());
        }
    }

    /**
     * Process "in-out" MessageExchange as Provider.<br />
     * This pattern is special, as it is bound to the request/response mode in
     * WSDL, the provider MUST return a response, no acknowledgment is allowed
     * at this step of the pattern
     * 
     * @param exchange
     *            the MessageExchange
     * @param asyncContext
     *            The asynchronous context
     * @param provides
     *            The provides
     * @throws MessagingException
     *             When errors occur during processing
     * 
     */
    private final void processInOutAsProvider(final Exchange exchange, final Provides provides)
            throws MessagingException {
        boolean reply = false;
        boolean activeStatusReceived = exchange.isActiveStatus();
        boolean faultReceived = exchange.getFault() != null;

        try {
            reply = this.jbiListener.onJBIMessage(exchange);

            if (provides != null && exchange.isOutMessage()) {
                forwardMessageData(exchange, provides);
            }
        } catch (final Exception e) {
            this.logger.log(Level.WARNING,
                    "An unexpected exception occured during an exchange processing", e);
            // set Fault message if reply is allowed, an error has been raised
            // and the exchange is not terminated
            if (exchange.getFault() == null) {
                exchange.setFault(new SOAP11FaultServerException(e.getMessage(),
                        Constants.Exception.NAMESPACE_URI.resolve("/" + this.componentName), e));
                reply = true;
            }
        }
        if (reply && activeStatusReceived) {
            try {
                // set Fault message if no Fault or Out message
                if (exchange.getFault() == null && !exchange.isOutMessage()) {
                    // an error may have been set as in contrary of the status,
                    // it is not forbidden to set an error on the exchange in
                    // this transition of this pattern
                    if (exchange.getError() != null) {
                        exchange.setFault(new SOAP11FaultServerException(exchange.getError()
                                .getMessage(), Constants.Exception.NAMESPACE_URI.resolve("/"
                                + this.componentName), exchange.getError()));
                    } else {
                        exchange.setFault(new SOAP11FaultServerException(
                                "The Provider do not return any response",
                                Constants.Exception.NAMESPACE_URI.resolve("/" + this.componentName)));
                    }
                }

                interceptAndSend(exchange, provides, faultReceived);
            } finally {
                Utils.updateStatsOnEndOfServiceProviderInvokation(exchange,
                        this.probeServiceProviderResponseTimes, this.serviceProviderStartDateItems,
                        this.probeServiceProviderInvokations, this.serviceProviderInvokationKeys,
                        this.logger);
            }
        }
    }

    /**
     * Process "In-only" MessageExchange as consumer
     * 
     * @param exchange
     * @param asyncContext
     *            The asynchronous context
     * @throws MessagingException
     */
    private final void processRobustInOnlyAsConsumer(final Exchange exchange,
            final AsyncContext asyncContext, final boolean isAnAsynchronousExchange) throws MessagingException {

        // This can be only an acknowledgment, according to the InOnly MEP.
        if (asyncContext == null) {
            if (isAnAsynchronousExchange && exchange.isActiveStatus()) {
                this.logger.warning("A response message do not match any Asynchronous Context. Its bound Asynchronous Context has probably already expired!");
                exchange.setError(new Exception(
                        "A response message do not match any Asynchronous Context. Its bound Asynchronous Context has probably already expired!"));
                exchange.setProperty(AsyncMessageManager.ASYNC_EXPIRED_MESSAGE_PROPERTY, null);
                this.deliveryChannel.send(((ExchangeImpl) exchange).getMessageExchange());
            }
            else {
                this.jbiListener.onJBIMessage(exchange);
            }
        } else {
            if (exchange.getProperty(AsyncMessageManager.ASYNC_EXPIRED_MESSAGE_PROPERTY) != null) {
                // the asynchronous message previously sent has expired
                this.jbiListener.onExpiredAsyncJBIMessage(asyncContext.getOriginalExchange(),
                        asyncContext);
            } else {
                // the exchange is a response of the asynchronous message
                // previously sent
                this.jbiListener.onAsyncJBIMessage(exchange, asyncContext);
            }
        }
    }

    /**
     * Process "robust-in-only" MessageExchange as Provider
     * 
     * @param exchange
     *            the MessageExchange
     * @param provides
     *            The provides
     * @throws MessagingException
     *             When errors occur during processing
     * 
     */
    private final void processRobustInOnlyAsProvider(final Exchange exchange,
            final Provides provides) throws MessagingException {

        boolean activeStatusReceived = exchange.isActiveStatus();
        boolean faultReceived = exchange.getFault() != null;

        final boolean reply = this.jbiListener.onJBIMessage(exchange);

        if (reply && activeStatusReceived) {
            try {
                if (exchange.getFault() == null && exchange.isActiveStatus()) {
                    // Set Status DONE if reply is allowed, no fault and ACTIVE
                    // status
                    exchange.setDoneStatus();
                }

                interceptAndSend(exchange, provides, faultReceived);
            } finally {
                Utils.updateStatsOnEndOfServiceProviderInvokation(exchange,
                        this.probeServiceProviderResponseTimes, this.serviceProviderStartDateItems,
                        this.probeServiceProviderInvokations, this.serviceProviderInvokationKeys,
                        this.logger);
            }
        }
    }

    private final void interceptAndSend(final Exchange exchange, final Provides provides,
            final boolean faultReceived) throws MessagingException {
        if (provides != null) {
            try {
                this.jbiListener.getMessageInterceptorProcessor().intercept(exchange, provides,
                        InterceptorPhase.ACCEPT_RESPONSE);
            } catch (final PEtALSCDKException e) {
                throw new MessagingException(e);
            }
            this.setProperties(provides);
        }

        MessageExchange messageExchange = ((ExchangeImpl) exchange).getMessageExchange();

        // necessary to test also that the received message is not a fault for
        // the pattern InOptionalOut and the log must be only present if there
        // is no fault after processing the message
        if (!faultReceived) {
            // test if the flow instance id was initialized
            FlowAttributes flowAttributes = messageExchange.getFlowAttributes();

            Utils.addMonitEndOrFailureTrace(logger, exchange, flowAttributes);

            if (flowAttributes != null) {
                messageExchange.setFlowAttributes(new FlowAttributes(flowAttributes
                        .getFlowInstanceId(), ExecutionContext.getProperties().getProperty(
                        PetalsExecutionContext.FLOW_PREVIOUS_STEP_ID_PROPERTY_NAME)));
            }
        }

        this.deliveryChannel.send(messageExchange);
    }

    /**
     * Get the extensions for the given provides/consumes.
     * 
     * @param service
     * @return The extensions
     */
    private final ConfigurationExtensions getExtensions(final Object service) {
        ConfigurationExtensions extensions = null;
        if (service == null) {
            return null;
        }
        final ServiceUnitDataHandler suDataHandler = this.suManager
                .getSUDataHandlerForService(service);
        if (suDataHandler != null) {
            extensions = suDataHandler.getConfigurationExtensions(service);
        }
        return extensions;
    }

    /**
     * Get the Consumes element for the given exchange.
     * 
     * @param exchange
     * @return {@code Consumes}
     */
    private final Consumes getConsumesFromExchange(final Exchange exchange) {
        final ServiceEndpoint endpoint = exchange.getEndpoint();
        return this.suManager.getConsumesFromDestination(endpoint.getEndpointName(),
                exchange.getService(), exchange.getInterfaceName());
    }

    /**
     * Get the Provides element for the given exchange.
     * 
     * @param exchange
     * @return {@code Provides}
     */
    private final Provides getProvidesFromExchange(final Exchange exchange) {
        final ServiceEndpoint ep = exchange.getEndpoint();
        final Provides provides = this.suManager.getProvidesFromEndpoint(ep);
        return provides;
    }

    /**
     * Set the current transaction.
     */
    private final void setCurrentTransaction() {
        if (this.exchange.isTransacted()) {
            // PEtALS Container
            Transaction currentTransaction = (Transaction) this.exchange
                    .getProperty(MessageExchange.JTA_TRANSACTION_PROPERTY_NAME);

            // Potential other JBI container
            if (currentTransaction == null) {
                try {
                    // FIXME: Bug as the generated class of the Fractal binding
                    // to the transaction manager is not in the classloader of
                    // the component when this latest is isolated. It is not
                    // possible to cast it in the component!
                    currentTransaction = ((TransactionManager) this.transactionManager)
                            .getTransaction();
                } catch (final SystemException e) {
                    this.logger.log(Level.WARNING, "Failed to retrieve Transaction for exchange '"
                            + this.exchange.getExchangeId() + "'", e);
                }
            }
            this.jbiListener.setCurrentTransaction(currentTransaction);
        }
    }

    /**
     * Set the configured Exchange and Message Properties.
     * 
     * @param exchangeProperties
     * @throws MessagingException
     */
    private final void setProperties(final Provides provides) throws MessagingException {
        final ExchangeProperties exchangeProperties = provides.getExchangeProperties();
        if (exchangeProperties != null) {
            for (final Param exchangeProperty : exchangeProperties.getExchangeProperty()) {
                this.exchange.setProperty(exchangeProperty.getName(), exchangeProperty.getValue());
            }
        }

        final MessageProperties messageProperties = provides.getMessageProperties();
        if (messageProperties != null) {
            for (final MessageParam messageProperty : messageProperties.getMessageProperty()) {
                final NormalizedMessage message = this.exchange.getMessageExchange().getMessage(
                        messageProperty.getMessage().name());
                message.setProperty(messageProperty.getName(), messageProperty.getValue());
            }
        }
    }

    /**
     * Forward the data from the In to the Out message.
     * 
     * @param exchange
     * @param provides
     * @throws MessagingException
     */
    private static final void forwardMessageData(final Exchange exchange, final Provides provides)
            throws MessagingException {

        if (provides == null) {
            return;
        }

        // forward security subject
        if (provides.isForwardSecuritySubject() != null
                && provides.isForwardSecuritySubject().booleanValue()) {
            if (exchange.getInSubject() != null && exchange.getOutSubject() == null) {
                exchange.setOutSubject(exchange.getInSubject());
            }
        }

        // forward message properties
        if (provides.isForwardMessageProperties() != null
                && provides.isForwardMessageProperties().booleanValue()) {
            for (final String propertyName : exchange.getInMessagePropertyNames()) {
                if (exchange.getOutMessageProperty(propertyName) == null) {
                    exchange.setOutMessageProperty(propertyName,
                            exchange.getInMessageProperty(propertyName));
                }
            }
        }

        // forward message attachment
        if (provides.isForwardAttachments() != null
                && provides.isForwardAttachments().booleanValue()) {
            final Set<String> attachmentNames = exchange.getInMessageAttachmentNames();
            for (final String attachmentName : attachmentNames) {
                if (exchange.getOutMessageAttachment(attachmentName) == null) {
                    exchange.setOutMessageAttachment(attachmentName,
                            exchange.getInMessageAttachment(attachmentName));
                }
            }
        }
    }
    


    /**
     * Update monitoring probes when the service provider invokation starts.
     * 
     * @param exchange
     *            The message exchange associated to the service provider
     *            invokation.
     * @param probeServiceProviderResponseTimes
     *            The probe about response times of service providers
     * @param serviceProviderStartDateItems
     *            The pending start date items of the probe about response times
     *            of service providers
     * @param probeServiceProviderInvokations
     *            The probe counting invokations of service providers
     * @param serviceProviderInvokationKeys
     *            The pending invokation keys of the probe counting service
     *            provider invokations
     * @throws MessagingException
     */
    public final void updateStatsOnStartOfServiceProviderInvokation(
            final Exchange exchange,
            final KeyedResponseTimeProbe<ServiceProviderKey> probeServiceProviderResponseTimes,
            final ConcurrentHashMap<String, KeyedStartDateItem<ServiceProviderKey>> serviceProviderStartDateItems,
            final KeyedCounterProbeWithExecutionStatus<ProbeKey> probeServiceProviderInvokations,
            final ConcurrentHashMap<String, ServiceProviderKey> serviceProviderInvokationKeys)
            throws MessagingException {

        final String exchangeId = exchange.getExchangeId();
        this.logger.log(Level.FINEST, "Process an exchange as PROVIDER with id : " + exchangeId);

        final ServiceProviderKey serviceProviderKey = new ServiceProviderKey(
                exchange.getInterfaceName(), exchange.getService(), exchange.getOperation(),
                exchange.getExchangePattern());
        try {
            this.serviceProviderInvokationKeys.put(exchangeId, serviceProviderKey);
            this.probeServiceProviderInvokations.incPending(serviceProviderKey);
        } catch (final ProbeNotStartedException e1) {
            this.logger
                    .warning("The probe counting service provider invokations is not started. Metric values can be uncorrect.");
        }
        try {
            this.serviceProviderStartDateItems.put(exchangeId,
                    this.probeServiceProviderResponseTimes.newExecution(serviceProviderKey));
        } catch (final StartDateItemLostException e2) {
            this.logger
                    .warning("Unable to add a new start date item into the probe about response times of service providers. Metric values can be uncorrect.");
        } catch (final ProbeNotStartedException e2) {
            this.logger
                    .warning("The probe about response times of service providers is not started. Metric values can be uncorrect.");
        } catch (final ResponseTimeCollectionStoppedException e2) {
            this.logger
                    .warning("The sample collecting of the probe about response times of service providers is stopped. Metric values can be uncorrect.");
        }
    }

}
