/**
 * Copyright (c) 2008-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.pool.impl.GenericObjectPool;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.probes.macro.ObjectPoolProbe;

/**
 * JBI Message processor pool
 * 
 * @author Christophe DENEUX (EBMWebsourcing)
 * 
 */
class MessageExchangeProcessorThreadPoolExecutor extends ThreadPoolExecutor {

    /**
     * The message exchange processor object pool
     */
    private final GenericObjectPool processorObjectPool;

    /**
     * The probe attached to the message exchange processor object pool
     */
    private final ObjectPoolProbe processorObjectPoolProbe;

    private final Logger logger;

    private final List<JBIProcessorManagerObserver> jbiProccesorManagerObservers;

    public MessageExchangeProcessorThreadPoolExecutor(final int corePoolSize,
            final int maximumPoolSize, final long keepAliveTime, final TimeUnit unit,
            final BlockingQueue<Runnable> workQueue, final ThreadFactory threadFactory,
            final GenericObjectPool processorObjectPool,
            final ObjectPoolProbe processorObjectPoolProbe,
            final List<JBIProcessorManagerObserver> jbiProccesorManagerObservers,
            final Logger logger) {
        super(corePoolSize, maximumPoolSize, keepAliveTime, unit, workQueue, threadFactory,
                new ThreadPoolExecutor.AbortPolicy());
        this.processorObjectPool = processorObjectPool;
        this.processorObjectPoolProbe = processorObjectPoolProbe;
        this.jbiProccesorManagerObservers = jbiProccesorManagerObservers;
        this.logger = logger;
    }

    @Override
    protected void beforeExecute(Thread t, Runnable r) {

        super.beforeExecute(t, r);

        if (r instanceof MessageExchangeProcessor) {
            final MessageExchangeProcessor mexProcessor = (MessageExchangeProcessor) r;

            mexProcessor.beforeExecute();

            final ExchangeImpl exchange = mexProcessor.getMessageExchange();
            this.notifyObserversBeforeProcessing(exchange);
        }
    }

    @Override
    protected void afterExecute(final Runnable r, final Throwable t) {

        if (r instanceof MessageExchangeProcessor) {
            final MessageExchangeProcessor mexProcessor = (MessageExchangeProcessor) r;

            // Clean the Message Exchange Processor
            final ExchangeImpl exchange = mexProcessor.getMessageExchange();
            this.notifyObserversAfterProcessing(exchange);

            mexProcessor.afterExecute();
            mexProcessor.clean();
        }

        try {
            this.processorObjectPool.returnObject(r);
            try {
                this.processorObjectPoolProbe.pick();
            } catch (final ProbeNotStartedException e1) {
                this.logger
                        .warning("The message exchange processor thread pool probe is not started. Metric values will not be valid.");
            }
        } catch (final Exception e) {
            this.logger
                    .log(Level.SEVERE,
                            "Failed to return a Message Exchange Processor instance to the Object pool",
                            e);
        }
        super.afterExecute(r, t);
    }

    private final void notifyObserversBeforeProcessing(ExchangeImpl exchange) {
        synchronized (this.jbiProccesorManagerObservers) {
            for (JBIProcessorManagerObserver jbiProccesorManagerObserver : jbiProccesorManagerObservers) {
                jbiProccesorManagerObserver.beforeProcessingMessage(exchange);
            }
        }
    }

    private final void notifyObserversAfterProcessing(ExchangeImpl exchange) {
        synchronized (this.jbiProccesorManagerObservers) {
            for (JBIProcessorManagerObserver jbiProccesorManagerObserver : jbiProccesorManagerObservers) {
                jbiProccesorManagerObserver.afterProcessingMessage(exchange);
            }
        }
    }

}