/**
 * Copyright (c) 2008-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process.async;

import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Settablepositivestrictlong;
import org.ow2.petals.component.framework.process.AbstractMessageManager;
import org.ow2.petals.component.framework.process.JBIProcessorManager;



/**
 * An asynchronous message Manager.
 * @author Christophe Hamerling - EBM WebSourcing
 * @since 3.0
 */
public class AsyncMessageManager extends AbstractMessageManager {

    /**
     * Default value of the time between async message cleaner runs, in
     * milliseconds.
     */
    protected static final int DEFAULT_TIME_BETWEEN_ASYNC_CLEANER_RUNS = 2000;

    /**
     * Name of the property set on a message exchange to define it as sent
     * asynchronously.
     */
    public static final String ASYNC_MESSAGE_PROPERTY = "org.ow2.petals.component.framework.asynchronous";

    /**
     * Name of the property set on an original message exchange to be able to
     * process expired messages sent asynchronously. The value of this property
     * is the message exchange id of the message sent.
     */
    public static final String ASYNC_EXPIRED_MESSAGE_PROPERTY = "org.ow2.petals.component.framework.asynchronous.expired";

    private final JBIProcessorManager jbiProcessorManager;

    /**
     * Map of contexts and message exchanges sent asynchronously, indexed by
     * identifier of message exchange sent.
     */
    private final ConcurrentMap<String, AsyncContextExchange> asynchronousContexts;
    
    /**
     * The timer needed to run cleaning task.
     */
    private final Timer timer;

    /**
     * {@link TimerTask} used to detect expired messages.
     */
    private TimerTask cleanerTask;

    /**
     * 
     * @param component
     * @param logger
     * @param jbiProcessorManager
     */
    public AsyncMessageManager(final AbstractComponent component,
            final JBIProcessorManager jbiProcessorManager, final Logger logger) {
        super(component, logger);
        this.asynchronousContexts = new ConcurrentHashMap<String, AsyncContextExchange>();
        this.jbiProcessorManager = jbiProcessorManager;
        this.timer = new Timer(this.component.getContext().getComponentName()
                + " - Asynchonous message exchange cleaner");
    }

    /**
     * Start the Async Message Manager.
     */
    public void start() {
        this.cleanerTask = new AsyncContextCleaner(this.asynchronousContexts,
                this.jbiProcessorManager, this.getRetryMaxTry(),
                this.getRetryBaseDuration(), this.logger);
        this.timer.schedule(this.cleanerTask, 0, this.getTimeBeetweenAsyncCleanerRuns());
    }

    /**
     * Stop the Async Message Manager.
     */
    public void stop() {
        this.cleanerTask.cancel();
    }

    /**
     * Shutdown the Async Message Manager.
     */
    public void shutdown() {
        this.timer.cancel();
    }

    private final long getTimeBeetweenAsyncCleanerRuns() {
        final Settablepositivestrictlong timeBeetweenAsyncCleanerRuns = this.component
                .getComponentConfiguration().getTimeBeetweenAsyncCleanerRuns();
        long timeBeetweenAsyncCleanerRunsValue;
        if (timeBeetweenAsyncCleanerRuns == null) {
            timeBeetweenAsyncCleanerRunsValue = DEFAULT_TIME_BETWEEN_ASYNC_CLEANER_RUNS;
        } else {
            timeBeetweenAsyncCleanerRunsValue = timeBeetweenAsyncCleanerRuns.getValue();
            if (timeBeetweenAsyncCleanerRunsValue <= 0) {
                timeBeetweenAsyncCleanerRunsValue = DEFAULT_TIME_BETWEEN_ASYNC_CLEANER_RUNS;
                this.logger
                        .warning("Invalid value for attribute 'time-beetween-async-cleaner-runs': The value must be upper or equals to 1. Default value used ("
                                + DEFAULT_TIME_BETWEEN_ASYNC_CLEANER_RUNS + ").");
            }
        }
        return timeBeetweenAsyncCleanerRunsValue;
    }

    /**
     * <p>
     * Reference an message as an asynchronous message:
     * <ul>
     * <li>adding the context to the asynchronous contexts map,</li>
     * <li>setting the expiration time of the message.</li>
     * </ul>
     * </p>
     * 
     * @param exchange
     *            The exchange sent asynchronously
     * @param asyncContext
     *            The asynchronous context associated to the exchange.
     */
    public void referenceAsyncContext(final Exchange exchange, final AsyncContext asyncContext) {
        if (this.logger.isLoggable(Level.FINE)) {
            this.logger.fine("Add AsyncContext for exchange id: " + exchange.getExchangeId());
        }
        exchange.setProperty(ASYNC_MESSAGE_PROPERTY, Boolean.toString(true));
        asyncContext.setExpirationTime();
        this.asynchronousContexts.put(exchange.getExchangeId(), new AsyncContextExchange(asyncContext, exchange));
    }

    /**
     * <p>
     * Unreference an asynchronous context:
     * <ul>
     * <li>removing the context from the asynchronous messages map.</li>
     * </ul>
     * </p>
     * 
     * @param exchangeId
     *            Unique identifier of the asynchronous message to unreference
     * @return null if the asynchronous context is not referenced (not found in
     *         the map)
     */
    public AsyncContext unreferenceAsyncContext(final String exchangeId) {
        final AsyncContextExchange asyncContextExchange = this.asynchronousContexts.remove(exchangeId);
        if (asyncContextExchange != null) {
            return asyncContextExchange.asyncContext;
        }
        else {
            return null;
        }
    }
}
