/**
 * Copyright (c) 2005-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.su;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.management.DeploymentException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.stream.StreamSource;

import org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.Description;
import org.ow2.easywsdl.wsdl.api.Endpoint;
import org.ow2.easywsdl.wsdl.api.Service;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.petals.commons.exception.ExceptionUtil;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.util.PropertiesUtil;
import org.ow2.petals.component.framework.interceptor.MessageInterceptorManager;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorException;
import org.ow2.petals.component.framework.jbidescriptor.JBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.jbidescriptor.generated.SUInterceptor;
import org.ow2.petals.component.framework.jbidescriptor.generated.SUInterceptors;
import org.ow2.petals.component.framework.util.EndpointUtil;
import org.ow2.petals.component.framework.util.ManagementMessageUtil;
import org.ow2.petals.component.framework.util.ServiceEndpointKey;
import org.ow2.petals.component.framework.util.SourceUtil;
import org.ow2.petals.component.framework.util.WSDLUtilImpl;
import org.w3c.dom.Document;

import com.ebmwebsourcing.easycommons.lang.StringHelper;
import com.ebmwebsourcing.easycommons.properties.PropertiesException;

/**
 * The common part of the ServiceUnitManager. 
 * <p>
 * It handles service units JBI descriptor during their deployment, life-cycle and undeployment.
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Adrien Louis - EBM WebSourcing
 */
public abstract class AbstractServiceUnitManager implements javax.jbi.component.ServiceUnitManager {

    /**
     * The reference to the component
     */
    protected final AbstractComponent component;

    /**
     * The component's Logger
     */
    protected Logger logger;

    /**
     * The service unit data handlers. Each service unit is linked to a
     * datahandler. Key is the service unit name.
     * 
     * @see {@link getServiceUnitDataHandlers()}
     * 
     */
    private final ConcurrentMap<String, ServiceUnitDataHandler> serviceUnitDataHandlers;

    /**
     * Creates a new instance of {@link AbstractServiceUnitManager}
     * 
     * @param component
     */
    public AbstractServiceUnitManager(final AbstractComponent component) {
        super();
        this.component = component;
        this.logger = component.getLogger();
        this.serviceUnitDataHandlers = new ConcurrentHashMap<String, ServiceUnitDataHandler>();
    }

    /**
     * Search the wsdl description of the activated endpoint in the Provides
     * extensions. If no wsdl description can be found, it creates a light
     * weight Wsdl description from Provides Interface, Service and Endpoint
     * names. No generation of endpoint is processed to know after the generated
     * name to update the provides too.
     * 
     * 
     * @param wsdlLocation
     *            the location of the wsdl document. Can be {@code null}
     * @param provides
     *            the Provides node
     * @return the WSDL description
     * @throws PEtALSCDKException
     */
    private static final Description getWsdlDescription(final URL wsdlLocation,
            final Provides provides) throws PEtALSCDKException {
        if (wsdlLocation != null) {
            try {
                return WSDLUtilImpl.createWsdlDescription(wsdlLocation, false);
            } catch (final WSDLException e) {
                throw new PEtALSCDKException(e);
            }
        } else {
            // if the WSDL field is null, generate a basic WSDL
            try {
                return WSDLUtilImpl.createLightWSDL20Description(provides.getInterfaceName(),
                        provides.getServiceName(), provides.getEndpointName());
            } catch (final WSDLException e) {
                throw new PEtALSCDKException("Can not auto-generate a WSDL document", e);
            }
        }
    }

    /**
     * Get the location of the wsdl file.
     * 
     * @param installRoot
     *            the root install used to search locally the file if needed
     * @param provides
     *            the provides read to get the wsdl file name
     * @return the valid URL of the Wsdl file, or null
     * @throws PEtALSCDKException
     */
    private static final URL getWsdlLocation(final String installRoot, final Provides provides)
            throws PEtALSCDKException {
        String wsdlFileLocation = null;
        URL wsdlLocation = null;

        wsdlFileLocation = provides.getWsdl();
        if (StringHelper.isNullOrEmpty(provides.getWsdl())) {
            return null;
        }
        try {
            wsdlLocation = new URL(provides.getWsdl());
        } catch (final MalformedURLException e) {
            final File wsdlFile = new File(installRoot, wsdlFileLocation);
            if (!wsdlFile.exists()) {
                throw new PEtALSCDKException("Can not retrieve WSDL document from location : "
                        + wsdlFileLocation);
            }
            try {
                wsdlLocation = wsdlFile.toURI().toURL();
            } catch (final MalformedURLException e1) {
                // not possible
            }
        }
        return wsdlLocation;
    }

    /**
     * Validate the WSDL description; Check that it match with the endpointName,
     * serviceName, and interfaceName of the {@code Provides} section.
     * 
     * @param wsdlDescription
     * @param provides
     * @throws JBIException
     */
    private static final void validateDescription(final Description wsdlDescription,
            final Provides provides) throws JBIException {

        try {
            final Service ss = wsdlDescription.getService(provides.getServiceName());
            if (ss == null) {
                throw new JBIException("Failed to find provided service '"
                        + provides.getServiceName() + "' in the WSDL description");
            }

            final Endpoint e = ss.getEndpoint(provides.getEndpointName());
            if (e == null) {
                throw new JBIException("Failed to find provided endpoint '"
                        + provides.getEndpointName() + "' in the WSDL description");
            }

            if (ss.getInterface() != null && ss.getInterface().getQName() != null) {
                if (!ss.getInterface().getQName().equals(provides.getInterfaceName())) {
                    throw new JBIException("Failed to match provided interface '"
                            + provides.getInterfaceName() + "'. Found '"
                            + ss.getInterface().getQName() + "' in the WSDL description");
                }
            } else {
                throw new JBIException(
                        "Failed to find an interface in WSDL description bound the provided service "
                                + ss.getQName());
            }
        } catch (final WSDLException e) {
            throw new JBIException(
                    "Failed to parse the WSDL description bound to the provided service '"
                            + provides.getEndpointName() + "'", e);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#deploy(java.lang.String,
     * java.lang.String)
     */
    @Override
    public String deploy(final String serviceUnitName, final String suRootPath)
            throws DeploymentException {
        this.logger.log(Level.INFO, "Deploy Service Unit '" + serviceUnitName + "'");

        if (this.serviceUnitDataHandlers.containsKey(serviceUnitName)) {
            this.logger.log(Level.SEVERE, "Service Unit '" + serviceUnitName + "' not deployed");
            return ManagementMessageUtil.getComponentTaskResult(this.component.getContext()
                    .getComponentName(), "deploy", ManagementMessageUtil.TASK_RESULT_FAILED, "1",
                    "A service unit with the name {1} is already deployed",
                    new String[] { serviceUnitName });
        }

        final File jbiXmlFile = new File(suRootPath, Constants.Component.META_INF + File.separator
                + Constants.Component.JBI_XML);

        Exception exception = null;
        Jbi jbiDescriptor = null;
        InputStream jbiDescriptorInputStream = null;
        try {
            final String jbiXmlString = this.processProperties(jbiXmlFile);
            final StringReader jbiXmlStringReader = new StringReader(jbiXmlString);
            jbiDescriptor = JBIDescriptorBuilder.buildJavaJBIDescriptor(jbiXmlStringReader);
        } catch (final CDKJBIDescriptorException e) {
            exception = e;
        } catch (final PEtALSCDKException e) {
            exception = e;
        } catch (final IOException e) {
            exception = e;
        } finally {
            try {
                if (jbiDescriptorInputStream != null) {
                    jbiDescriptorInputStream.close();
                }
            } catch (IOException ex) {
                // do nothing
            }
        }

        if (exception == null) {
            final ServiceUnitDataHandler handler = new ServiceUnitDataHandler(serviceUnitName,
                    suRootPath, jbiDescriptor);

            // register the datahandler here, because of the call of the method
            // 'getServiceDescription' when activating an endpoint
            this.logger.log(Level.FINE, "Register a data handler for SU '" + serviceUnitName + "'");
            this.serviceUnitDataHandlers.put(serviceUnitName, handler);

            try {
                this.processConsumesNodes(handler);
                this.processProvidesNodes(handler);
                this.doDeploy(serviceUnitName, suRootPath, jbiDescriptor);
            } catch (final Exception e) {
                this.logger.log(Level.FINE, "Unregister the data handler for SU : "
                        + serviceUnitName);
                for (ServiceEndpoint endpoint : handler.getServiceEndpoints().values()) {
                    try {
                        this.component.getContext().deactivateEndpoint(endpoint);
                    } catch (final JBIException e1) {
                        this.logger.log(Level.WARNING,
                                "Error during deployment cleanup of Service unit '"
                                        + serviceUnitName + "'", e);
                    }
                }
                this.serviceUnitDataHandlers.remove(serviceUnitName);
                exception = e;
            }
        }

        String result;
        if (exception == null) {
            result = ManagementMessageUtil.getComponentTaskResult(this.component.getContext()
                    .getComponentName(), "deploy", ManagementMessageUtil.TASK_RESULT_SUCCESS);
            this.logger.log(Level.INFO, "Service Unit '" + serviceUnitName + "' deployed");
        } else {
            // instead of throwing an exception, return an error message
            result = ManagementMessageUtil.getComponentTaskResult(this.component.getContext()
                    .getComponentName(), "deploy", ManagementMessageUtil.TASK_RESULT_FAILED,
                    ManagementMessageUtil.MESSAGE_TYPE_ERROR, "1",
                    "Failed to deploy Service Unit : {1}", new String[] { serviceUnitName },
                    ExceptionUtil.getExtendedMessage(exception));
            this.logger.log(Level.SEVERE,
                    "Service Unit '" + serviceUnitName + "' failed to deploy", exception);
        }

        return result;
    }

    /**
     * Get the consumes matching the given endpoint name.
     * 
     * @param serviceName
     *            The endpoint name to match
     * @return The {@code Consumes} instance if matched, @ null} otherwise
     */
    public Consumes getConsumesForEndpoint(final String endpointName) {
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            for (final Consumes consumes : dataHandler.getDescriptor().getServices().getConsumes()) {
                if (endpointName.equals(consumes.getEndpointName())) {
                    return consumes;
                }
            }
        }
        return null;
    }

    /**
     * Get the consumes matching the given service name.
     * 
     * @param serviceName
     *            The service qualified name to match
     * @return The {@code Consumes} instance if matched, @ null} otherwise
     */
    public Consumes getConsumesForInterface(final QName interfaceName) {
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            for (final Consumes consumes : dataHandler.getDescriptor().getServices().getConsumes()) {
                if (interfaceName.equals(consumes.getInterfaceName())) {
                    return consumes;
                }
            }
        }
        return null;
    }

    /**
     * Get the consumes matching the given service name.
     * 
     * @param serviceName
     *            The service qualified name to match
     * @return The {@code Consumes} instance if matched, @ null} otherwise
     */
    public Consumes getConsumesForService(final QName serviceName) {
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            for (final Consumes consumes : dataHandler.getDescriptor().getServices().getConsumes()) {
                if (serviceName.equals(consumes.getServiceName())) {
                    return consumes;
                }
            }
        }
        return null;
    }

    /**
     * Retrieve the Consumes node matching the given destination information
     * (endpoint name, service name, interface name). Search order is by
     * ascending priority : endpointName, serviceName, interfaceName. If there
     * are many consumes matching information, return the first one found.
     * 
     * @param endpointName
     *            an endpoint name, can be null
     * @param serviceName
     *            a service {@link QName}, can be null
     * @param interfaceName
     *            an interface {@link QName}, can be null
     * @return the {@link Consumes} element matching information, can be null
     */
    public Consumes getConsumesFromDestination(final String endpointName, final QName serviceName,
            final QName interfaceName) {
        Consumes result = null;
        if (endpointName != null) {
            result = this.getConsumesForEndpoint(endpointName);
        }
        if (result == null) {
            if (serviceName != null) {
                result = this.getConsumesForService(serviceName);
            }
            if (result == null) {
                if (interfaceName != null) {
                    result = this.getConsumesForInterface(interfaceName);
                }
            }
        }
        return result;
    }

    /**
     * Get the {@code set} of endpoints exposed by the given Service Unit.
     * 
     * @param serviceUnitName
     *            The name of the Service Unit
     * @return a {@code set} of {@code ServiceEndpoint}
     */
    public Set<ServiceEndpoint> getEndpointsForServiceUnit(final String serviceUnitName) {
        final Set<ServiceEndpoint> result = new HashSet<ServiceEndpoint>();
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (final ServiceEndpointKey key : this.serviceUnitDataHandlers.get(serviceUnitName)
                .getServiceEndpoints().keySet()) {
            result.add(this.serviceUnitDataHandlers.get(serviceUnitName).getServiceEndpoints()
                    .get(key));
        }
        return result;
    }

    /**
     * Retrieve the Provides node that has activated the given jbi Endpoint
     * 
     * @param serviceEndpoint
     *            the previously activated jbi Endpoint
     * @return the {@link Provides} element
     */
    public Provides getProvidesFromEndpoint(final ServiceEndpoint serviceEndpoint) {
        if (serviceEndpoint != null) {
            final ServiceEndpointKey serviceEndpointKey = new ServiceEndpointKey(
                    serviceEndpoint.getServiceName(), serviceEndpoint.getEndpointName());
            // No synchronization on serviceUnitDataHandlers is needed because
            // it's a ConcurrentMap
            for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
                final Provides provides = dataHandler.getEpJBIDesc().get(serviceEndpointKey);
                if (provides != null) {
                    return provides;
                }
            }
        }
        return null;
    }

    /**
     * Retrieve the Service description (as DOM Document) of the given jbi
     * Endpoint
     * 
     * @param serviceEndpoint
     *            the jbi Endpoint
     * @return the service description as Dom Document
     */
    public Document getServiceDescription(final ServiceEndpoint serviceEndpoint) {
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            final Document serviceDescription = dataHandler.getEndpointDescription(serviceEndpoint);
            if (serviceDescription != null) {
                return serviceDescription;
            }
        }
        return null;
    }

    /**
     * Return all {@link ServiceUnitDataHandler}s as a Map
     * 
     * @return all {@link ServiceUnitDataHandler}s as a Map
     */
    public Map<String, ServiceUnitDataHandler> getServiceUnitDataHandlers() {
        return this.serviceUnitDataHandlers;
    }

    /**
     * Get the SU Data Handler matching the given Provides.
     * 
     * @param consumes
     *            The @ code Consumes} instance
     * @return
     */
    public ServiceUnitDataHandler getSUDataHandlerForConsumes(final Consumes consumes) {
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            for (final Consumes consumesElement : dataHandler.getDescriptor().getServices()
                    .getConsumes()) {
                if (consumes.equals(consumesElement)) {
                    return dataHandler;
                }
            }
        }
        return null;
    }

    /**
     * Retrieve the data handler of the Service Unit that has activated the
     * given endpoint.
     * 
     * @param endpoint
     *            the
     * @return the {@link ServiceUnitDataHandler}
     */
    public ServiceUnitDataHandler getSUDataHandlerForEndpoint(final ServiceEndpoint endpoint) {
        final ServiceEndpointKey key = new ServiceEndpointKey(endpoint.getServiceName(),
                endpoint.getEndpointName());
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            for (final ServiceEndpointKey k : dataHandler.getEpJBIDesc().keySet()) {
                if (key.equals(k)) {
                    return dataHandler;
                }
            }
        }

        return null;
    }

    /**
     * Get the SU Data Handler matching the given Consumes.
     * 
     * @param provides
     *            The @ code Provides} instance
     * @return
     */
    public ServiceUnitDataHandler getSUDataHandlerForProvides(final Provides provides) {
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            for (final Provides providesElement : dataHandler.getDescriptor().getServices()
                    .getProvides()) {
                if (provides.equals(providesElement)) {
                    return dataHandler;
                }
            }
        }
        return null;
    }

    /**
     * Get the SU Data Handler matching the given Consumes/Provides service.
     * 
     * @param consumes
     *            The @ code Consumes} instance
     * @return
     */
    public ServiceUnitDataHandler getSUDataHandlerForService(final Object service) {
        ServiceUnitDataHandler suDataHandler = null;

        if (service instanceof Consumes) {
            suDataHandler = this.getSUDataHandlerForConsumes((Consumes) service);
        } else if (service instanceof Provides) {
            suDataHandler = this.getSUDataHandlerForProvides((Provides) service);
        }

        return suDataHandler;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#init(java.lang.String,
     * java.lang.String)
     */
    @Override
    public void init(final String serviceUnitName, final String suRootPath)
            throws DeploymentException {
        this.logger.log(Level.INFO, "Init Service Unit '" + serviceUnitName + "'");

        // Call specific processing on stop phase
        try {
            this.doInit(serviceUnitName, suRootPath);
        } catch (final PEtALSCDKException e) {
            throw new DeploymentException("Failed to init the Service Unit '" + serviceUnitName
                    + "'", e);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#shutDown(java.lang.String)
     */
    @Override
    public void shutDown(final String serviceUnitName) throws DeploymentException {
        this.logger.log(Level.INFO, "Shut down Service Unit '" + serviceUnitName + "'");

        try {
            // Call specific processing on shutting down phase
            this.doShutdown(serviceUnitName);
        } catch (final PEtALSCDKException e) {
            throw new DeploymentException("Failed to shut down the Service Unit '"
                    + serviceUnitName + "'", e);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#start(java.lang.String)
     */
    @Override
    public void start(final String serviceUnitName) throws DeploymentException {
        this.logger.log(Level.INFO, "Start Service Unit '" + serviceUnitName + "'");

        // Call specific processing on starting phase
        try {
            this.doStart(serviceUnitName);
        } catch (final PEtALSCDKException e) {
            throw new DeploymentException("Failed to start the Service Unit '" + serviceUnitName
                    + "'", e);
        }

    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#stop(java.lang.String)
     */
    @Override
    public void stop(final String serviceUnitName) throws DeploymentException {
        this.logger.log(Level.INFO, "Stop Service Unit '" + serviceUnitName + "'");

        // Call specific processing on stop phase
        try {
            this.doStop(serviceUnitName);
        } catch (final PEtALSCDKException e) {
            throw new DeploymentException("Failed to stop the Service Unit '" + serviceUnitName
                    + "'", e);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#undeploy(java.lang.String,
     * java.lang.String)
     */
    @Override
    public String undeploy(final String serviceUnitName, final String suRootPath)
            throws DeploymentException {
        this.logger.log(Level.INFO, "Undeploy Service Unit " + serviceUnitName);
        Exception exception = null;

        final ServiceUnitDataHandler handler = this.serviceUnitDataHandlers.get(serviceUnitName);

        for (ServiceEndpoint endpoint : handler.getServiceEndpoints().values()) {
            try {
                this.component.getContext().deactivateEndpoint(endpoint);
            } catch (final JBIException e) {
                if (exception == null) {
                    exception = e;
                } else {
                    this.logger.log(Level.WARNING,
                            "Several errors during undeployment of Service unit '"
                                    + serviceUnitName + "'", e);
                }
            }
        }

        try {
            this.doUndeploy(serviceUnitName);
        } catch (final PEtALSCDKException e) {
            if (exception == null) {
                exception = e;
            } else {
                this.logger.log(Level.WARNING,
                        "Several errors during undeployment of Service unit '" + serviceUnitName
                                + "'", e);
            }
        } finally {
            this.serviceUnitDataHandlers.remove(serviceUnitName);
        }

        final String result;
        if (exception == null) {
            result = ManagementMessageUtil.getComponentTaskResult(this.component.getContext()
                    .getComponentName(), "undeploy", ManagementMessageUtil.TASK_RESULT_SUCCESS);
            this.logger.log(Level.INFO, "Service unit '" + serviceUnitName + "' undeploy");
        } else {
            // instead of throwing an exception, return an error message
            result = ManagementMessageUtil.getComponentTaskResult(this.component.getContext()
                    .getComponentName(), "undeploy", ManagementMessageUtil.TASK_RESULT_FAILED,
                    ManagementMessageUtil.MESSAGE_TYPE_ERROR, "1",
                    "Failed to undeploy Service Unit {1}", new String[] { serviceUnitName },
                    ExceptionUtil.getExtendedMessage(exception));
            this.logger.log(Level.SEVERE, "Service unit '" + serviceUnitName
                    + "' failed to undeploy", exception);
        }

        return result;
    }

    /**
     * Override this method to implement your component logic on the deploy
     * step.
     */
    protected void doDeploy(final String serviceUnitName, final String suRootPath,
            final Jbi jbiDescriptor) throws PEtALSCDKException {
        // do nothing
    }

    /**
     * Override this method to implement your component logic on the init step.
     */
    protected void doInit(final String serviceUnitName, final String suRootPath)
            throws PEtALSCDKException {
        // do nothing
    }

    /**
     * Override this method to implement your component logic on the shut down
     * step.
     */
    protected void doShutdown(final String serviceUnitName) throws PEtALSCDKException {
        // do nothing
    }

    /**
     * Override this method to implement your component logic on the start step.
     */
    protected void doStart(final String serviceUnitName) throws PEtALSCDKException {
        // do nothing
    }

    /**
     * Override this method to implement your component logic on the stop step.
     */
    protected void doStop(final String serviceUnitName) throws PEtALSCDKException {
        // do nothing
    }

    /**
     * Override this method to implement your component logic on the undeploy
     * step.
     */
    protected void doUndeploy(final String serviceUnitName) throws PEtALSCDKException {
        // do nothing
    }

    /**
     * Process the consumes elements.
     * 
     * @param handler
     * @throws JBIException
     */
    private void processConsumesNodes(final ServiceUnitDataHandler handler) throws JBIException {
        final Jbi descriptor = handler.getDescriptor();
        final List<Consumes> consumesList = descriptor.getServices().getConsumes();
        for (final Consumes consumes : consumesList) {
            this.processInterceptors(handler, consumes);
        }
    }

    /**
     * Process the interceptors for a consumes/provides service.
     * 
     * @param handler
     * @param service
     *            the {@code Provides} or {@code Consumes} service
     * @throws JBIException
     */
    private void processInterceptors(final ServiceUnitDataHandler handler, final Object service)
            throws JBIException {

        final SUInterceptors suInterceptors;

        if (service instanceof Consumes) {
            suInterceptors = ((Consumes) service).getSuInterceptors();
        } else if (service instanceof Provides) {
            suInterceptors = ((Provides) service).getSuInterceptors();
        } else {
            throw new PEtALSCDKException(
                    "The service parameter must either a Consumes class or a Provides class");
        }

        // create the manager for this extension element. Each element has a
        // linked manager with preloaded interceptors.

        final MessageInterceptorManager interceptorManager;
        if (this.component.getComponentConfiguration().getPropertiesFile() != null
                && !StringHelper.isNullOrEmpty(this.component.getComponentConfiguration()
                        .getPropertiesFile().getValue())) {
            interceptorManager = new MessageInterceptorManager(this.component
                    .getComponentConfiguration().getPropertiesFile().getValue(), this.logger);
        } else {
            interceptorManager = new MessageInterceptorManager(this.logger);
        }

        List<SUInterceptor> sendSUInterceptors = null;
        List<SUInterceptor> sendResponseSUInterceptors = null;
        List<SUInterceptor> acceptSUInterceptors = null;
        List<SUInterceptor> acceptResponseSUInterceptors = null;
        if (suInterceptors != null) {
            if (suInterceptors.getSend() != null) {
                sendSUInterceptors = suInterceptors.getSend().getInterceptor();
            }
            if (suInterceptors.getSendResponse() != null) {
                sendResponseSUInterceptors = suInterceptors.getSendResponse().getInterceptor();
            }
            if (suInterceptors.getAccept() != null) {
                acceptSUInterceptors = suInterceptors.getAccept().getInterceptor();
            }
            if (suInterceptors.getAcceptResponse() != null) {
                acceptResponseSUInterceptors = suInterceptors.getAcceptResponse().getInterceptor();
            }
        }

        interceptorManager.init(this.component.getInterceptors(), acceptSUInterceptors,
                sendSUInterceptors, acceptResponseSUInterceptors, sendResponseSUInterceptors);

        handler.addMessageInterceptorManager(service, interceptorManager);
    }

    /**
     * Create a JBI descriptor from a file and replace the component referenced
     * properties by their values.
     * 
     * @param jbiFile
     * @return the JBI descriptor in String form
     * @throws IOException
     * @throws PEtALSCDKException
     */
    private String processProperties(final File jbiFile) throws IOException, PEtALSCDKException {
        final FileInputStream jbiFileInputStream = new FileInputStream(jbiFile);
        String jbiString;
        try {
            jbiString =SourceUtil.createString(
                    new StreamSource(jbiFileInputStream));
        } finally {
            jbiFileInputStream.close();
        }
        
        if (this.component.getComponentConfiguration().getPropertiesFile() != null
                && !StringHelper.isNullOrEmpty(this.component.getComponentConfiguration()
                        .getPropertiesFile().getValue())) {

            try {
                final Properties componentProperties = PropertiesUtil.loadProperties(this.component
                        .getComponentConfiguration().getPropertiesFile().getValue());
                for (final Entry<Object, Object> entry : componentProperties.entrySet()) {
                    jbiString = jbiString.replaceAll("\\$\\{" + entry.getKey() + "\\}",
                            (String) entry.getValue());
                }
            } catch (PropertiesException pe) {
                throw new PEtALSCDKException(pe);
            }
        }

        return jbiString;
    }

    /**
     * Process the given provides node. Search the valid location of the wsdl
     * document, or generate one new if needed generate the endpoints and then
     * update the provides, the description and the wsdl file if needed Activate
     * the new endpoint
     * 
     * @param handler
     *            the service unit handler
     * @param provides
     *            the provides of the su
     * @return The resulted Service Endpoint
     * @throws JBIException
     * @throws CDKJBIDescriptorException
     * @throws WSDLException
     * @throws IOException
     * @throws JBIException
     */
    private ServiceEndpoint processProvidesNode(final ServiceUnitDataHandler handler,
            final Provides provides) throws CDKJBIDescriptorException, WSDLException, IOException,
            JBIException {

        boolean bSaveJBI = false;
        boolean bSaveWsdl = false;

        final URL wsdlLocation = getWsdlLocation(handler.getInstallRoot(), provides);

        final Description wsdlDescription = getWsdlDescription(wsdlLocation, provides);
        String endpointName = provides.getEndpointName();
        if (Constants.Component.AUTOGENERATED_ENDPOINT_NAME.equals(endpointName)) {
            endpointName = EndpointUtil.generateEndpointName();
            // update the provides object to set the real endpoint
            provides.setEndpointName(endpointName);
            bSaveJBI = true;

            // change arbitrary all the endpoint named 'autogenerate' found to
            // the given auto-generated value
            final Service service = wsdlDescription.getService(provides.getServiceName());
            if (service != null) {
                final Endpoint autogeneratedEndpoint = service
                        .getEndpoint(Constants.Component.AUTOGENERATED_ENDPOINT_NAME);
                if (autogeneratedEndpoint != null) {
                    autogeneratedEndpoint.setName(endpointName);
                }
                bSaveWsdl = true;
            } else {
                throw new JBIException("Failed to find provided service '"
                        + provides.getServiceName() + "' in the WSDL description");
            }
        }

        // Update dynamically the service name of the WSDL
        final QName originalServiceName = provides.getServiceName();
        final QName newServiceName = this.getServiceName(originalServiceName);
        if (newServiceName != null) {
            // Update the JBI descriptor with the new service name
            provides.setServiceName(newServiceName);
            bSaveJBI = true;

            // Update the WSDL with the new service name
            final Service wsdlService = wsdlDescription.getService(originalServiceName);
            wsdlService.setQName(newServiceName);
            bSaveWsdl = true;
        }

        // Save changes of JBI descriptor and WSDL
        if (bSaveJBI) {
            AbstractServiceUnitManager.saveJBI(handler);
        }
        if (bSaveWsdl) {
            AbstractServiceUnitManager.saveWsdl(wsdlLocation, wsdlDescription);
        }

        final Document wsdlDocument = WSDLUtilImpl.convertDescriptionToDocument(wsdlDescription);

        if (provides.getValidateWsdl() == null || provides.getValidateWsdl().isValue()) {
            this.logger.fine("Validating the provided WSDL description");
            validateDescription(wsdlDescription, provides);
        }

        this.processInterceptors(handler, provides);

        // store in the handler first since the container can try to get the
        // service description first...
        handler.addEndpointDescription(provides.getServiceName(), endpointName, wsdlDocument);
        handler.addJBIDescription(provides.getServiceName(), endpointName, provides);
        handler.addEndpointOperationsList(provides.getServiceName(), endpointName, WSDLUtilImpl
                .getEndpointOperations(wsdlDescription, provides.getServiceName(), endpointName));

        // register and get the new endpoint instance
        final ServiceEndpoint ep = this.component.getContext().activateEndpoint(
                provides.getServiceName(), endpointName);

        handler.addServiceEndpoint(ep);
        return ep;

    }

    private static void saveJBI(final ServiceUnitDataHandler handler)
            throws CDKJBIDescriptorException, IOException {
        final File jbiXmlFile = new File(handler.getInstallRoot(), Constants.Component.META_INF
                + File.separator + Constants.Component.JBI_XML);
        final OutputStream jbiOutputStream = new FileOutputStream(jbiXmlFile);
        try {
            JBIDescriptorBuilder.buildXmlJBIdescriptor(handler.getDescriptor(), jbiOutputStream);
        } finally {
            jbiOutputStream.close();
        }
    }

    private static void saveWsdl(final URL wsdlLocation, final Description wsdlDescription)
            throws WSDLException {
        if (wsdlLocation != null) {
            wsdlDescription.deleteImportedDocumentsInWsdl();

            WSDLUtilImpl
                    .convertDescriptionToFile(wsdlDescription, new File(wsdlLocation.getFile()));

            wsdlDescription.addImportedDocumentsInWsdl();
        }
    }

    /**
     * Extension point to update the service name according to the component
     * needs.
     * 
     * @param originalServiceName
     *            The service name read from the SU archive
     * @return The service name to use, or <code>null</code> if the original
     *         service name must be kept.
     */
    protected QName getServiceName(final QName originalServiceName) {
        return null;
    }

    /**
     * Process the Provides elements.
     * 
     * @param handler
     * @throws IOException
     * @throws CDKJBIDescriptorException
     * @throws WSDLException
     * @throws JBIException
     */
    private void processProvidesNodes(final ServiceUnitDataHandler handler) throws WSDLException,
            CDKJBIDescriptorException, IOException, JBIException {
        final Jbi descriptor = handler.getDescriptor();

        final List<Provides> providesList = descriptor.getServices().getProvides();
        for (final Provides provides : providesList) {
            final ServiceEndpoint sep = this.processProvidesNode(handler, provides);
            this.logger.log(Level.INFO, "New Service Endpoint deployed : " + sep);
        }
    }
}
