/**
 * Copyright (c) 2005-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.su;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.interceptor.MessageInterceptorManager;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.util.ServiceEndpointKey;
import org.w3c.dom.Document;

/**
 * Handles useful information about a Service Unit.
 * <ul>
 * <li>The service unit description (jbi descriptor)</li>
 * <li>Links between Provides nodes and activated endpoints.</li>
 * <li>Links between activated endpoints and corresponding service description.</li>
 * <li>Links between Consumes nodes and corresponding addresses.</li>
 * <li>The installation root.</li>
 * <li>The service unit name - the message interceptor manager.</li>
 * </ul>
 * <p>
 * A SUDataHandler instance is available for each SU of the component.
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 */
public class ServiceUnitDataHandler {

    private Jbi descriptor;

    private Map<ServiceEndpointKey, Provides> epJBIDesc;

    private Map<ServiceEndpointKey, Document> epServiceDesc;

    private final Map<ServiceEndpointKey, List<QName>> epOperations;

    private final Map<ServiceEndpointKey, ServiceEndpoint> endpoints;

    /**
     * Map of service/MessageInterceptorManager
     */
    private final Map<Object, MessageInterceptorManager> messageInterceptorManagersMap;

    /**
     * Map of service/ConfigurationExtensions
     */
    private final Map<Object, ConfigurationExtensions> configurationExtensionsMap;

    /**
     * Map of Consumes/ExchangeProperties
     */
    private final Map<Consumes, ConfigurationExtensions> exchangePropertiesMap;

    private String installRoot;

    private String name;

    /**
     * Creates a new instance of {@link ServiceUnitDataHandler}
     * 
     * @throws PEtALSCDKException
     * 
     */
    private ServiceUnitDataHandler() {
        super();
        this.epServiceDesc = new HashMap<ServiceEndpointKey, Document>();
        this.epJBIDesc = new HashMap<ServiceEndpointKey, Provides>();
        this.endpoints = new HashMap<ServiceEndpointKey, ServiceEndpoint>();
        this.epOperations = new HashMap<ServiceEndpointKey, List<QName>>();
        this.messageInterceptorManagersMap = new HashMap<Object, MessageInterceptorManager>();
        this.configurationExtensionsMap = new HashMap<Object, ConfigurationExtensions>();
        this.exchangePropertiesMap = new HashMap<Consumes, ConfigurationExtensions>();
    }

    /**
     * Creates a new instance of {@link ServiceUnitDataHandler}
     * 
     * @throws PEtALSCDKException
     * 
     */
    public ServiceUnitDataHandler(String suName, String installRoot, Jbi jbiDescriptor) {
        this();
        this.name = suName;
        this.installRoot = installRoot;
        this.descriptor = jbiDescriptor;

        for (Consumes consumes : this.descriptor.getServices().getConsumes()) {
            this.configurationExtensionsMap.put(consumes,
                    new ConfigurationExtensions(consumes.getAny()));
            if (consumes.getExchangeProperties() != null) {
                this.exchangePropertiesMap.put(consumes, new ConfigurationExtensions(consumes
                        .getExchangeProperties().getExchangeProperty()));
            }
        }

        for (Provides provides : this.descriptor.getServices().getProvides()) {
            this.configurationExtensionsMap.put(provides,
                    new ConfigurationExtensions(provides.getAny()));
        }
    }

    /**
     * 
     * @param endpoint
     * @param description
     */
    public void addEndpointDescription(final ServiceEndpoint endpoint, final Document description) {
        this.addEndpointDescription(endpoint.getServiceName(), endpoint.getEndpointName(),
                description);
    }

    /**
     * Return wsdl description of the given endpoint
     * 
     * @param endpoint
     */
    public Document getEndpointDescription(final ServiceEndpoint endpoint) {
        return this.epServiceDesc.get(new ServiceEndpointKey(endpoint.getServiceName(), endpoint
                .getEndpointName()));
    }

    public void addEndpointDescription(final QName serviceName, final String endpointName,
            final Document description) {
        this.epServiceDesc.put(new ServiceEndpointKey(serviceName, endpointName), description);
    }

    /**
     * 
     * @param endpoint
     * @param provides
     */
    public void addJBIDescription(final ServiceEndpoint endpoint, final Provides provides) {
        this.addJBIDescription(endpoint.getServiceName(), endpoint.getEndpointName(), provides);
    }

    public void addJBIDescription(final QName serviceName, final String endpointName,
            final Provides provides) {
        this.epJBIDesc.put(new ServiceEndpointKey(serviceName, endpointName), provides);
    }

    public void addEndpointOperationsList(final QName serviceName, final String endpointName,
            final List<QName> operations) {
        this.epOperations.put(new ServiceEndpointKey(serviceName, endpointName), operations);
    }

    public List<QName> getEndpointOperationsList(final ServiceEndpoint endpoint) {
        return this.epOperations.get(new ServiceEndpointKey(endpoint.getServiceName(), endpoint
                .getEndpointName()));
    }

    public void addServiceEndpoint(final ServiceEndpoint serviceEndpoint) {
        this.endpoints.put(
                new ServiceEndpointKey(serviceEndpoint.getServiceName(), serviceEndpoint
                        .getEndpointName()), serviceEndpoint);
    }

    public Map<ServiceEndpointKey, ServiceEndpoint> getServiceEndpoints() {
        return this.endpoints;
    }

    /**
     * 
     * @return
     */
    public Jbi getDescriptor() {
        return this.descriptor;
    }

    /**
     * 
     * @return
     */
    public Map<ServiceEndpointKey, Provides> getEpJBIDesc() {
        return this.epJBIDesc;
    }

    /**
     * 
     * @return
     */
    public Map<ServiceEndpointKey, Document> getEpServiceDesc() {
        return this.epServiceDesc;
    }

    /**
     * 
     * @return
     */
    public String getInstallRoot() {
        return this.installRoot;
    }

    /**
     * 
     * @return
     */
    public String getName() {
        return this.name;
    }

    /**
     * 
     * @param endpoint
     */
    public void removeEndpoint(final ServiceEndpoint endpoint) {
        ServiceEndpointKey key = new ServiceEndpointKey(endpoint.getServiceName(),
                endpoint.getEndpointName());
        this.epServiceDesc.remove(key);
        this.epJBIDesc.remove(key);
        this.endpoints.remove(key);
        this.epOperations.remove(key);
    }

    /**
     * 
     * @param epJBIDesc
     */
    public void setEpJBIDesc(final Map<ServiceEndpointKey, Provides> epJBIDesc) {
        this.epJBIDesc = epJBIDesc;
    }

    /**
     * 
     * @param epServiceDesc
     */
    public void setEpServiceDesc(final Map<ServiceEndpointKey, Document> epServiceDesc) {
        this.epServiceDesc = epServiceDesc;
    }

    /**
     * Add an interceptor manager.
     * 
     * @param service
     *            A {@code Consumes} or {@code Provides} instance
     * @param manager
     */
    public void addMessageInterceptorManager(Object service, MessageInterceptorManager manager) {

        if ((service instanceof Consumes) || (service instanceof Provides)) {
            this.messageInterceptorManagersMap.put(service, manager);
        } else {
            throw new IllegalArgumentException(
                    "The parameter service  is not an instance of Provides or Consumes Classes");
        }

    }

    /**
     * Get the manager for the given provide/consume element.
     * 
     * @param extensible
     * @return
     */
    public MessageInterceptorManager getMessageInterceptorManager(Object service) {
        return this.messageInterceptorManagersMap.get(service);
    }

    /**
     * Return the map of Interceptor managers.
     * 
     * @return the messageInterceptorManager map
     */
    public Map<Object, MessageInterceptorManager> getMessageInterceptorManagersMap() {
        return this.messageInterceptorManagersMap;
    }

    /**
     * Return the extensions for the given Consumes/Provides service.
     * 
     * @param service
     * @return
     */
    public ConfigurationExtensions getConfigurationExtensions(Object service) {
        return this.configurationExtensionsMap.get(service);
    }

}
