/**
 * Copyright (c) 2007-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.util;

import java.io.File;
import java.io.FilenameFilter;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Pattern;

/**
 * Utilitary class providing java.io FilenameFilters.
 * @author Adrien Louis - EBM WebSourcing
 */
public class FileNamePatternUtil {

    private static FileNamePatternUtil instance = new FileNamePatternUtil();

    private FileNamePatternUtil() {
    }

    /**
     * Return the singleton for this utility class
     * @return
     */
    public static FileNamePatternUtil getInstance() {
        return instance;
    }

    /**
     * Return a FilenameFilter with the given regex filter
     * @param regex
     * @return
     */
    public RegexFileNameFilter buildFileNameFilterFromRegex(String regex) {
        return new RegexFileNameFilter(regex);
    }

    /**
     * Return a FilenameFilter with the given wildchar filter (using * and ?)
     * @param wildChar
     * @return
     */
    public RegexFileNameFilter buildFileNameFilterFromWildChar(String wildChar) {
        return new RegexFileNameFilter(wildcharToRegex(wildChar));
    }

    /**
     * Return a FilenameFilter matching all filters
     * @param filters
     * @return
     */
    public RegexFileNameFilter buildFileNameFilterFromFilters(List<RegexFileNameFilter> filters) {
        return new RegexFileNameFilterList(filters);
    }

    /**
     * convert a regex expression to a standard file wildchar filter
     * @param regex
     * @return
     */
    public String regexToWildchar(String regex) {
        String a = regex.replaceAll("\\\\.", "²");
        String b = a.replaceAll("\\.", "?");
        String c = b.replaceAll("²", ".");
        return c;
    }

    /**
     * convert a standard file wildchar filter to a regex expression
     * @param wildChar
     * @return
     */
    public String wildcharToRegex(String wildChar) {
        return wildChar.replaceAll("\\.", "\\\\.").replace("*", ".*").replaceAll("\\?", ".");
    }

    public static class RegexFileNameFilter implements FilenameFilter {

        private static final String ANY_VALUE = ".*";

        private String regex;

        public RegexFileNameFilter() {
            this(ANY_VALUE);
        }

        public RegexFileNameFilter(String regex) {
            this.regex = regex;
        }

        public boolean accept(File dir, String name) {
            return Pattern.compile(regex, Pattern.CASE_INSENSITIVE).matcher(name).matches();
        }
    }

    public static class RegexFileNameFilterList extends RegexFileNameFilter {

        private List<RegexFileNameFilter> list;

        public RegexFileNameFilterList(List<RegexFileNameFilter> list) {
            this.list = list;
        }

        public boolean accept(File dir, String name) {
            boolean accept = false;
            Iterator<RegexFileNameFilter> iterator = list.iterator();
            while (iterator.hasNext() && !accept) {
                accept = iterator.next().accept(null, name);
            }
            return accept;
        }
    }

   
}
