/**
 * Copyright (c) 2005-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.URI;
import java.nio.charset.Charset;

import javax.xml.namespace.QName;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.component.framework.api.exception.DocumentException;
import org.ow2.petals.component.framework.api.exception.FaultException;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.ebmwebsourcing.easycommons.lang.StringHelper;
import com.ebmwebsourcing.easycommons.lang.UncheckedException;
import com.ebmwebsourcing.easycommons.pooling.PoolException;
import com.ebmwebsourcing.easycommons.xml.DocumentBuilders;
import com.ebmwebsourcing.easycommons.xml.SourceHelper;
import com.ebmwebsourcing.easycommons.xml.Transformers;
import com.ebmwebsourcing.easycommons.xml.XMLHelper;

/**
 * Sample class used to manage the mapping between
 * <code>javax.xml.transform.Source</code> from String. the mapping is a quicly
 * "string" -> "<content>string</content>" conversion.
 * 
 * @author Adrien Louis - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 * @author Nicolas Oddoux - EBM WebSourcing
 */
public class SourceUtil {

    private final static String SOAPENV_NAMESPACE = "http://schemas.xmlsoap.org/soap/envelope/";

    /**
     * Creates a new instance of {@link SourceUtil}
     * 
     */
    public SourceUtil() {
        super();
    }

    /**
     * Fork the given StreamSource
     * 
     * @param streamSource
     * @return
     */
    private static final InputStream forkStreamSource(final StreamSource streamSource) {
    	
    	try {
			return ((StreamSource) SourceHelper.fork(streamSource)).getInputStream();
		} catch (IOException e) {
          throw new UncheckedException(e);
		}
    }

    /**
     * Create a {@link Document} from a {@link Source}. <br>
     * -If the Source is a {@link StreamSource}, the source would be still
     * available.<br>
     * -If the Source is a {@link DomSource} and contains a {@link Document},
     * the returned document is a clone of this document.
     * 
     * @param source
     *            the {@link Source}
     * @return the DOM document built from the Source
     * @throws PEtALSCDKException
     */
    public static final Document createDocument(final Source source) throws PEtALSCDKException {
        return createDocument(source, true);
    }

    /**
     * Create a {@link Document} from a {@link Source}. <br>
     * -If the Source is a {@link StreamSource} and {@code forkSource} is set to
     * {@code true}, the source would be still available. Otherwise the source
     * is no more available.<br>
     * If the Source is a {@link DomSource} and contains a {@link Document} and
     * {@code forkSource} is set to {@code true}, the returned document is a
     * clone of this document. Otherwise the original {@link Document} is
     * returned.
     * 
     * @param source
     *            the {@link Source}
     * @param forkSource
     *            {@code true} to fork the content of the source if necessary
     * @return a DOM document
     * @throws PEtALSCDKException
     */
    public static final Document createDocument(final Source source, final boolean forkSource)
            throws PEtALSCDKException {
        Document document = null;
        try {
            if (source instanceof DOMSource) {
                final Node originalNode = ((DOMSource) source).getNode();
                if (originalNode instanceof Document) {
                    final Document originalDocument = (Document) originalNode;
                    if (forkSource) {
                        // clone the document
                        document = DocumentBuilders.newDocument();
                        document.appendChild(document.importNode(
                                originalDocument.getDocumentElement(), true));
                    } else {
                        // refer the document
                        document = originalDocument;
                    }
                } else {
                    document = DocumentBuilders.newDocument();
                    if (forkSource) {
                        // clone the node
                        document.appendChild(document.importNode(originalNode, true));
                    } else {
                        // adopt the node
                        document.appendChild(document.adoptNode(originalNode));
                    }

                }
            }

            if (document == null) {
                Source tempSource = source;
                if (forkSource && source instanceof StreamSource) {
                    tempSource = new StreamSource(forkStreamSource((StreamSource) source));
                }
                document = DocumentBuilders.newDocument();
                DOMResult domResult = new DOMResult(document);
                Transformer transformer = Transformers.takeTransformer();
                try {
                    transformer.transform(tempSource, domResult);
                } finally {
                    Transformers.releaseTransformer(transformer);
                }
            }
        } catch (final TransformerException e) {
            throw new PEtALSCDKException(e);
        } catch (final PoolException e) {
            throw new PEtALSCDKException(e);
        }
        
        return document;
    }

    /**
     * Create a DOM Source from the DOM document.
     * 
     * @param document
     *            A DOM document
     * @return the DOM source built from the DOM document
     */
    public static final DOMSource createDOMSource(final Document document) {
        // normalize the document to assure the resolution of each node
        // namespace
        document.normalizeDocument();
        return new DOMSource(document);
    }

    /**
     * <p>
     * Creates an SOAP fault, as a DOM document, according to provided exception
     * and code:<br>
     * <code>&lt;s:Fault xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;faultcode&gt;{org.objectweb.petals}ERROR&lt;/faultcode&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;faultstring&gt;mesage&lt;/faultstring&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;faultactor&gt;component&lt;/faultactor&gt;<br>
     * &nbsp;&nbsp;&nbsp;&lt;detail&gt;<br>
     * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;...<br>
     * &nbsp;&nbsp;&nbsp;&lt;/detail&gt;<br>
     * &lt;s:Fault&gt;</code>
     * </p>
     * 
     * <p>
     * If the provided exception is an instance of {@link DocumentException}:
     * <ul>
     * <li>the fault code is set with the value returned by
     * {@link DocumentException#getFaultCode()},</li>
     * <li>the fault string is set with the value returned by
     * {@link DocumentException#getMessage()},</li>
     * <li>the fault actor is set with the value returned by
     * {@link DocumentException#getFaultActor()},</li>
     * <li>the detail is set with the exception document returned by
     * {@link DocumentException#getDocument()}.</li>
     * </ul>
     * </p>
     * 
     * <p>
     * If the provided exception is an instance of {@link FaultException}:
     * <ul>
     * <li>the fault code is set with the value returned by
     * {@link FaultException#getFaultCode()},</li>
     * <li>the fault string is set with the value returned by
     * {@link FaultException#getMessage()},</li>
     * <li>the fault actor is set with the value returned by
     * {@link FaultException#getFaultActor()},</li>
     * <li>the detail is set with the exception stack trace enclosed in the node
     * <code>&lt;stackTrace&gt;</code>.</li>
     * </ul>
     * </p>
     * 
     * <p>
     * otherwise:
     * <ul>
     * <li>the fault code is set with the value of <code>errorCode</code>,</li>
     * <li>the fault string is set with the value returned by
     * {@link Exception#getMessage()},</li>
     * <li>the fault actor is set to <code>null</code>,</li>
     * <li>the detail is set with the exception stack trace enclosed in the node
     * <code>&lt;stackTrace&gt;</code>.</li>
     * </ul>
     * </p>
     * 
     * @param e
     *            An exception used to provide faultstring and detail
     * @param errorCode
     *            The QName of the fault code. It is not used if the exception
     *            is an instance of {@link FaultException} or
     *            {@link DocumentException}.
     * @return A Document containing the Soap Fault
     */
    public static final Document createSoapFaultDocument(final Throwable e, final QName errorCode) {
        QName faultCode = errorCode;
        String faultString = e.getMessage();
        if (faultString == null) {
            faultString = "null message";
        }
        String details;
        URI faultActor = null;
        if (e instanceof DocumentException) {
            faultCode = ((DocumentException) e).getFaultCode();
            faultActor = ((DocumentException) e).getFaultActor();
            final Document document = ((DocumentException) e).getDocument();
            if (document != null) {
                try {
                    details = XMLHelper.createStringFromDOMDocument(document);
                } catch (final TransformerException e1) {
                    details = "An error occured when retrieving the exception document ("
                            + e1.getMessage() + ").";
                } catch (PoolException e1) {
                    details = "An error occured when retrieving the exception document ("
                        + e1.getMessage() + ").";
                }
            } else {
                details = "";
            }
        } else if (e instanceof FaultException) {
            faultCode = ((FaultException) e).getFaultCode();
            faultActor = ((FaultException) e).getFaultActor();
            // The stacktrace will be enclosed in a section CDATA when adding
            // the DOM node
            final StringWriter sw = new StringWriter();
            final PrintWriter pw = new PrintWriter(sw);
            e.printStackTrace(pw);
            details = sw.toString();

        } else {
            // The stacktrace will be enclosed in a section CDATA when adding
            // the DOM node
            final StringWriter sw = new StringWriter();
            final PrintWriter pw = new PrintWriter(sw);
            e.printStackTrace(pw);
            details = sw.toString();
        }
        // warning, the building of the document must be compatible DOM3 AND
        // DOM2
        final Document faultDocument = DocumentBuilders.newDocument();

        final Element faultElement = faultDocument.createElementNS(SOAPENV_NAMESPACE,
                "soapenv:Fault");
        // the NameSpace is not fill in an attribute with old XALAN DOM2
        // transformation
        if (!isDOM3(faultDocument)) {
            final Attr attr = faultDocument.createAttributeNS("http://www.w3.org/2000/xmlns/",
                    "xmlns:soapenv");
            attr.appendChild(faultDocument.createTextNode(SOAPENV_NAMESPACE));
            faultElement.setAttributeNode(attr);
        }
        faultDocument.appendChild(faultElement);

        final Element faultCodeElement = faultDocument.createElement("faultcode");
        if (faultCode != null) {
            if (faultCode.getNamespaceURI() != null) {
                if (faultCode.getNamespaceURI().equals(SOAPENV_NAMESPACE)) {
                    faultCodeElement.appendChild(faultDocument.createTextNode("soapenv:"
                            + faultCode.getLocalPart()));
                } else {
                    String faultCodePrefix = faultCode.getPrefix();
                    if (StringHelper.isNullOrEmpty(faultCodePrefix)) {
                        faultCodePrefix = "fcns";
                    }
                    final Attr attr = faultDocument.createAttributeNS(
                            "http://www.w3.org/2000/xmlns/", "xmlns:" + faultCodePrefix);
                    attr.appendChild(faultDocument.createTextNode(faultCode.getNamespaceURI()));
                    faultCodeElement.setAttributeNode(attr);
                    faultCodeElement.appendChild(faultDocument.createTextNode(faultCodePrefix + ":"
                            + faultCode.getLocalPart()));
                }
            } else {
                faultCodeElement
                        .appendChild(faultDocument.createTextNode(faultCode.getLocalPart()));
            }
        }
        
        // The node "faultstring" value must be enclosed in a CDATA section to
        // prevent encoding problems
        final Element faultStringElement = faultDocument.createElement("faultstring");
        faultStringElement.appendChild(faultDocument.createCDATASection(faultString));
        final Element faultDetailElement = faultDocument.createElement("detail");
        if (e instanceof DocumentException) {
            faultDetailElement.appendChild(faultDocument.createCDATASection(details));
        } else {
            final Element stackTrace = faultDocument.createElement("stackTrace");
            faultDetailElement.appendChild(stackTrace);
            stackTrace.appendChild(faultDocument.createCDATASection(details));
        }
        final Element faultActorElement = faultDocument.createElement("faultactor");
        if (faultActor != null) {
            faultActorElement.appendChild(faultDocument.createTextNode(faultActor.toString()));
        }
        faultElement.appendChild(faultCodeElement);
        faultElement.appendChild(faultStringElement);
        faultElement.appendChild(faultActorElement);
        faultElement.appendChild(faultDetailElement);
        return faultDocument;
    }

    /**
     * Create a Source from the String. The message must be a valid XML String.
     * If an encoding is specified in the XML String header it is used to create
     * the XML bytes stream, otherwise the default JVM encoding is used.
     * 
     * @param message
     *            a valid XML String
     * @return the Source built from the XML String
     * @throws PEtALSCDKException
     *             if an error occurs during Source creation
     * 
     */
    public static final Source createSource(final String message) throws PEtALSCDKException {
        final String charsetEncoding = Charset.defaultCharset().displayName();
        return createSource(message, charsetEncoding);
    }

    /**
     * Create a Source from the String with a specified encoding format. The
     * message must be a valid XML String. The encoding must match the XML
     * declaration in the message.
     * 
     * @param message
     *            a valid XML String
     * @param charFormat
     *            the encoding format, must be non {@code null}
     * 
     * @return the Source built from the XML String
     * @throws PEtALSCDKException
     *             if an error occurs during Source creation
     */
    public static final Source createSource(final String message, final String charFormat)
            throws PEtALSCDKException {
        final StreamSource source = new StreamSource();
        try {
            final byte[] msgByte = message.getBytes(charFormat);
            final ByteArrayInputStream in = new ByteArrayInputStream(msgByte);
            source.setInputStream(in);
        } catch (final IOException e) {
            throw new PEtALSCDKException(e);
        }
        return source;
    }

    /**
     * Create a Stream Source from the DOM document.
     * 
     * @param document
     *            A DOM document
     * @return the Stream source built from the DOM document
     */
    public static final StreamSource createStreamSource(final Document document) throws PEtALSCDKException {
        ByteArrayOutputStream bufferOut = new ByteArrayOutputStream();
        Result sResult = new StreamResult(bufferOut);
        Transformer transformer = null;
        try {
            transformer = Transformers.takeTransformer();
            transformer.transform(new DOMSource(document), sResult);
        } catch (final TransformerException e) {
            throw new PEtALSCDKException("Error while transform DOM document to StreamSource", e);
        } catch (PoolException e) {
            throw new PEtALSCDKException("Error while transform DOM document to StreamSource", e);
        } finally {
            if (transformer != null) {
                Transformers.releaseTransformer(transformer);
            }
        }
        
        // TODO (CDE): Use CircularBuffer to increase performances.
        final ByteArrayInputStream bufferIn = new ByteArrayInputStream(bufferOut.toByteArray());
        return new StreamSource(bufferIn); 
    }

    /**
     * Create a {@link String} from a {@link Source}.<br>
     * -If the Source is a {@code StreamSource}, the source would be still
     * available.<br>
     * 
     * @param source
     *            the {@link Source}
     * @return the String built from the Source
     * @throws PEtALSCDKException
     */
    public static final String createString(final Source source) throws PEtALSCDKException {
        return createString(source, true);
    }

    /**
     * Create a {@link String} from a {@link Source}.<br>
     * -If the Source is a {@link StreamSource} and {@code forkSource} is set to
     * {@code true}, the source would be still available. Otherwise the source
     * is no more available.<br>
     * 
     * @param source
     *            the {@link Source}
     * @return the String built from the Source
     * @throws PEtALSCDKException
     */
    public static final String createString(final Source source, final boolean forkSource)
            throws PEtALSCDKException {
        Source tempSource = source;
        if (forkSource && source instanceof StreamSource) {
            tempSource = new StreamSource(forkStreamSource((StreamSource) source));
        }

        try {
            StringWriter buffer = new StringWriter();
            Result sresult = new StreamResult(buffer);
            Transformer transformer = Transformers.takeTransformer();
            try {
                transformer.transform(tempSource, sresult);
            } finally {
                Transformers.releaseTransformer(transformer);
            }

            return buffer.toString();
        } catch (final TransformerException e) {
            throw new PEtALSCDKException(e);
        } catch (PoolException e) {
            throw new PEtALSCDKException(e);
        }
    }

    /**
     * Test if the document is a DOM3 document
     * 
     * @param document
     *            the document to test
     * @return true if the document is DOM3, false else
     */
    public static final boolean isDOM3(final Document document) {
        boolean result = true;

        try {
            document.getClass().getMethod("getBaseURI", (Class[]) null);
        } catch (final NoSuchMethodException e) {
            result = false;
        }

        return result;
    }

}
