/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.lang.ref.PhantomReference;
import java.lang.reflect.InvocationTargetException;
import java.util.List;
import java.util.logging.Logger;

import javax.jbi.JBIException;

import junit.framework.TestCase;

import org.junit.Test;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.mbean.RuntimeConfigurationNotifier;
import org.ow2.petals.probes.api.exceptions.MultipleProbesFactoriesFoundException;
import org.ow2.petals.probes.api.exceptions.NoProbesFactoryFoundException;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

public class JBIAcceptorManagerTest extends AbstractJBIManager<JBIAcceptorManager> {
    
    private final static Logger LOG = Logger.getLogger(JBIAcceptorManagerTest.class.getName());

    private void assertAcceptorThreadGroupIsFreed()
            throws IllegalArgumentException, IllegalAccessException {
        
        try {
            // Wait thread context switches
            Thread.sleep(100);
        } catch (InterruptedException e) {
            // NOP
        }
        System.gc();
        
        final ThreadGroup[] childrenThreadGroup = new ThreadGroup[50];
        Thread.currentThread().getThreadGroup().enumerate(childrenThreadGroup);
        for (final ThreadGroup childThreadGroup : childrenThreadGroup) {
            if (childThreadGroup != null
                    && childThreadGroup.getName().startsWith("my-component-JBIListener")) {
                TestCase.fail("The JBI Acceptor Thread group is not freed.");
                break;
            }
        }
    }

    /**
     * Checks the JBIAcceptorManager is correctly garbaged.
     */
    @Test
    public void testGarbaged_001() throws JBIException, IllegalArgumentException,
            IllegalAccessException, MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null,
                null, 10, 5, null, null, null);

        final JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);
        
        JBIAcceptorManager acceptorMgr = new JBIAcceptorManager(componentMock, procMgr,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        acceptorMgr.start();

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIAcceptorManager> refAcceptorMgr = new PhantomReference<JBIAcceptorManager>(
                acceptorMgr, this.queueGC);
        acceptorMgr.stop();

        acceptorMgr = null;
        this.assertIsGarbaged(refAcceptorMgr);
        this.assertAcceptorThreadGroupIsFreed();

    }
    
    /**
     * Checks the start against default values.
     * Expected result: all parameter values are defined with their default values.
     */
    @Test
    public void testStart_000() throws JBIException, IllegalArgumentException,
            IllegalAccessException, InvocationTargetException,
            MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {
        
        final int acceptorPoolSize = 10;
        final AbstractComponent componentMock = this.createComponentMock(acceptorPoolSize, null,
                null, null, null, 10, null, null, null, null);
    
        final JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);
    
        JBIAcceptorManager acceptorMgr = new JBIAcceptorManager(componentMock, procMgr,
                new RuntimeConfigurationNotifier(null, null, null), LOG);
    
        acceptorMgr.start();
        
        final List<MessageExchangeAcceptor> acceptorPool = (List<MessageExchangeAcceptor>) ReflectionHelper
                .getPrivateFieldValue(acceptorMgr.getClass(), acceptorMgr, "acceptorPool");
        TestCase.assertEquals("Acceptor pool size is not the expected one.", acceptorPoolSize,
                acceptorPool.size());

        TestCase.assertEquals("Acceptor retry policy max number has not the default value.",
                JBIAcceptorManager.DEFAULT_ACCEPTOR_RETRY_MAX_TRY, acceptorMgr.getRetryMaxTry());

        TestCase.assertEquals("Acceptor retry policy base duration has not the default value.",
                JBIAcceptorManager.DEFAULT_ACCEPTOR_RETRY_BASE_DURATION,
                acceptorMgr.getRetryBaseDuration());
        
        acceptorMgr.stop();
    
        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIAcceptorManager> refAcceptorMgr = new PhantomReference<JBIAcceptorManager>(
                acceptorMgr, this.queueGC);
        acceptorMgr.stop();
    
        acceptorMgr = null;
        this.assertIsGarbaged(refAcceptorMgr);
        this.assertAcceptorThreadGroupIsFreed();
    
    }

    /**
     * Checks the start against a value of acceptor retry max number set to a
     * negative value. Expected result: the default value is used.
     */
    @Test
    public void testStart_001() throws JBIException, IllegalArgumentException,
            IllegalAccessException, InvocationTargetException,
            MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final AbstractComponent componentMock = this.createComponentMock(10, null, -10, null, null,
                10, null, null, null, null);

        final JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        JBIAcceptorManager acceptorMgr = new JBIAcceptorManager(componentMock, procMgr,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        acceptorMgr.start();

        TestCase.assertEquals("Acceptor retry policy max number has not the default value.",
                JBIAcceptorManager.DEFAULT_ACCEPTOR_RETRY_MAX_TRY, acceptorMgr.getRetryMaxTry());

        acceptorMgr.stop();

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIAcceptorManager> refAcceptorMgr = new PhantomReference<JBIAcceptorManager>(
                acceptorMgr, this.queueGC);
        acceptorMgr.stop();

        acceptorMgr = null;
        this.assertIsGarbaged(refAcceptorMgr);
        this.assertAcceptorThreadGroupIsFreed();

    }

    /**
     * Checks the start against a value of acceptor retry max number set to 0.
     * Expected result: the default value is used.
     */
    @Test
    public void testStart_002() throws JBIException, IllegalArgumentException,
            IllegalAccessException, InvocationTargetException,
            MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final AbstractComponent componentMock = this.createComponentMock(10, null, 0, null, null,
                10, null, null, null, null);

        final JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        JBIAcceptorManager acceptorMgr = new JBIAcceptorManager(componentMock, procMgr,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        acceptorMgr.start();

        TestCase.assertEquals("Acceptor retry policy max number has not the default value.",
                JBIAcceptorManager.DEFAULT_ACCEPTOR_RETRY_MAX_TRY, acceptorMgr.getRetryMaxTry());

        acceptorMgr.stop();

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIAcceptorManager> refAcceptorMgr = new PhantomReference<JBIAcceptorManager>(
                acceptorMgr, this.queueGC);
        acceptorMgr.stop();

        acceptorMgr = null;
        this.assertIsGarbaged(refAcceptorMgr);
        this.assertAcceptorThreadGroupIsFreed();

    }

    /**
     * Checks the start against a value of acceptor retry base duration set to a
     * negative value. Expected result: the default value is used.
     */
    @Test
    public void testStart_003() throws JBIException, IllegalArgumentException,
            IllegalAccessException, InvocationTargetException,
            MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, -750l,
                null, 10, null, null, null, null);

        final JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        JBIAcceptorManager acceptorMgr = new JBIAcceptorManager(componentMock, procMgr,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        acceptorMgr.start();

        TestCase.assertEquals("Acceptor retry policy base duration has not the default value.",
                JBIAcceptorManager.DEFAULT_ACCEPTOR_RETRY_BASE_DURATION,
                acceptorMgr.getRetryBaseDuration());

        acceptorMgr.stop();

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIAcceptorManager> refAcceptorMgr = new PhantomReference<JBIAcceptorManager>(
                acceptorMgr, this.queueGC);
        acceptorMgr.stop();

        acceptorMgr = null;
        this.assertIsGarbaged(refAcceptorMgr);
        this.assertAcceptorThreadGroupIsFreed();

    }

    /**
     * Checks the start against a value of acceptor retry base duration set to
     * 0. Expected result: the default value is used.
     */
    @Test
    public void testStart_004() throws JBIException, IllegalArgumentException,
            IllegalAccessException, InvocationTargetException,
            MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, 0l, null,
                10, null, null, null, null);

        final JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        JBIAcceptorManager acceptorMgr = new JBIAcceptorManager(componentMock, procMgr,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        acceptorMgr.start();

        TestCase.assertEquals("Acceptor retry policy base duration has not the default value.",
                JBIAcceptorManager.DEFAULT_ACCEPTOR_RETRY_BASE_DURATION,
                acceptorMgr.getRetryBaseDuration());

        acceptorMgr.stop();

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIAcceptorManager> refAcceptorMgr = new PhantomReference<JBIAcceptorManager>(
                acceptorMgr, this.queueGC);
        acceptorMgr.stop();

        acceptorMgr = null;
        this.assertIsGarbaged(refAcceptorMgr);
        this.assertAcceptorThreadGroupIsFreed();

    }

    /**
     * Checks the stop against a value of 'acceptor-stop-max-wait' not set.
     * Expected result: no error thrown and the duration of stop of the
     * JBIAcceptorManager is lower than the default value (500ms).
     */
    @Test
    public void testStop_001() throws JBIException, IllegalArgumentException,
            IllegalAccessException, MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null,
                null, 10, null, null, null, null);

        final JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        JBIAcceptorManager acceptorMgr = new JBIAcceptorManager(componentMock, procMgr,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        acceptorMgr.start();

        final long start = System.currentTimeMillis();

        acceptorMgr.stop();

        TestCase.assertTrue(
                "The stop duration the acceptor pool is higher than the default value (500s).",
                System.currentTimeMillis() - start <= 10 * JBIAcceptorManager.DEFAULT_ACCEPTOR_STOP_MAX_WAIT);

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIAcceptorManager> refAcceptorMgr = new PhantomReference<JBIAcceptorManager>(
                acceptorMgr, this.queueGC);
        acceptorMgr.stop();

        acceptorMgr = null;
        this.assertIsGarbaged(refAcceptorMgr);
        this.assertAcceptorThreadGroupIsFreed();

    }

    /**
     * Checks the stop against a value of 'acceptor-stop-max-wait' set to 0.
     * Expected result: no error thrown and the duration of stop of the
     * JBIAcceptorManager is lower than the default value (500ms).
     */
    @Test
    public void testStop_002() throws JBIException, IllegalArgumentException,
            IllegalAccessException, MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        final AbstractComponent componentMock = this.createComponentMock(10, 0l, null, null, null,
                10, null, null, null, null);

        final JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        JBIAcceptorManager acceptorMgr = new JBIAcceptorManager(componentMock, procMgr,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        acceptorMgr.start();

        final long start = System.currentTimeMillis();

        acceptorMgr.stop();

        TestCase.assertTrue(
                "The stop duration the acceptor pool is higher than the default value (500s).",
                System.currentTimeMillis() - start <= 10 * JBIAcceptorManager.DEFAULT_ACCEPTOR_STOP_MAX_WAIT);

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIAcceptorManager> refAcceptorMgr = new PhantomReference<JBIAcceptorManager>(
                acceptorMgr, this.queueGC);
        acceptorMgr.stop();

        acceptorMgr = null;
        this.assertIsGarbaged(refAcceptorMgr);
        this.assertAcceptorThreadGroupIsFreed();

    }

    /**
     * Checks the stop against a value of 'acceptor-stop-max-wait' set to a
     * negative value. Expected result: no error thrown and the duration of stop
     * of the JBIAcceptorManager is lower than the default value (500ms).
     */
    @Test
    public void testStop_003() throws JBIException, IllegalArgumentException,
            IllegalAccessException, MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        final AbstractComponent componentMock = this.createComponentMock(10, -500l, null, null,
                null, 10, null, null, null, null);

        final JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        JBIAcceptorManager acceptorMgr = new JBIAcceptorManager(componentMock, procMgr,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        acceptorMgr.start();

        final long start = System.currentTimeMillis();

        acceptorMgr.stop();

        TestCase.assertTrue(
                "The stop duration the acceptor pool is higher than the default value (500s).",
                System.currentTimeMillis() - start <= 10 * JBIAcceptorManager.DEFAULT_ACCEPTOR_STOP_MAX_WAIT);

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIAcceptorManager> refAcceptorMgr = new PhantomReference<JBIAcceptorManager>(
                acceptorMgr, this.queueGC);
        acceptorMgr.stop();

        acceptorMgr = null;
        this.assertIsGarbaged(refAcceptorMgr);
        this.assertAcceptorThreadGroupIsFreed();

    }
}
