/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.lang.ref.PhantomReference;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import javax.jbi.JBIException;

import junit.framework.TestCase;

import org.junit.Ignore;
import org.junit.Test;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.mbean.RuntimeConfigurationNotifier;
import org.ow2.petals.probes.api.exceptions.MultipleProbesFactoriesFoundException;
import org.ow2.petals.probes.api.exceptions.NoProbesFactoryFoundException;

public class JBIProcessorManagerTest extends AbstractJBIManager<JBIProcessorManager> {
    
    private final static Logger LOG = Logger.getLogger(JBIProcessorManagerTest.class.getName());

    /**
     * Checks the start against valid configuration parameters of processor core
     * pool. Expected result: configuration parameter values are used
     */
    @Test
    public void testStart_000() throws JBIException, MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        final int acceptorCoreSize = 10;
        final long acceptorStopMaxWait = 500l;
        final int acceptorRetryNb = 15;
        final long acceptorRetryWait = 750;
        final int messageProcessorsMaxPoolSize = 61;
        final int processorCoreSize = 10;
        final int processorMaxSize = 50;
        final int processorKeepAliveTime = 250;
        final long processorStopMaxWait = 670l;

        final AbstractComponent componentMock = this.createComponentMock(acceptorCoreSize,
                acceptorStopMaxWait, acceptorRetryNb, acceptorRetryWait, messageProcessorsMaxPoolSize, processorCoreSize, processorMaxSize,
                processorKeepAliveTime, processorStopMaxWait, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        // TODO: add a check about acceptor core pool size
        // TODO: add a check about acceptor pool stop max wait
        TestCase.assertEquals("The processor core pool size is not set to the configured value.",
                processorCoreSize, procMgr.getProcessorThreadPool().getCorePoolSize());
        TestCase.assertEquals("The processor max pool size is not set to the configured value.",
                processorMaxSize, procMgr.getProcessorThreadPool().getMaximumPoolSize());
        TestCase.assertEquals("The processor keep alive time is not set to the configured value.",
                processorKeepAliveTime, procMgr.getProcessorThreadPool().getKeepAliveTime(TimeUnit.SECONDS));
        // TODO: add a check about processor pool stop max wait

        TestCase.assertEquals(
                "The processor object pool max size is not set to the configured value.",
                messageProcessorsMaxPoolSize, procMgr.getProcessorObjectPool().getMaxActive());
        TestCase.assertEquals(
                "The processor object pool min size is not set to the configured value.",
                processorCoreSize, procMgr.getProcessorObjectPool().getMinIdle());
        TestCase.assertEquals(
                "The processor object pool min size is not set to the configured value.",
                processorKeepAliveTime, procMgr.getProcessorObjectPool().getMinEvictableIdleTimeMillis());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * Checks the start against a value of processor core pool size higher than
     * the max pool size. Expected result: the core pool size is used as max
     * pool size.
     */
    @Test
    public void testStart_001() throws JBIException, MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null,
                null, 10, 5, null, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);
        
        procMgr.start();

        TestCase.assertEquals("The processor max pool size is not set to the core pool size.",
                procMgr.getProcessorThreadPool().getCorePoolSize(), procMgr.getProcessorThreadPool()
                        .getMaximumPoolSize());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * Checks the start against a value of processor max pool size not set.
     * Expected result: the default value is used.
     */
    @Test
    public void testStart_002() throws JBIException, MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null,
                null, 10, null, null, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        TestCase.assertEquals("The processor max pool size is not set to the default value.",
                JBIProcessorManager.DEFAULT_PROCESSOR_POOL_SIZE_MAX, procMgr.getProcessorThreadPool()
                        .getMaximumPoolSize());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);
        
    }

    /**
     * Checks the start against a value of keep alive time of the processor pool not set.
     * Expected result: the default value is used (5min).
     */
    @Test
    public void testStart_003() throws JBIException, MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null,
                null, 10, null, null, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        TestCase.assertEquals("The keep alive time of the processor pool is not set to the default value.",
                JBIProcessorManager.DEFAULT_PROCESSOR_POOL_SIZE_KEEP_ALIVE_TIME, procMgr.getProcessorThreadPool()
                        .getKeepAliveTime(TimeUnit.SECONDS));

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);
        
    }

    /**
     * Checks the start against a value of keep alive time of the processor pool
     * that is 0. Expected result: the default value is used (5min).
     */
    @Test
    public void testStart_004() throws JBIException, MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null,
                null, 10, null, 0, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        TestCase.assertEquals(
                "The keep alive time of the processor pool is not set to the default value.",
                JBIProcessorManager.DEFAULT_PROCESSOR_POOL_SIZE_KEEP_ALIVE_TIME, procMgr
                        .getProcessorThreadPool().getKeepAliveTime(TimeUnit.SECONDS));

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * Checks the start against a value of keep alive time of the processor pool
     * that is negative. Expected result: the default value is used (5min).
     */
    @Test
    public void testStart_005() throws JBIException, MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null,
                null, 10, null, -300, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        TestCase.assertEquals(
                "The keep alive time of the processor pool is not set to the default value.",
                JBIProcessorManager.DEFAULT_PROCESSOR_POOL_SIZE_KEEP_ALIVE_TIME, procMgr
                        .getProcessorThreadPool().getKeepAliveTime(TimeUnit.SECONDS));

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * Checks the start against a value of message processor max pool size set
     * to a negative value. Expected result: the default value is used (max size
     * of the processor thread pool).
     */
    @Test
    public void testStart_006() throws JBIException, MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        final int processorMaxSize = 43;

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null, -15,
                10, processorMaxSize, null, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        TestCase.assertEquals(
                "The message processor object pool max size is not set to its default value.",
                processorMaxSize, procMgr.getProcessorObjectPool().getMaxActive());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * Checks the start against a value of message processor max pool size set
     * to 0. Expected result: the default value is used (max size of the
     * processor thread pool).
     */
    @Test
    public void testStart_007() throws JBIException, MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        final int processorMaxSize = 43;

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null, 0,
                10, processorMaxSize, null, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        TestCase.assertEquals(
                "The message processor object pool max size is not set to its default value.",
                processorMaxSize, procMgr.getProcessorObjectPool().getMaxActive());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * Checks the start against a value of message processor max pool size set
     * to a value lower than the max size of the processor thread pool. Expected
     * result: the default value is used (max size of the processor thread
     * pool).
     */
    @Test
    public void testStart_008() throws JBIException, MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        final int processorMaxSize = 43;

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null, 5,
                10, processorMaxSize, null, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        TestCase.assertEquals(
                "The message processor object pool max size is not set to its default value.",
                processorMaxSize, procMgr.getProcessorObjectPool().getMaxActive());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * Checks the stop against a value of 'processor-stop-max-wait' not set.
     * Expected result: no error thrown and the duration of stop of the
     * JBIProcessorManager is lower than the default value (15s).
     */
    @Test
    public void testStop_001() throws JBIException, MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null,
                null, 10, null, null, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        final long start = System.currentTimeMillis();

        procMgr.stop();

        TestCase.assertTrue(
                "The stop duration the processor pool is higher than the default value (15s).",
                System.currentTimeMillis() - start <= JBIProcessorManager.DEFAULT_PROCESSOR_STOP_MAX_WAIT);

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * Checks the stop against a value of 'processor-stop-max-wait' set to 0.
     * Expected result: no error thrown and the duration of stop of the
     * JBIProcessorManager is lower than the default value (15s).
     */
    @Test
    public void testStop_002() throws JBIException, MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null,
                null, 10, null, null, Long.valueOf(0), null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        final long start = System.currentTimeMillis();

        procMgr.stop();

        TestCase.assertTrue(
                "The stop duration the processor pool is higher than the default value (15s).",
                System.currentTimeMillis() - start <= JBIProcessorManager.DEFAULT_PROCESSOR_STOP_MAX_WAIT);

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * Checks the stop against a value of 'processor-stop-max-wait' set to a
     * negative value. Expected result: no error thrown and the duration of stop
     * of the JBIProcessorManager is lower than the default value (15s).
     */
    @Test
    public void testStop_003() throws JBIException, MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null,
                null, 10, null, null, Long.valueOf(-1500), null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        final long start = System.currentTimeMillis();

        procMgr.stop();

        TestCase.assertTrue(
                "The stop duration the processor pool is higher than the default value (15s).",
                System.currentTimeMillis() - start <= JBIProcessorManager.DEFAULT_PROCESSOR_STOP_MAX_WAIT);

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * <p>
     * Checks the resize of the message exchange processors, increasing the core
     * pool size.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the core size of the message exchange processors thread pool is size
     * to the new value,</li>
     * <li>the min idle number of message exchange processors pool is set to the
     * new value,</li>
     * <li>the number of started threads corresponds to the new value.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testSetCorePoolSize_000() throws JBIException,
            MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final int originalCorePoolSize = 5;
        final AbstractComponent componentMock = this.createComponentMock(null, null, null, null,
                null, originalCorePoolSize, null, null, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        TestCase.assertEquals(
                "The processor thread pool core size is not set to the expected value.",
                originalCorePoolSize, procMgr.getProcessorThreadPool().getCorePoolSize());
        TestCase.assertEquals(
                "The number of running thread of the message exchange processor thread pool is not the expected value.",
                originalCorePoolSize, procMgr.getProcessorThreadPool().getPoolSize());

        // Set a new value to the core pool size of the message exchange
        // processor thread pool
        final int newCorePoolSize = 12;
        procMgr.setCorePoolSize(newCorePoolSize);

        TestCase.assertEquals(
                "The processor thread pool core size is not correctly updated to the new value.",
                newCorePoolSize, procMgr.getProcessorThreadPool().getCorePoolSize());
        TestCase.assertEquals(
                "The number of running thread of the message exchange processor thread pool is not correctly updated to the new value.",
                newCorePoolSize, procMgr.getProcessorThreadPool().getPoolSize());
        TestCase.assertEquals(
                "The min idle of the message exchange processor pool is not correctly updated to the new value.",
                newCorePoolSize, procMgr.getProcessorObjectPool().getMinIdle());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * <p>
     * Checks the resize of the message exchange processors, decreasing the core
     * pool size.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the core size of the message exchange processors thread pool is sized
     * to the new value,</li>
     * <li>the min idle number of message exchange processors pool is set to the
     * new value,</li>
     * <li>the number of started threads corresponds to the new value.</li>
     * </ul>
     * </p>
     */
    @Ignore("The processor thread pool is not reduce because of JVM bug 6450200, 6576792: http://bugs.sun.com/view_bug.do?bug_id=6576792")
    @Test
    public void testSetCorePoolSize_001() throws JBIException,
            MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final int originalCorePoolSize = 50;
        final AbstractComponent componentMock = this.createComponentMock(null, null, null, null,
                null, originalCorePoolSize, null, null, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        TestCase.assertEquals(
                "The processor thread pool core size is not set to the expected value.",
                originalCorePoolSize, procMgr.getProcessorThreadPool().getCorePoolSize());
        TestCase.assertEquals(
                "The number of running thread of the message exchange processor thread pool is not the expected value.",
                originalCorePoolSize, procMgr.getProcessorThreadPool().getPoolSize());

        // Set a new value to the core pool size of the message exchange
        // processor thread pool
        final int newCorePoolSize = 12;
        procMgr.setCorePoolSize(newCorePoolSize);

        TestCase.assertEquals(
                "The processor thread pool core size is not correctly updated to the new value.",
                newCorePoolSize, procMgr.getProcessorThreadPool().getCorePoolSize());
        TestCase.assertEquals(
                "The number of running thread of the message exchange processor thread pool is not correctly updated to the new value.",
                newCorePoolSize, procMgr.getProcessorThreadPool().getPoolSize());
        TestCase.assertEquals(
                "The min idle of the message exchange processor pool is not correctly updated to the new value.",
                newCorePoolSize, procMgr.getProcessorObjectPool().getMinIdle());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * <p>
     * Checks the resize of the message exchange processors, increasing the max
     * pool size.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the max size of the message exchange processors thread pool is size
     * to the new value,</li>
     * <li>the max number of message exchange processors pool is set to the new
     * value added with the queue size.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testSetMaxCorePoolSize_000() throws JBIException,
            MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final int originalMaxPoolSize = 50;
        final AbstractComponent componentMock = this.createComponentMock(null, null, null, null,
                null, 5, originalMaxPoolSize, null, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        TestCase.assertEquals(
                "The processor thread pool max size is not set to the expected value.",
                originalMaxPoolSize, procMgr.getProcessorThreadPool().getMaximumPoolSize());

        // Set a new value to the core pool size of the message exchange
        // processor thread pool
        final int newMaxPoolSize = 62;
        procMgr.setMaxCorePoolSize(newMaxPoolSize);

        TestCase.assertEquals(
                "The processor thread pool max size is not correctly updated to the new value.",
                newMaxPoolSize, procMgr.getProcessorThreadPool().getMaximumPoolSize());
        TestCase.assertEquals(
                "The max number of the message exchange processor pool is not correctly updated to the new value.",
                newMaxPoolSize, procMgr.getProcessorObjectPool().getMaxActive());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * <p>
     * Checks the resize of the message exchange processors, decreasing the max
     * pool size.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the max size of the message exchange processors thread pool is sized
     * to the new value,</li>
     * <li>the max number of message exchange processors pool is set to the new
     * value added with the queue size.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testSetMaxCorePoolSize_001() throws JBIException,
            MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final int originalMaxPoolSize = 50;
        final AbstractComponent componentMock = this.createComponentMock(null, null, null, null,
                null, 5, originalMaxPoolSize, null, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        TestCase.assertEquals(
                "The processor thread pool max size is not set to the expected value.",
                originalMaxPoolSize, procMgr.getProcessorThreadPool().getMaximumPoolSize());

        // Set a new value to the core pool size of the message exchange
        // processor thread pool
        final int newMaxPoolSize = 42;
        procMgr.setMaxCorePoolSize(newMaxPoolSize);

        TestCase.assertEquals(
                "The processor thread pool max size is not correctly updated to the new value.",
                newMaxPoolSize, procMgr.getProcessorThreadPool().getMaximumPoolSize());
        TestCase.assertEquals(
                "The max number of the message exchange processor pool is not correctly updated to the new value.",
                newMaxPoolSize, procMgr.getProcessorObjectPool().getMaxActive());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * <p>
     * Checks the reconfiguration of the message exchange processors, increasing
     * the keep alive time.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the keep alive time of the message exchange processors thread pool is
     * size to the new value,</li>
     * <li>the max number of message exchange processors pool is set to the new
     * value added with the queue size.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testSetKeepAliveTime_000() throws JBIException,
            MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final int originalKeepAliveTime = 500;
        final AbstractComponent componentMock = this.createComponentMock(null, null, null, null,
                null, 5, null, originalKeepAliveTime, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        TestCase.assertEquals(
                "The processor thread pool keep alive time is not set to the expected value.",
                originalKeepAliveTime, procMgr.getProcessorThreadPool().getKeepAliveTime(TimeUnit.SECONDS));

        // Set a new value to the core pool size of the message exchange
        // processor thread pool
        final int newKeepAliveTime = 999;
        procMgr.setPoolKeepAliveTime(newKeepAliveTime);

        TestCase.assertEquals(
                "The processor thread pool keep alive time is not correctly updated to the new value.",
                newKeepAliveTime, procMgr.getProcessorThreadPool().getKeepAliveTime(TimeUnit.SECONDS));
        TestCase.assertEquals(
                "The min eviectable time of the message exchange processor pool is not correctly updated to the new value.",
                newKeepAliveTime * 1000, procMgr.getProcessorObjectPool().getMinEvictableIdleTimeMillis());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * <p>
     * Checks the reconfiguration of the message exchange processors, decreasing
     * the keep alive time.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the keep alive time of the message exchange processors thread pool is
     * size to the new value,</li>
     * <li>the max number of message exchange processors pool is set to the new
     * value added with the queue size.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testSetKeepAliveTime_001() throws JBIException,
            MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final int originalKeepAliveTime = 500;
        final AbstractComponent componentMock = this.createComponentMock(null, null, null, null,
                null, 5, null, originalKeepAliveTime, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        TestCase.assertEquals(
                "The processor thread pool keep alive time is not set to the expected value.",
                originalKeepAliveTime, procMgr.getProcessorThreadPool().getKeepAliveTime(TimeUnit.SECONDS));

        // Set a new value to the core pool size of the message exchange
        // processor thread pool
        final int newKeepAliveTime = 249;
        procMgr.setPoolKeepAliveTime(newKeepAliveTime);

        TestCase.assertEquals(
                "The processor thread pool keep alive time is not correctly updated to the new value.",
                newKeepAliveTime, procMgr.getProcessorThreadPool().getKeepAliveTime(TimeUnit.SECONDS));
        TestCase.assertEquals(
                "The min eviectable time of the message exchange processor pool is not correctly updated to the new value.",
                newKeepAliveTime * 1000, procMgr.getProcessorObjectPool().getMinEvictableIdleTimeMillis());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * <p>
     * Checks the resize of the message exchange processor pool max size with a
     * value set to 0.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no exception thrown,</li>
     * <li>message exchange processor pool max size is not changed,</li>
     * </ul>
     * </p>
     */
    @Test
    public void testSetMessageProcessorMaxPoolSize_000() throws JBIException,
            MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final int initialMessageExchangeMaxPoolSize = 50;

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null,
                initialMessageExchangeMaxPoolSize, 10, 30, null, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        procMgr.setMessageProcessorMaxPoolSize(0);

        TestCase.assertEquals("The message processor object pool max size has changed.",
                initialMessageExchangeMaxPoolSize, procMgr.getProcessorObjectPool().getMaxActive());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * <p>
     * Checks the resize of the message exchange processor pool max size with a
     * value set to a negative value.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no exception thrown,</li>
     * <li>message exchange processor pool max size is not changed,</li>
     * </ul>
     * </p>
     */
    @Test
    public void testSetMessageProcessorMaxPoolSize_001() throws JBIException,
            MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final int initialMessageExchangeMaxPoolSize = 50;

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null,
                initialMessageExchangeMaxPoolSize, 10, 30, null, null, null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        procMgr.setMessageProcessorMaxPoolSize(-10);

        TestCase.assertEquals("The message processor object pool max size has changed.",
                initialMessageExchangeMaxPoolSize, procMgr.getProcessorObjectPool().getMaxActive());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * <p>
     * Checks the resize of the message exchange processor pool max size with a
     * value upper than the max size of the processor thread pool.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>message exchange processor pool max is set to the value set.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testSetMessageProcessorMaxPoolSize_002() throws JBIException,
            MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final int initialMessageExchangeMaxPoolSize = 50;
        final int initialProcessorMaxPoolSize = 30;

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null,
                initialMessageExchangeMaxPoolSize, 10, initialProcessorMaxPoolSize, null, null,
                null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        final int newMessageExchangeMaxPoolSize = initialProcessorMaxPoolSize + 10;
        procMgr.setMessageProcessorMaxPoolSize(newMessageExchangeMaxPoolSize);

        TestCase.assertEquals("The message processor object pool max size is not correctly set",
                newMessageExchangeMaxPoolSize, procMgr.getProcessorObjectPool().getMaxActive());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

    /**
     * <p>
     * Checks the resize of the message exchange processor pool max size with a
     * value set to a value lower than the max size of the processor thread
     * pool.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no exception thrown,</li>
     * <li>message exchange processor pool max size is not changed,</li>
     * </ul>
     * </p>
     * s
     */
    @Test
    public void testSetMessageProcessorMaxPoolSize_003() throws JBIException,
            MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final int initialMessageExchangeMaxPoolSize = 50;
        final int initialProcessorMaxPoolSize = 30;

        final AbstractComponent componentMock = this.createComponentMock(10, null, null, null,
                null, initialMessageExchangeMaxPoolSize, initialProcessorMaxPoolSize, null, null,
                null);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock,
                new RuntimeConfigurationNotifier(null, null, null), LOG);

        procMgr.start();

        final int newMessageExchangeMaxPoolSize = initialProcessorMaxPoolSize - 10;
        procMgr.setMessageProcessorMaxPoolSize(newMessageExchangeMaxPoolSize);

        TestCase.assertEquals("The message processor object pool max size has changed",
                initialMessageExchangeMaxPoolSize, procMgr.getProcessorObjectPool().getMaxActive());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);

    }

}
