/**
 * Copyright (c) 2007-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.util;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.fail;

import java.net.URI;
import java.net.URISyntaxException;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;

import org.junit.Test;
import org.ow2.petals.component.framework.api.exception.FaultException;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.exception.SOAP11FaultClientException;
import org.ow2.petals.component.framework.api.exception.SOAP11FaultServerException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * @author Christophe Deneux - Cap Gemini
 */
public class SourceUtilTest {

    /**
     * Test the soap fault generation based on a SOAP11FAultServerException
     */
    @Test
    public void testcreateSoapFault_001() {
        try {
            final SourceUtil sourceUtilImpl = new SourceUtil();
            final Exception e = new SOAP11FaultServerException("My error",
                    new URI("http://myActor"));
            final Document soapFaultDocument = sourceUtilImpl.createSoapFaultDocument(e, new QName(
                    ""));
            this.compareSOAPFault(soapFaultDocument, new QName(
                    "http://schemas.xmlsoap.org/soap/envelope/", "Server"), // faultcode
                    "My error", // faultstring
                    new URI("http://myActor") // faultactor
            );
        } catch (final URISyntaxException e) {
            fail(e.getMessage());
        }
    }

    /**
     * Test the soap fault generation based on a SOAP11FaultClientException
     */
    @Test
    public void testcreateSoapFault_002() {
        try {
            final SourceUtil sourceUtilImpl = new SourceUtil();
            final Exception e = new SOAP11FaultClientException("My error",
                    new URI("http://myActor"));
            final Document soapFaultDocument = sourceUtilImpl.createSoapFaultDocument(e, new QName(
                    ""));

            this.compareSOAPFault(soapFaultDocument, new QName(
                    "http://schemas.xmlsoap.org/soap/envelope/", "Client"), // faultcode
                    "My error", // faultstring
                    new URI("http://myActor") // faultactor
            );

        } catch (final URISyntaxException e) {
            fail(e.getMessage());
        }
    }

    /**
     * Test the soap fault generation based on a SOAP11FaultClientException
     * containing a message including special characters nit authorized in XML.
     */
    @Test
    public void testcreateSoapFault_003() {
        try {
            final String faultString = "My error containg special characters [><&]";
            final SourceUtil sourceUtilImpl = new SourceUtil();
            final Exception e = new SOAP11FaultClientException(faultString, new URI(
                    "http://myActor"));
            final Document soapFaultDocument = sourceUtilImpl.createSoapFaultDocument(e, new QName(
                    ""));
            this.compareSOAPFault(soapFaultDocument, new QName(
                    "http://schemas.xmlsoap.org/soap/envelope/", "Client"), // faultcode
                    faultString, // faultstring
                    new URI("http://myActor") // faultactor
            );

        } catch (final URISyntaxException e) {
            fail(e.getMessage());
        }
    }

    /**
     * Test the soap fault generation based on a custom FaultException
     */
    @Test
    public void testcreateSoapFault_004() {
        try {
            final SourceUtil sourceUtilImpl = new SourceUtil();
            final Exception e = new MyCustomFaultException("My error", new URI("http://myActor"));

            final Document soapFaultDocument = sourceUtilImpl.createSoapFaultDocument(e, new QName(
                    ""));
            this.compareSOAPFault(soapFaultDocument, new QName("http://my-namespace.com",
                    "my-local-part"), // faultcode
                    "My error", // faultstring
                    new URI("http://myActor") // faultactor
            );
        } catch (final URISyntaxException e) {
            fail(e.getMessage());
        }
    }

    private void compareSOAPFault(final Document document, final QName expectedFaultCode,
            final String expectedFaultString, final URI expectedFaultActor) {
        assertNotNull(document);
        final Element faultElement = document.getDocumentElement();
        assertNotNull(faultElement);
        assertEquals("Fault", faultElement.getLocalName());
        assertEquals("http://schemas.xmlsoap.org/soap/envelope/", faultElement.getNamespaceURI());
        final Node faultCodeNode = faultElement.getFirstChild();
        assertNotNull(faultCodeNode);
        assertEquals("faultcode", faultCodeNode.getNodeName());
        assertNull(faultCodeNode.getNamespaceURI());
        final Node faultCodeValueNode = faultCodeNode.getFirstChild();
        assertNotNull(faultCodeValueNode);
        final String faultCodeValuePrefix = faultCodeValueNode.lookupPrefix(expectedFaultCode
                .getNamespaceURI());
        assertEquals(faultCodeValuePrefix + ":" + expectedFaultCode.getLocalPart(),
                faultCodeValueNode.getNodeValue());
        final Node faultStringNode = faultCodeNode.getNextSibling();
        assertNotNull(faultStringNode);
        assertEquals("faultstring", faultStringNode.getNodeName());
        assertNull(faultStringNode.getNamespaceURI());
        final Node faultStringValueNode = faultStringNode.getFirstChild();
        assertNotNull(faultStringValueNode);
        assertEquals(expectedFaultString, faultStringValueNode.getNodeValue());
        final Node faultActorNode = faultStringNode.getNextSibling();
        assertNotNull(faultActorNode);
        assertEquals("faultactor", faultActorNode.getNodeName());
        assertNull(faultActorNode.getNamespaceURI());
        final Node faultActorValueNode = faultActorNode.getFirstChild();
        assertNotNull(faultActorValueNode);
        assertEquals(expectedFaultActor.toString(), faultActorValueNode.getNodeValue());
    }

    /**
     * test the createDocument method
     */
    @Test
    public void testCreateDocument() {
        try {
            final SourceUtil sourceUtilImpl = new SourceUtil();
            final Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder()
                    .newDocument();
            final Element element = doc.createElement("test");
            element.setTextContent("coucou");
            doc.appendChild(element);
            final Document finalDocument = sourceUtilImpl.createDocument(new DOMSource(doc));
            assertNotNull(finalDocument);
            assertEquals("coucou", finalDocument.getDocumentElement().getTextContent());
        } catch (final ParserConfigurationException e) {
            fail(e.getMessage());
        } catch (final PEtALSCDKException e) {
            fail(e.getMessage());
        }
    }

    /**
     * test the createDocument method without fork
     */
    @Test
    public void testCreateDocumen2() {
        try {
            final SourceUtil sourceUtilImpl = new SourceUtil();
            final Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder()
                    .newDocument();
            final Element element = doc.createElement("test");
            element.setTextContent("coucou");
            doc.appendChild(element);
            final Document finalDocument = sourceUtilImpl.createDocument(new DOMSource(doc), false);
            assertNotNull(finalDocument);
            assertEquals(doc, finalDocument);
        } catch (final ParserConfigurationException e) {
            fail(e.getMessage());
        } catch (final PEtALSCDKException e) {
            fail(e.getMessage());
        }
    }

    @Test
    public void testCreateSourceString() {
        final String content = "<?xml version=\"1.0\" encoding=\"UTF-8\"?><result xmlns=\"http://petals.ow2.org/petals-se-eip/scatter-gather\"><EnvoiESB xmlns=\"http://finances.gouv.fr/dgme/esb/webapp/webservice/wsdl/envoiesb\" xmlns:env=\"http://finances.gouv.fr/dgme/esb/webapp/webservice/wsdl/envoiesb\" xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\">"
                + "    <informationsGF xmlns=\"\">"
                + "        <idEnveloppe>125</idEnveloppe>"
                + "        <numeroDossier>GFDG</numeroDossier>"
                + "        <destinataire>principal</destinataire>"
                + "        <idEmetteur>GF</idEmetteur>"
                + "    </informationsGF>"
                + "    <informationsSI xmlns=\"\">"
                + "        <idDestinataire>11111111111125</idDestinataire>"
                + "        <idService>3</idService>"
                + "        <idGuichet>1</idGuichet>"
                + "    </informationsSI>"
                + "    <mail xmlns=\"\">"
                + "        <expediteur>component-test@devmail.com</expediteur>"
                + "        <destinataire>component-test@devmail.com</destinataire>"
                + "        <sujet>sujet</sujet>"
                + "        <corps>corps</corps>"
                + "    </mail>"
                + "    <listeFichiers xmlns=\"\">"
                + "        <fichier>"
                + "                <nom>38001-A-8-4E0SNRYB-ent-0.xml</nom>"
                + "        </fichier>"
                + "        <fichier>"
                + "                <nom>38002-A-8-4E0SNRYB-ent-0.xml</nom>"
                + "        </fichier>"
                + "        <fichier>"
                + "                <nom>38003-A-8-4E0SNRYB-ent-0.xml</nom>"
                + "        </fichier>"
                + "    </listeFichiers>" + "</EnvoiESB></result>";
        try {
            final Source source = SourceUtil.createSource(content);
            assertNotNull(source);
        } catch (final PEtALSCDKException e) {
            fail("unable to create the source: " + e.getMessage());
        }
    }

    @Test
    public void testCreateDocumentSource() {
        final String content = "<?xml version=\"1.0\" encoding=\"UTF-8\"?><result xmlns=\"http://petals.ow2.org/petals-se-eip/scatter-gather\"><EnvoiESB xmlns=\"http://finances.gouv.fr/dgme/esb/webapp/webservice/wsdl/envoiesb\" xmlns:env=\"http://finances.gouv.fr/dgme/esb/webapp/webservice/wsdl/envoiesb\" xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\">"
                + "    <informationsGF xmlns=\"\">"
                + "        <idEnveloppe>125</idEnveloppe>"
                + "        <numeroDossier>GFDG</numeroDossier>"
                + "        <destinataire>principal</destinataire>"
                + "        <idEmetteur>GF</idEmetteur>"
                + "    </informationsGF>"
                + "    <informationsSI xmlns=\"\">"
                + "        <idDestinataire>11111111111125</idDestinataire>"
                + "        <idService>3</idService>"
                + "        <idGuichet>1</idGuichet>"
                + "    </informationsSI>"
                + "    <mail xmlns=\"\">"
                + "        <expediteur>component-test@devmail.com</expediteur>"
                + "        <destinataire>component-test@devmail.com</destinataire>"
                + "        <sujet>sujet</sujet>"
                + "        <corps>corps</corps>"
                + "    </mail>"
                + "    <listeFichiers xmlns=\"\">"
                + "        <fichier>"
                + "                <nom>38001-A-8-4E0SNRYB-ent-0.xml</nom>"
                + "        </fichier>"
                + "        <fichier>"
                + "                <nom>38002-A-8-4E0SNRYB-ent-0.xml</nom>"
                + "        </fichier>"
                + "        <fichier>"
                + "                <nom>38003-A-8-4E0SNRYB-ent-0.xml</nom>"
                + "        </fichier>"
                + "    </listeFichiers>" + "</EnvoiESB></result>";
        try {
            final Source source = SourceUtil.createSource(content);
            final Document document = SourceUtil.createDocument(source);
            assertNotNull(document);
        } catch (final PEtALSCDKException e) {
            fail("unable to create the source: " + e.getMessage());
        }
    }

    private static class MyCustomFaultException extends FaultException {
        private static final long serialVersionUID = 5724917270299632484L;

        public MyCustomFaultException(final String faultString, final URI faultActor) {
            super(new QName("http://my-namespace.com", "my-local-part"), faultString, faultActor);
        }
    }
}
